(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)


(***************************************************************************)
(* Simplified Typed TAL *)
(***************************************************************************)

(***************************************************************************)
(* Utilities *)
Require Export Coq.Lists.PolyList.
Require Export map.


(***************************************************************************)
(* Syntax of TAL *)

Require Export talreg.

Syntactic Definition label := nat.
Syntactic Definition int   := nat.
Syntactic Definition initflag := bool.


(* the allocation pointer register is not merged into the normal register
   file - it just contains a label with a special type *)
Definition AP := label.   
Inductive APTy : Set := fresh : APTy  
      	       	     |  used : nat -> APTy.


(* lifted types *)
Inductive OmegaL : nat -> Set
  := varL   : (i:nat) (OmegaL (S i))
  |  liftL  : (i:nat) (OmegaL i) -> (OmegaL (S i))
  |  inttyL : (OmegaL O)
  | codetyL : (i:nat) (OmegaL_Map i) -> APTy -> (OmegaL i)
  |  tuptyL : (i:nat) (OmegaL_List i) -> (list initflag) -> (OmegaL i)
  |  rectyL : (i:nat) (OmegaL (S i)) -> (OmegaL i)
with OmegaL_Map : nat -> Set
  := memptyL : (i:nat) (OmegaL_Map i)
  |  melemL  : (i:nat) Reg -> (OmegaL i) -> (OmegaL_Map i) -> (OmegaL_Map i)
with OmegaL_List : nat -> Set
  := nilL : (i:nat) (OmegaL_List i)
  | consL : (i:nat) (OmegaL i) -> (OmegaL_List i) -> (OmegaL_List i)
.

Syntactic Definition OmegaR := (OmegaL (S O)).

(* the actual types *)
Inductive Omega  : Set := intty  : Omega
      	       	       | codety : (Map Reg Omega) -> APTy -> Omega
      	       	       | tupty  : (list Omega) -> (list initflag) -> Omega
		       | recty  : (OmegaL (S O)) -> Omega
.

Load rftal.

Syntactic Definition RegFileTy := (Map Reg Omega).

Inductive WordVal : Set
  := wl : label -> WordVal
  |  wi : int -> WordVal
  |  wuninit : Omega -> WordVal
  |  wfold : WordVal -> Omega -> WordVal.


Inductive Instr : Set
  := add  : Reg -> Reg -> Reg -> Instr
  |  addi : Reg -> Reg -> int -> Instr
  |  alloc : Reg -> (list Omega) -> Instr
  |  bgt  : Reg -> Reg -> label -> Instr
  |  bump : int -> Instr
  |  fold : Reg -> Omega -> Reg -> Instr
  |  ld   : Reg -> Reg -> int -> Instr
  |  mov  : Reg -> Reg -> Instr
  |  movi : Reg -> int -> Instr
  |  movl : Reg -> label -> Instr
  |  st   : Reg -> int -> Reg -> Instr
  |  unfold : Reg -> Reg -> Instr
.

Inductive InstrSeq : Set
  := iseq : Instr -> InstrSeq -> InstrSeq
  |  jd   : label -> InstrSeq
  |  jmp  : Reg -> InstrSeq.


Inductive HeapVal : Set
  := tuple : (list WordVal) -> HeapVal
  |  code  : RegFileTy -> APTy -> InstrSeq -> HeapVal.


(* RegFile environment ---------------*)
Definition RegFile := (Map Reg WordVal).

(* RegFileTy defined with Omega above *)


(* Heap environment ------------------*)
Definition Heap := (WFMap label HeapVal).

Syntactic Definition hunwrap := (unwrapMap label HeapVal).
Syntactic Definition hsize := (msize label HeapVal).
Syntactic Definition hindom  := (mindom label HeapVal).
Syntactic Definition hnindom := (mnotindom label HeapVal).
Syntactic Definition hlookup := (mlookup label HeapVal).
Syntactic Definition hupdate := (mupdate label HeapVal).
Syntactic Definition hextend := (mextend label HeapVal).
Syntactic Definition hupdext := (mupdext label HeapVal).


(* Heap type environment -------------*)
Definition HeapTy := (WFMap label Omega).

Syntactic Definition htunwrap := (unwrapMap label Omega).
Syntactic Definition htsize := (msize label Omega).
Syntactic Definition htindom  := (mindom label Omega).
Syntactic Definition htnindom := (mnotindom label Omega).
Syntactic Definition htlookup := (mlookup label Omega).
Syntactic Definition htupdate := (mupdate label Omega).
Syntactic Definition htextend := (mextend label Omega).
Syntactic Definition htupdext := (mupdext label Omega).


(* Register stuff ------------------- *)
Syntactic Definition reglookup := (mlookup ? ?).
Syntactic Definition regupdext := (mupdext ? ?).


(* TAL Program -----------------------*)
Definition Program := (Heap * (RegFile * (AP * InstrSeq))).


(***************************************************************************)
(* more stuff for the heap *)
Syntactic Definition HeapMap := (Map label HeapVal).

Inductive OrdHeap : HeapMap -> Prop
  := ordheap0 : (OrdHeap (mempty ? ?))
  |  ordheap1 : (l:label; hv:HeapVal; h:(Map label HeapVal))
		(msize ? ? h l) ->
      	       	(OrdHeap h) ->
      	       	(OrdHeap (melem ? ? l hv h))
.

Inductive PropSubHeap : HeapMap -> HeapMap -> Prop
  := propsubh0 : (h:HeapMap; l:label; hv:HeapVal) 
      	       	 (PropSubHeap h (melem ? ? l hv h))
  |  propsubh1 : (h,h':HeapMap; l:label; hv:HeapVal)
      	       	 (PropSubHeap h h') ->
		 (PropSubHeap h (melem ? ? l hv h')).

Inductive SubHeap : HeapMap -> HeapMap -> Prop
  := subheap0 : (h:HeapMap) (SubHeap h h)
  |  subheap1 : (h,h':HeapMap) (PropSubHeap h h') -> (SubHeap h h').




(***************************************************************************)
(* length and positions of an instruction sequence *)
(* and other utilities *)
Inductive ISubDepth : InstrSeq -> InstrSeq -> nat -> Prop
  := isubd0 : (I:InstrSeq) (ISubDepth I I O)
  |  isubdS : (I,I':InstrSeq; i:Instr; n:nat)
      	      (ISubDepth I' I n) ->
	      (ISubDepth I' (iseq i I) (S n)).

Fixpoint lenInstrSeq [I:InstrSeq] : nat
  := Cases I of (iseq _ I') => (S (lenInstrSeq I'))
              | _ => (S O)
     end.

Fixpoint getInstrPos [n:nat] : InstrSeq -> InstrSeq
  := [I] Cases n of O => I
      	       	 | (S n') => Cases I of (iseq _ I') => (getInstrPos n' I')
		                      | _ => I end
         end.

Inductive ListNth [A:Set] : (list A) -> nat -> A -> Prop
  := listnth0 : (V:(list A); a:A) (ListNth A (cons a V) O a)
  |  listnth1 : (V:(list A); a,a':A; n:nat) 
      	       	(ListNth A V n a') ->
		(ListNth A (cons a V) (S n) a').


Fixpoint makeUninitTup [V:(list Omega)] : (list WordVal)
  := Cases V of nil => (nil WordVal)
              | (cons t V') => (cons (wuninit t) (makeUninitTup V'))
     end.

Fixpoint makeUninitTupty [V:(list Omega)] : (list initflag)
  := Cases V of nil => (nil initflag)
      	      | (cons t V') => (cons false (makeUninitTupty V'))
     end.

Inductive updatetuple
  : (list WordVal) -> int -> WordVal -> (list WordVal) -> Prop
  := updtup0 : (V:(list WordVal); w,w':(WordVal))
      	       (updatetuple (cons w V) O w' (cons w' V))
  |  updtup1 : (V,V':(list WordVal); w,w':(WordVal); i:int)
      	       (updatetuple V i w' V') ->
	       (updatetuple (cons w V) (S i) w' (cons w V')).


Inductive updatetupty  (* used for static semantics *)
  : (list initflag) -> int -> (list initflag) -> Prop
  := updtpt0 : (V:(list initflag); b:initflag)
      	       (updatetupty (cons b V) O (cons true V))
  |  updtpt1 : (V,V':(list initflag); b:initflag; i:int)
      	       (updatetupty V i V') ->
	       (updatetupty (cons b V) (S i) (cons b V')).




(***************************************************************************)
(* Operational Semantics *)
Inductive Eval : Program -> Program -> Prop
  := ev_add    : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	 (rd,rs,rs':Reg; rsval,rsval':int)
		 (reglookup R rs (wi rsval)) ->
		 (reglookup R rs' (wi rsval')) ->
		 (regupdext R rd (wi (plus rsval rsval')) R') ->
      	       	 (Eval (H, (R, (r31, (iseq (add rd rs rs') I'))))
		       (H, (R', (r31, I'))))
  |  ev_addi   : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	 (rd,rs:Reg; i:int; rsval:int)
		 (reglookup R rs (wi rsval)) ->
		 (regupdext R rd (wi (plus rsval i)) R') ->
		 (Eval (H, (R, (r31, (iseq (addi rd rs i) I'))))
		       (H, (R', (r31, I'))))

  |  ev_alloc  : (H,H':Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	 (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h'))
      	       	 (rd:Reg; V:(list Omega))
		 H=(wfmap ? ? h wfh) ->
		 H'=(wfmap ? ? h' wfh') ->
		 (regupdext R rd (wl r31) R') ->
		 (hextend h r31 (tuple (makeUninitTup V)) h') ->
      	       	 (Eval (H, (R, (r31, (iseq (alloc rd V) I'))))
		       (H', (R', (r31, I'))))
  | ev_bgt0   : (H:Heap; R:RegFile; r31:AP; I':InstrSeq)
      	       	(rs,rt:Reg; l:label; rsval,rtval:int)
		(reglookup R rs (wi rsval)) ->
		(reglookup R rt (wi rtval)) ->
		(le rsval rtval) ->
		(Eval (H, (R, (r31, (iseq (bgt rs rt l) I'))))
		      (H, (R, (r31, I'))))
  | ev_bgt1   : (H:Heap; R:RegFile; r31:AP; I':InstrSeq)
      	       	(G:RegFileTy; T:APTy; I'':InstrSeq)
      	       	(rs,rt:Reg; l:label; rsval,rtval:int)
		(reglookup R rs (wi rsval)) ->
		(reglookup R rt (wi rtval)) ->
		(gt rsval rtval) ->
		(hlookup (hunwrap H) l (code G T I'')) ->
		(Eval (H, (R, (r31, (iseq (bgt rs rt l) I'))))
		      (H, (R, (r31, I''))))
		
  | ev_bump   : (H:Heap; R:RegFile; r31:AP; I':InstrSeq)
      	       	(i:int; l:nat)
		(hsize (hunwrap H) l) ->
		(Eval (H, (R, (r31, (iseq (bump i) I'))))
		      (H, (R, (l, I'))))
  | ev_fold   : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	 (rd,rs:Reg; t:Omega; rsval:WordVal)
		 (reglookup R rs rsval) ->
		 (regupdext R rd (wfold rsval t) R') ->
		 (Eval (H, (R, (r31, (iseq (fold rd t rs) I'))))
		       (H, (R', (r31, I'))))

  | ev_jd     : (H:Heap; R:RegFile; r31:AP)
     	       	(l:label; G:RegFileTy; T:APTy; I':InstrSeq)
		(hlookup (hunwrap H) l (code G T I')) ->
		(Eval (H, (R, (r31, (jd l))))
		      (H, (R, (r31, I'))))
  | ev_jmp    : (H:Heap; R:RegFile; r31:AP)
      	       	(r:Reg; l:label; G:RegFileTy; T:APTy; I':InstrSeq)
		(reglookup R r (wl l)) ->
		(hlookup (hunwrap H) l (code G T I')) ->
		(Eval (H, (R, (r31, (jmp r))))
		      (H, (R, (r31, I'))))

  | ev_ld     : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	(rd,rs:Reg; i:int)
		(l:label; V:(list WordVal); v:WordVal)
		(reglookup R rs (wl l)) ->
		(hlookup (hunwrap H) l (tuple V)) ->
		(ListNth ? V i v) ->
		(regupdext R rd v R') ->		
      	       	(Eval (H, (R, (r31, (iseq (ld rd rs i) I'))))
		      (H, (R',(r31, I'))))

  | ev_mov    : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	(rd,rs:Reg; v:WordVal)
		(reglookup R rs v) ->
		(regupdext R rd v R') ->
		(Eval (H, (R, (r31, (iseq (mov rd rs) I'))))
		      (H, (R',(r31, I'))))
  | ev_movi   : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	(rd:Reg; i:int)
		(regupdext R rd (wi i) R') ->
		(Eval (H, (R, (r31, (iseq (movi rd i) I'))))
		      (H, (R',(r31, I'))))
  | ev_movl   : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	(rd:Reg; l:label)
		(regupdext R rd (wl l) R') ->
		(Eval (H, (R, (r31, (iseq (movl rd l) I'))))
		      (H, (R',(r31, I'))))

  | ev_store  : (H,H':Heap; R:RegFile; r31:AP; I':InstrSeq)
      	       	(h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h'))
		(rd,rs:Reg; i:int; l:label; V,V':(list WordVal); w:WordVal)
		H=(wfmap ? ? h wfh) ->
		H'=(wfmap ? ? h' wfh') ->
		(reglookup R rd (wl l)) ->
		(reglookup R rs w) ->
		(hlookup h l (tuple V)) ->
      	       	(updatetuple V i w V') ->
		(hupdate h l (tuple V') h') ->
      	       	(Eval (H, (R, (r31, (iseq (st rd i rs) I'))))
      	       	      (H',(R, (r31, I'))))

  | ev_unfold : (H:Heap; R,R':RegFile; r31:AP; I':InstrSeq)
      	       	(rd,rs:Reg; v:WordVal; t:Omega)
		(reglookup R rs (wfold v t)) ->
		(regupdext R rd v R') ->
		(Eval (H, (R, (r31, (iseq (unfold rd rs) I'))))
		      (H, (R', (r31, I'))))

.




(***************************************************************************)
(* Static Semantics *)

(* register file subtyping *)                                  (*  G <= G' *)
Inductive RegFileSubtype : RegFileTy -> RegFileTy -> Prop
  := weaken      : (G,G':RegFileTy)
		   ((r:Reg; t:Omega)
                    (mlookup ? ? G' r t) -> (mlookup ? ? G r t)) ->
      	       	   (RegFileSubtype G G'). 


(* subtyping *)                                              (* |- t <= t' *)
Inductive SubtypeB : initflag -> initflag -> Prop
  := subb_refl : (i:initflag) (SubtypeB i i)
  |  subb_true : (SubtypeB true false).


Inductive SubtypeBList : (list initflag) -> (list initflag) -> Prop
  := subb_nil  : (SubtypeBList (nil ?) (nil ?))
  |  subb_cons : (L1,L2:(list initflag); i1,i2:initflag)
      	       	 (SubtypeB i1 i2) ->
		 (SubtypeBList L1 L2) ->
		 (SubtypeBList (cons i1 L1) (cons i2 L2)).


Inductive Subtype : Omega -> Omega -> Prop
  := reflex       : (t:Omega) (Subtype t t)
  |  tuple_sub    : (ol:(list Omega); il1,il2:(list initflag))
      	       	    (SubtypeBList il1 il2) ->
		    (Subtype (tupty ol il1) (tupty ol il2)).


(* well-formed word values of type *)                  (* HT |- w : t wval *)
Inductive WFWordVal : HeapTy -> WordVal -> Omega -> Prop
  := int_wval   : (HT:HeapTy; i:int) (WFWordVal HT (wi i) intty)
  |  label_wval : (HT:HeapTy; l:label; t,t':Omega)
      	       	  (htlookup (htunwrap HT) l t') ->
		  (Subtype t' t) ->
		  (WFWordVal HT (wl l) t)
  |  fold_word_wval
      	       	: (HT:HeapTy)
		  (w:WordVal;  t:OmegaR; t':Omega)
      	       	  (RUnlift (RUnfold t))=t' ->
		  (WFWordVal HT w t') ->
		  (WFWordVal HT (wfold w (recty t)) (recty t)).


(* well-formed initialized word values of type *)
Inductive WFWordValinit : HeapTy -> WordVal -> Omega -> initflag -> Prop
  := init       : (HT:HeapTy)
      	       	  (w:WordVal; t:Omega; f:initflag)
		  (WFWordVal HT w t) ->
		  (WFWordValinit HT w t f)
  |  uninit     : (HT:HeapTy)
      	       	  (t:Omega)
      	       	  (WFWordValinit HT (wuninit t) t false).



(* well-formed instruction sequences *)                     (* HT; G |- I *)
Inductive WFInstrSeq : HeapTy -> RegFileTy -> APTy -> InstrSeq -> Prop
  := s_add      : (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd,rs,rs':Reg)
		  (mlookup ? ? G rs intty) ->
		  (mlookup ? ? G rs' intty) ->
		  (mupdext ? ? G rd intty G') ->
	          (WFInstrSeq HT G' T I) ->
		  (WFInstrSeq HT G T (iseq (add rd rs rs') I))
  |  s_addi     : (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd,rs:Reg; i:int)
		  (mlookup ? ? G rs intty) ->
		  (mupdext ? ? G rd intty G') ->
	          (WFInstrSeq HT G' T I) ->
		  (WFInstrSeq HT G T (iseq (addi rd rs i) I))

  |  s_alloc    : (HT:HeapTy; G,G':RegFileTy; I:InstrSeq)
      	       	  (rd:Reg; n:nat; V:(list Omega))
		  n=(length V) ->
		  (mupdext ? ? G rd (tupty V (makeUninitTupty V)) G') ->
		  (WFInstrSeq HT G' (used n) I) ->
      	       	  (WFInstrSeq HT G fresh (iseq (alloc rd V) I))
  |  s_bgt      : (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rs,rt:Reg; l:label)
		  (mlookup ? ? G rs intty) ->
		  (mlookup ? ? G rt intty) ->
		  (htlookup (htunwrap HT) l (codety G' T)) ->
		  (RegFileSubtype G G') ->
		  (WFInstrSeq HT G T I) ->
		  (WFInstrSeq HT G T (iseq (bgt rs rt l) I))

  |  s_bump     : (HT:HeapTy; G:RegFileTy; I:InstrSeq; n:nat)
      	       	  (WFInstrSeq HT G fresh I) ->  
      	       	  (WFInstrSeq HT G (used n) (iseq (bump n) I))
  |  s_fold     : (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd,rs:Reg; f:OmegaR; rst:Omega)
		  (mlookup ? ? G rs rst) ->
		  (RUnlift (RUnfold f))=rst ->
      	       	  (mupdext ? ? G rd (recty f) G') ->
		  (WFInstrSeq HT G' T I) ->
      	       	  (WFInstrSeq HT G T (iseq (fold rd (recty f) rs) I))

  |  s_jd       : (HT:HeapTy; G,G':RegFileTy; T:APTy)
       	       	  (l:label)
		  (htlookup (htunwrap HT) l (codety G' T)) ->
		  (RegFileSubtype G G') ->  
      	       	  (WFInstrSeq HT G T (jd l))
  |  s_jmp      : (HT:HeapTy; G,G':RegFileTy; T:APTy)
       	       	  (r:Reg)
		  (mlookup ? ? G r (codety G' T)) ->
		  (RegFileSubtype G G') ->  
      	       	  (WFInstrSeq HT G T (jmp r))

  |  s_ld	: (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd,rs:Reg; i:int)
      	       	  (ol:(list Omega); il:(list initflag); t:Omega)
		  (mlookup ? ? G rs (tupty ol il)) ->
		  (ListNth ? ol i t) ->
		  (ListNth ? il i true) ->
		  (mupdext ? ? G rd t G') ->
		  (WFInstrSeq HT G' T I) ->		  
      	       	  (WFInstrSeq HT G T (iseq (ld rd rs i) I))

  |  s_mov	: (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd,rs:Reg; t:Omega)
		  (mlookup ? ? G rs t) ->
		  (mupdext ? ? G rd t G') ->
		  (WFInstrSeq HT G' T I) ->
		  (WFInstrSeq HT G T (iseq (mov rd rs) I))
  |  s_movi	: (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd:Reg; i:int)
		  (mupdext ? ? G rd intty G') ->
		  (WFInstrSeq HT G' T I) ->
		  (WFInstrSeq HT G T (iseq (movi rd i) I))
  |  s_movl     : (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd:Reg; l:label; t,t':Omega)
		  (htlookup (htunwrap HT) l t) ->
		  (Subtype t t') ->
		  (mupdext ? ? G rd t' G') ->
		  (WFInstrSeq HT G' T I) ->
		  (WFInstrSeq HT G T (iseq (movl rd l) I))

  |  s_st       : (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd,rs:Reg; i:int; 
      	       	   V,V':(list initflag); Ts:(list Omega); t:Omega)
		  (mlookup ? ? G rd (tupty Ts V)) ->
		  (mlookup ? ? G rs t) ->
		  (ListNth ? Ts i t) ->
		  (updatetupty V i V') ->
		  (mupdate ? ? G rd (tupty Ts V') G') ->
		  (WFInstrSeq HT G' T I) ->
      	       	  (WFInstrSeq HT G T (iseq (st rd i rs) I))

  |  s_unfold   : (HT:HeapTy; G,G':RegFileTy; T:APTy; I:InstrSeq)
      	       	  (rd,rs:Reg; f:OmegaR;  unf,rst:Omega)
		  (mlookup ? ? G rs rst) ->
		  rst=(recty f) ->
		  (RUnlift (RUnfold f))=unf ->
      	       	  (mupdext ? ? G rd unf G') ->
		  (WFInstrSeq HT G' T I) ->
		  (WFInstrSeq HT G T (iseq (unfold rd rs) I))
.


Inductive WFWordValinitList 
  :  HeapTy -> (list WordVal) -> (list Omega) -> (list initflag) -> Prop
  := wfwvilist0 : (HT:HeapTy)
      	       	  (WFWordValinitList HT (nil ?) (nil ?) (nil ?))
  |  wfwvilist1 : (HT:HeapTy; 
      	       	   wl:(list WordVal); tl:(list Omega); il:(list initflag))
      	       	  (w:WordVal; t:Omega; i:initflag)
      	       	  (WFWordValinit HT w t i) ->
		  (WFWordValinitList HT wl tl il) ->
		  (WFWordValinitList HT (cons w wl) (cons t tl) (cons i il)).


(* well-formed heap values of type *)
Inductive WFHeapVal : HeapTy -> HeapVal -> Omega -> Prop
  := tuple_wf   : (HT:HeapTy; wl:(list WordVal); 
      	       	   tl:(list Omega); il:(list initflag))
      	       	  (WFWordValinitList HT wl tl il) ->
		  (WFHeapVal HT (tuple wl) (tupty tl il))
  | code_wf     : (HT:HeapTy; G:RegFileTy; I:InstrSeq; T:APTy)
      	       	  (WFInstrSeq HT G T I) -> 
      	       	  (WFHeapVal HT (code G T I) (codety G T)).


(* well-formed heap type *)
Inductive WFHeap : Heap -> HeapTy -> Prop
  := heap_wf     : (H:Heap; HT:HeapTy)
      	       	   (EX s | (hsize (hunwrap H) s) /\ 
                           (htsize (htunwrap HT) s) /\
			   ((n:label; h:HeapVal)
			    (hlookup (hunwrap H) n h) -> (lt n s)) /\
			   ((n:label; t:Omega)
			    (htlookup (htunwrap HT) n t) -> (lt n s)) /\
			   ((n:label) (lt n s) ->
			              (EX h | (hlookup (hunwrap H) n h))) /\
			   ((n:label) (lt n s) ->
			              (EX t | (htlookup (htunwrap HT) n t))) /\
			   ((n:label; h:HeapVal; t:Omega)
               	       	       	       (hlookup (hunwrap H) n h) ->
				       (htlookup (htunwrap HT) n t) ->
				       (WFHeapVal HT h t)) /\
		           (OrdHeap (hunwrap H))
      	       	   ) ->
		   (WFHeap H HT).
		   


(* well-formed allocation pointer *)
Inductive WFap : HeapTy -> AP -> APTy -> Prop
  := fresh_wf  : (HT:HeapTy; l:AP)
      	       	 (htsize (htunwrap HT) l) ->
		 (WFap HT l fresh)
  |  used_wf   : (HT:HeapTy; l:AP; n:nat; tl:(list Omega); ol:(list initflag))
      	       	 n=(length tl) ->
      	       	 (htsize (htunwrap HT) (S l)) ->
		 (WFWordVal HT (wl l) (tupty tl ol)) ->
		 (WFap HT l (used n)).


(* strip folds from word value *)
Fixpoint stripWV [w:WordVal] : WordVal
  := Cases w of (wfold w' t) => (stripWV w')
      	      | _ => w
     end.

(* well-formed register file of type *)
Inductive WFRegFile : HeapTy -> RegFile -> RegFileTy -> Prop
  := regfile_wf   : (HT:HeapTy; R:RegFile; G:RegFileTy)
      	       	    ((r:Reg; t:Omega)
		     (mlookup ? ? G r t) ->
		       (EX w | (mlookup ? ? R r w) /\ (WFWordVal HT w t))) ->

		    ((r:Reg; v:WordVal; l:label; n:nat)
		     (mlookup ? ? R r v) ->
      	       	     (stripWV v)=(wl l) ->
		     (htsize (htunwrap HT) n) ->
		     (lt l n)) ->

      	       	    (WFRegFile HT R G).

(* well-formed program *)
Inductive WFProgram : Program -> Prop
  := program_wf   : (H:Heap; HT:HeapTy; R:RegFile; G:RegFileTy; 
      	       	     l:AP; t:APTy; I:InstrSeq)
      	       	    (WFHeap H HT) ->
		    (WFRegFile HT R G) ->
		    (WFap HT l t) ->
		    (WFInstrSeq HT G t I) ->
		    (EX l | (EX G' | (EX T' | (EX I' | (EX n |
      	       	       	 (hlookup (hunwrap H) l (code G' T' I')) /\
			 (ISubDepth I I' n)))))) ->
      	       	    (WFProgram (H, (R, (l, I)))).


(***************************************************************************)
(* Proofs *)

Load lemmas_ftal.


Theorem preserv : (P,P':Program)(WFProgram P) -> (Eval P P') -> (WFProgram P').
Proof.
Intros; Inversion H; Inversion H0.
(* add case: *)
Apply (preserv_add H8 R0 I' r31 rd rs rs' (H8, (R', (r31,I')))); Replace (H8,(R0,(r31,(iseq (add rd rs rs') I')))) with P; Auto; Replace (H8,(R0,(r31,(iseq (add rd rs rs') I')))) with P; Replace  (H8,(R',(r31,I'))) with P'; Auto.
(* addi case: *)
Apply (preserv_addi H8 R0 I' r31 rd rs i (H8, (R', (r31,I')))); Replace (H8,(R0,(r31,(iseq (addi rd rs i) I')))) with P; Auto; Replace (H8,(R0,(r31,(iseq (addi rd rs i) I')))) with P; Replace  (H8,(R',(r31,I'))) with P'; Auto.
(* alloc case: *)
Apply (preserv_alloc H8 R0 I' r31 rd V (H',(R',(r31,I')))).
Replace (H8,(R0,(r31,(iseq (alloc rd V) I')))) with P.
Auto.
Replace (H8,(R0,(r31,(iseq (alloc rd V) I')))) with P.
Replace (H',(R',(r31,I'))) with P'.
Auto.
(* 2 bgt cases: *)
Apply (preserv_bgt H8 R0 I' r31 (H8,(R0,(r31,I'))) rs rt l0).
Replace (H8,(R0,(r31,(iseq (bgt rs rt l0) I')))) with P.
Auto.
Replace (H8,(R0,(r31,(iseq (bgt rs rt l0) I')))) with P.
Replace (H8, (R0,(r31,I'))) with P'.
Auto.
Apply (preserv_bgt H8 R0 I' r31 (H8,(R0,(r31,I''))) rs rt l0).
Replace (H8,(R0,(r31,(iseq (bgt rs rt l0) I')))) with P.
Auto.
Replace (H8,(R0,(r31,(iseq (bgt rs rt l0) I')))) with P.
Replace (H8, (R0,(r31,I''))) with P'.
Auto.
(* bump case; *)
Apply (preserv_bump H8 R0 I' r31 i (H8,(R0,(l0,I')))).
Replace (H8,(R0,(r31,(iseq (bump i) I')))) with P.
Auto.
Replace (H8,(R0,(r31,(iseq (bump i) I')))) with P.
Replace (H8,(R0,(l0,I'))) with P'.
Auto.
(* fold case: *)
Apply (preserv_fold H8 R0 I' r31 (H8,(R',(r31,I'))) rd rs t0).
Replace (H8,(R0,(r31,(iseq (fold rd t0 rs) I')))) with P.
Auto.
Replace (H8,(R0,(r31,(iseq (fold rd t0 rs) I')))) with P.
Replace (H8,(R',(r31,I'))) with P'.
Auto.
(* jd case: *)
Apply (preserv_jd H8 R0 r31 l0 (H8,(R0,(r31,I')))); Replace (H8,(R0,(r31,(jd l0)))) with P; Auto; Replace (H8,(R0,(r31,(jd l0)))) with P; Replace (H8,(R0,(r31,I'))) with P'; Auto.
(* jmp case: *)
Apply (preserv_jmp H8 R0 r31 r (H8,(R0,(r31,I')))); Replace (H8,(R0,(r31,(jmp r)))) with P; Auto; Replace (H8,(R0,(r31,(jmp r)))) with P; Replace (H8,(R0,(r31,I'))) with P'; Auto.
(* ld case: *)
Apply (preserv_ld H8 R0 I' r31 (H8,(R',(r31,I'))) rd rs i).
Rewrite H13.
Auto.
Rewrite H13.
Rewrite H14.
Auto.
(* mov, movi cases: *)
Apply (preserv_mov H8 R0 r31 I' (H8,(R',(r31,I'))) rd rs).
Rewrite H11.
Auto.
Rewrite H11.
Rewrite H12.
Auto.
Apply (preserv_movi H8 R0 r31 I' (H8,(R',(r31,I'))) rd i).
Rewrite H10.
Auto.
Rewrite H10.
Rewrite H11.
Auto.
(* movl case: *)
Apply (preserv_movl H8 R0 r31 I' (H8,(R',(r31,I'))) rd l0); Replace (H8,(R0,(r31,(iseq (movl rd l0) I')))) with P; Auto; Replace (H8,(R0,(r31,(iseq (movl rd l0) I')))) with P; Replace (H8,(R',(r31,I'))) with P'; Auto.
(* st case: *)
Apply (preserv_st H8 R0 I' r31 rd rs i (H',(R0,(r31,I')))); Replace (H8,(R0,(r31,(iseq (st rd i rs) I')))) with P; Auto; Replace (H8,(R0,(r31,(iseq (st rd i rs) I')))) with P; Replace (H',(R0,(r31,I'))) with P'; Auto.
(* unfold case: *)
Apply (preserv_unfold H8 R0 I' r31 (H8,(R',(r31,I'))) rd rs).
Replace (H8,(R0,(r31,(iseq (unfold rd rs) I')))) with P.
Auto.
Replace (H8,(R0,(r31,(iseq (unfold rd rs) I')))) with P.
Replace (H8,(R',(r31,I'))) with P'.
Auto.
Save.




Theorem progress : (P:Program) (WFProgram P) -> (EX P' | (Eval P P')).
Proof.
Intros.
Inversion H.
Clear H5.
Inversion H4.
Apply progress_add.
Rewrite H13.
Rewrite H6.
Auto.
Apply progress_addi.
Rewrite H12.
Rewrite H6.
Auto.
Apply progress_alloc.
Rewrite H12.
Rewrite H6.
Auto.
Apply progress_bgt.
Rewrite H14.
Rewrite H6.
Auto.
Apply progress_bump.
Rewrite H10.
Rewrite H6.
Auto.
Apply progress_fold.
Rewrite H13.
Rewrite H6.
Auto.
Apply progress_jd.
Rewrite H11.
Rewrite H6.
Auto.
Apply progress_jmp.
Rewrite H11.
Rewrite H6.
Auto.
Apply progress_ld.
Rewrite H14.
Rewrite H6.
Auto.
Apply progress_mov.
Rewrite H12. Rewrite H6. Auto.
Apply progress_movi.
Rewrite H11. Rewrite H6. Auto.
Apply progress_movl.
Rewrite H13.
Rewrite H6.
Auto.
Apply progress_st.
Rewrite H15.
Rewrite H6.
Auto.
Apply progress_unfold.
Rewrite H14.
Rewrite H6.
Auto.
Save.




(***************************************************************************)


