(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)
(* Thanks to Jean Goubault-Larrecq for some Coq help  *)

(* loaded by tal.v *)


Require Le.
Require Lt.
Require natutil.



(***************************************************************************)
(* Lemmas for tal.v *)
(***************************************************************************)


(* utilities *)
Definition left_and : (A,B:Prop) (A/\B) -> A
  := [A,B:Prop; H:(A/\B)](and_ind A B A [H0:A; _:B]H0 H).
Definition right_and : (A,B:Prop) (A/\B) -> B
  := [A,B:Prop; H:(A/\B)](and_ind A B B [_:A; H1:B]H1 H).


(* equality of subterms of equal terms of inductive definitions *)
Lemma eq_pair_compl : (A,B:Set; a,c:A; b,d:B) (a,b)=(c,d) -> a=c.
Intros; Inversion H; Auto.
Save.
Lemma eq_pair_compr : (A,B:Set; a,c:A; b,d:B) (a,b)=(c,d) -> b=d.
Intros; Inversion H; Auto.
Save.
Lemma eq_progs_H : (H1,H2:Heap; R1,R2:RegFile; a1,a2:AP; 
                    I1,I2:InstrSeq; P:Program)
                   (H1,(R1,(a1,I1)))=P -> (H2,(R2,(a2,I2)))=P -> H1=H2.
Intros; Induction H; Inversion H0; Auto.
Save.
Lemma eq_progs_R : (H1,H2:Heap; R1,R2:RegFile; a1,a2:AP; 
                    I1,I2:InstrSeq; P:Program)
                   (H1,(R1,(a1,I1)))=P -> (H2,(R2,(a2,I2)))=P -> R1=R2.
Intros; Induction H; Inversion H0; Auto.
Save.
Lemma eq_progs_I : (H1,H2:Heap; R1,R2:RegFile; a1,a2:AP; 
                    I1,I2:InstrSeq; P:Program)
                   (H1,(R1,(a1,I1)))=P -> (H2,(R2,(a2,I2)))=P -> I1=I2.
Intros; Induction H; Inversion H0; Auto.
Save.
Lemma eq_progs_a : (H1,H2:Heap; R1,R2:RegFile; a1,a2:AP; 
                    I1,I2:InstrSeq; P:Program)
                   (H1,(R1,(a1,I1)))=P -> (H2,(R2,(a2,I2)))=P -> a1=a2.
Intros; Induction H; Inversion H0; Auto.
Save.
Lemma eq_instrseq : (i1,i2:Instr; I1,I2:InstrSeq)
      	       	    (iseq i1 I1)=(iseq i2 I2) -> (i1=i2).
Intros; Inversion H; Auto.
Save.
Lemma eq_add_rd : (rd,rd1, rs,rs1, rs',rs1':Reg) (add rd rs rs')=(add rd1 rs1 rs1') -> (rd=rd1).
Intros; Inversion H; Auto.
Save.
Lemma eq_codety_G : (G,G':RegFileTy; T,T':APTy) (codety G T)=(codety G' T') -> G=G'.
Intros; Inversion H; Auto.
Save.

Lemma eq_recty : (f,f':OmegaR) (recty f)=(recty f') -> f=f'.
Intros; Inversion H; Auto.
Save.



(* used to argue contradictions in cases *)
(*
Definition liftb := [b:bool] if b then True else False.
Definition is_intty
  := [o:Omega] (Cases o of (intty) => true | _ => false end).
Definition is_codety 
  := [o:Omega] (Cases o of (codety _ _) => true | _ => false end).
Definition is_tupty
  := [o:Omega] (Cases o of (tupty _ _) => true | _ => false end).
Definition is_recty
  := [o:Omega] (Cases o of (recty _ _) => true | _ => false end).
*)
(* usually use them like this:
Absurd (tupty n V)=intty.                 <-- the absurd hypothesis
Unfold not.
Intros.
Change (liftb (is_intty (tupty n V))).    <-- fix appropriate here
Rewrite H25.                              <-- whatever says they *are* equal
Unfold is_intty.
Red.
Auto.
Auto.                                     <- this uses the H25 
*)




(* more major lemmas *)
Lemma listnth_lt_length
      	       	     : (A:Set; i:nat; l:(list A); a:A)
		       (ListNth A l i a) ->
		       (lt i (length l)).
Proof.
Induction i.
Intros.
Inversion H.
Unfold length.
Apply lt_O_Sn.

Intros.
Inversion H0.
Unfold length.
Apply lt_n_S.
Apply (H V a).
Auto.
Save.


Lemma lt_imp_listnth
  : (A:Set; i:nat; l:(list A))
    (lt i (length l)) ->
    (EX a | (ListNth A l i a)).
Proof.
Double Induction 2 3.
Intros.
Inversion H.
Intros.
Split with a.
Constructor.
Intros.
Inversion H0.
Intros.
Cut (EX a:? | (ListNth A l0 n a)).
Intros.
Inversion_clear H2.
Split with x.
Constructor.
Auto.
Apply H0.
Unfold length in H1.
Unfold length.
Apply lt_S_n.
Auto.
Save.



Lemma lt_imp_updatetuple
      	       : (i:int; wl:(list WordVal); w:WordVal)
		 (lt i (length wl)) ->
		 (EX wl' | (updatetuple wl i w wl')).
Proof.
Double Induction 1 2.
Intros.
Inversion H.

Intros.
Split with (cons w l).
Constructor.

Intros.
Inversion H0.

Intros.
Cut (EX wl'':? | (updatetuple l n w wl'')).
Intros.
Inversion_clear H2.
Split with (cons a x).
Constructor.
Auto.

Apply H0.
Unfold length in H1.
Unfold length.
Apply lt_S_n.
Auto.
Save.


Lemma lt_imp_updatetupty
      	       : (i:int; il:(list initflag))
		 (lt i (length il)) ->
		 (EX il' | (updatetupty il i il')).
Proof.
Double Induction 1 2.
Intros.
Inversion H.

Intros.
Split with (cons true l).
Constructor.

Intros.
Inversion H0.

Intros.
Cut (EX il'':? | (updatetupty l n il'')).
Intros.
Inversion_clear H2.
Split with (cons a x).
Constructor.
Auto.
Apply H0.
Unfold length.
Unfold length in H1.
Apply lt_S_n.
Auto.
Save.





Lemma wfwordvalinitlist_length_eq_wl_tl
      	       	    : (HT:HeapTy; wl:(list WordVal); 
      	       	       tl:(list Omega); il:(list initflag))
		      (WFWordValinitList HT wl tl il) ->
		      (length wl)=(length tl).
Proof.
Induction wl.
Intros.
Inversion H.
Unfold length; Auto.

Intros; Inversion H0.
Cut (length l)=(length tl0).
Unfold length.
Intros.
Apply eq_S.
Auto.

Apply H with il0.
Auto.
Save.


Lemma wfwordvalinitlist_length_eq_il_tl
      	       	    : (HT:HeapTy; wl:(list WordVal); 
      	       	       tl:(list Omega); il:(list initflag))
		      (WFWordValinitList HT wl tl il) ->
		      (length il)=(length tl).
Proof.
Induction wl.
Intros.
Inversion H.
Unfold length.
Auto.

Intros.
Inversion H0.
Cut (length il0)=(length tl0).
Unfold length.
Intros.
Apply eq_S.
Auto.

Apply H.
Auto.
Save.


Lemma wftupty_length : (HT:HeapTy; wl:(list WordVal); tl:(list Omega);
      	       	       	il:(list initflag))
		       (WFWordValinitList HT wl tl il) ->
		       (length tl)=(length il).
Proof.
Induction wl.
Intros.
Inversion H.
Unfold length.
Auto.
Intros.
Inversion H0.
Cut (length tl0)=(length il0).
Intros.
Unfold length in H8.
Unfold length.
Apply eq_S.
Auto.
Apply H.
Auto.
Save.


Lemma subtype_recty : (t,t':Omega; f:OmegaR)
      	       	       (Subtype t t') -> 
      	       	       t'=(recty f) ->
		       t=(recty f).
Proof.
Intros.
Induction H.
Auto.
Absurd (tupty ol il2)=(recty f).
Discriminate; Auto.
Auto.
Save.



Lemma subtype_codety : (t,t':Omega; R:RegFileTy; A:APTy) 
      	       	       (Subtype t t') -> 
      	       	       t'=(codety R A) ->
		       t=(codety R A).
Proof.
Intros.
Induction H.
Auto.
Absurd (tupty ol il2)=(codety R A).
Discriminate; Auto.
Auto.
Save.


Lemma subtype_intty : (t,t':Omega) (Subtype t t') -> t'=intty -> t=intty.
Intros.
Induction H.
Auto.
Absurd (tupty ol il2)=intty.
Discriminate; Auto.
Auto.
Save.


Lemma subtype_tupty : (t:Omega; tl:(list Omega); il:(list initflag))
      	       	      (Subtype t (tupty tl il)) ->
		      (EX il' | t=(tupty tl il')).
Proof.
Intros.
Inversion H.
Split with il.
Auto.
Split with il1.
Auto.
Save.





Lemma regfile_eq_subtype_eq
      	     : (G,G1,G2:RegFileTy)
	       ( (r:Reg; t:Omega) (mlookup ? ? G1 r t) ->
	                          (mlookup ? ? G2 r t) ) ->
	       (RegFileSubtype G1 G) ->
	       (RegFileSubtype G2 G).
Proof.
Intros.
Inversion H0.
Constructor.
Intros.
Apply H.
Apply H1.
Auto.
Save.


Lemma regfile_ext_eq_instr_wf
      	     : (I:InstrSeq; HT:HeapTy; G1,G2:RegFileTy; T:APTy)
	       ( (r:Reg; t:Omega) (mlookup ? ? G1 r t) ->
	                          (mlookup ? ? G2 r t) ) ->
	       (WFInstrSeq HT G1 T I) ->
	       (WFInstrSeq HT G2 T I).
Proof.
Induction I.
Induction i.
Intros.
Inversion H1.
Cut (EX G':? | (mupdext ? ? G2 r intty G')).
Intros.
Inversion_clear H13.
Apply s_add with x.
Apply H0.
Auto.
Apply H0.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r intty.
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.

Intros.
Inversion H1.
Cut (EX G2':? | (mupdext ? ? G2 r intty G2')).
Intros.
Elim H12; Clear H12; Intros G2' H12.
Apply s_addi with G2'.
Apply H0.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r intty.
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.

Intros.
Inversion H1.
Cut (EX G2':? | (mupdext ? ? G2 r (tupty l (makeUninitTupty l)) G2')).
Intros.
Elim H11; Clear H11; Intros G2' H11.
Apply s_alloc with G2' n.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq
 with G1 G2 G' r (tupty l (makeUninitTupty l)).
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.

Intros.
Inversion H1.
Apply s_bgt with G'.
Apply H0.
Auto.
Apply H0; Auto.
Auto.
Apply regfile_eq_subtype_eq with G1.
Auto.
Auto.
Apply H with G1.
Auto.
Auto.
Intros.
Inversion H1.
Constructor.
Apply H with G1.
Auto.
Auto.
Intros.
Inversion H1.
Cut (EX G2':? | (mupdext ? ? G2 r (recty f) G2')).
Intros.
Elim H13; Clear H13; Intros G2' H13.
Apply s_fold with G2' rst.
Apply H0; Auto.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r (recty f).
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.

Intros.
Inversion H1.
Cut (EX G2' | (mupdext ? ? G2 r t G2')).
Intros.
Elim H14; Clear H14; Intros G2' H14.
Apply s_ld with G2' ol il t.
Apply H0.
Auto.
Auto.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r t.
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.

Intros.
Inversion H1.
Cut (EX G2':? | (mupdext ? ? G2 r t G2')).
Intros.
Elim H11; Clear H11; Intros G2' H11.
Apply s_mov with G2' t.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r t.
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always; Apply eqorneq_reg.

Intros.
Inversion H1.
Cut (EX G2':? | (mupdext ? ? G2 r intty G2')).
Intros.
Elim H10; Clear H10; Intros G2' H10.
Apply s_movi with G2'.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r intty.
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always; Apply eqorneq_reg.

Intros.
Inversion H1.
Cut (EX G2':? | (mupdext ? ? G2 r t' G2')).
Intros.
Elim H12; Clear H12; Intros G2' H12.
Apply s_movl with G2' t t'.
Auto.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r t'.
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always; Apply eqorneq_reg.
Intros.
Inversion H1.
Cut (EX G2':? | (mupdate ? ? G2 r (tupty Ts V') G2')).
Intros.
Elim H15; Clear H15; Intros G2' H15.
Apply s_st with G2' V V' Ts t.
Apply H0.
Auto.
Apply H0; Auto.
Auto.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r (tupty Ts V').
Apply eqorneq_reg.
Auto.
Apply mupdate_imp_mupdext.
Apply eqorneq_reg.
Auto.
Apply mupdate_imp_mupdext.
Apply eqorneq_reg.
Auto.
Auto.
Auto.
Apply mlookup_imp_mupdate with (tupty Ts V).
Apply H0.
Auto.
Intros.
Inversion H1.
Cut (EX G2':? | (mupdext ? ? G2 r unf G2')).
Intros. 
Elim H13; Clear H13; Intros G2' H13.
Apply s_unfold with G2' f unf rst.
Apply H0; Auto.
Auto.
Auto.
Auto.
Apply H with G'.
Intros.
Apply ext_map_eq_updext_eq with G1 G2 G' r unf.
Apply eqorneq_reg; Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Intros.
Inversion H0.
Apply s_jd with G'.
Auto.
Apply regfile_eq_subtype_eq with G1.
Auto.
Auto.
Intros.
Inversion H0.
Apply s_jmp with G'.
Apply H; Auto.
Apply regfile_eq_subtype_eq with G1.
Auto.
Auto.
Save.




Lemma wfwordval_label_lt_heap_size
  : (H:Heap; HT:HeapTy; v:WordVal; l:label; t:Omega; n:nat)
    (WFHeap H HT) ->
    (WFWordVal HT v t) ->
    (stripWV v)=(wl l) ->
    (htsize (htunwrap HT) n) ->
    (lt l n).
Proof.
Induction v.
Intros.
Inversion H0.
Inversion_clear H5.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Clear H12.
Inversion H1.
Unfold stripWV in H2.
Inversion H2.
Rewrite <- H17.
Replace n0 with x.
Apply H10 with t'.
Auto.
Apply (msize_eq ? ? (htunwrap HT)).
Auto.
Auto.
Intros.
Unfold stripWV in H2.
Inversion H2.
Intros.
Unfold stripWV in H2.
Inversion H2.
Intros.
Unfold 1 stripWV in H3.
Fold stripWV in H3.
Inversion H2.
Apply H0 with t'.
Auto.
Auto.
Auto.
Auto.
Save.




Lemma regfile_upd_wf 
      	     : (H:Heap; HT:HeapTy; R,R':RegFile; G,G':(Map Reg Omega))
	       (rd:Reg; w:WordVal; t:Omega)
	       (WFHeap H HT) ->
	       (WFRegFile HT R G) ->
	       (WFWordVal HT w t) ->
	       (mupdext ? ? R rd w R') ->
	       (mupdext ? ? G rd t G') ->
	       (WFRegFile HT R' G').
Proof.
Intros.
Inversion H1.
Constructor.
Intros.
Cut r=rd\/~r=rd.
Intros.
Inversion_clear H11.
Split with w.
Split.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Rewrite H12.
Auto.
Replace t0 with t.
Auto.
Cut (mlookup ? ? G' r t).
Intros.
Apply (mlookup_eq Reg Omega r t t0 G').
Apply eqorneq_reg.
Auto.
Auto.
Apply mupdext_imp_mlookup with G.
Apply eqorneq_reg.
Rewrite H12.
Auto.
Cut (EX w:WordVal | (mlookup Reg WordVal R r w)/\(WFWordVal HT w t0)).
Intros.
Inversion_clear H11.
Inversion_clear H13.
Split with x.
Split.
Apply (mupdext_preserv Reg WordVal R R' r rd x w).
Auto.
Auto.
Auto.
Auto.
Cut (mlookup Reg Omega G r t0).
Intros.
Apply H5.
Auto.
Apply (mupdext_preserv_rev Reg Omega G G' r rd t0 t).
Auto.
Auto.
Auto.
Apply eqorneq_reg.
Intros.
Cut r=rd\/~r=rd.

Intros.
Inversion_clear H13.
Apply wfwordval_label_lt_heap_size with H HT v t.
Auto.
Replace v with w.
Auto.
Apply (mlookup_eq ? ? r w v R').
Apply eqorneq_reg.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Rewrite H14.
Auto.
Auto.
Auto.
Auto.
Apply H6 with r v.
Apply mupdext_preserv_rev with R' rd w.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply eqorneq_reg.
Save.


Lemma heap_lookup_not_recty
      	    : (H:Heap; HT:HeapTy; l:label; t:Omega; f:OmegaR)
	      (WFHeap H HT) ->
	      (htlookup (htunwrap HT) l t) ->
	      ~t=(recty f).
Proof.
Intros.
Induction H.
Induction HT.
Unfold unwrapMap in H1.
Inversion H0.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Unfold unwrapMap in H11.
Cut (lt l x).
Intros.
Unfold unwrapMap in H7.
Unfold unwrapMap in H8.
Unfold unwrapMap in H9.
Unfold unwrapMap in H2.
Unfold unwrapMap in H5.
Unfold unwrapMap in H6.
Cut (EX h:? | (mlookup nat HeapVal m l h)).
Intros.
Inversion_clear H12.
Cut (WFHeapVal (wfmap nat Omega m1 m2) x0 t).
Intros.
Inversion H12.
Discriminate.
Discriminate.
Inversion H11.
Apply (H12 l x0 t).
Auto.
Auto.
Apply H8.
Auto.
Unfold unwrapMap in H7.
Apply H7 with t.
Auto.
Save.


Lemma heap_lookup_not_intty
      	    : (H:Heap; HT:HeapTy; l:label; t:Omega)
	      (WFHeap H HT) ->
	      (htlookup (htunwrap HT) l t) ->
	      ~t=intty.
Proof.
Intros.
Induction H.
Induction HT.
Unfold unwrapMap in H1.
Inversion H0.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Unfold unwrapMap in H11.
Cut (lt l x).
Intros.
Unfold unwrapMap in H7.
Unfold unwrapMap in H8.
Unfold unwrapMap in H9.
Unfold unwrapMap in H2.
Unfold unwrapMap in H5.
Unfold unwrapMap in H6.
Move H1 after H10.
Cut (EX h:? | (mlookup nat HeapVal m l h)).
Intros.
Inversion_clear H12.
Cut (WFHeapVal (wfmap nat Omega m1 m2) x0 t).
Intros.
Inversion H12.
Discriminate.
Discriminate.
Inversion H11.
Apply (H12 l x0 t).
Auto.
Auto.
Apply H8.
Auto.
Unfold unwrapMap in H7.
Apply H7 with t.
Auto.
Save.





Lemma can_word_forms_rec
      	    : (H:Heap; HT:HeapTy; w:WordVal; f:OmegaR)
	      (WFHeap H HT) ->
	      (WFWordVal HT w (recty f)) ->
	      (EX v | (EX t | w=(wfold v t))).
Proof.
Intros.
Inversion H1.
Absurd intty=(recty f).
Discriminate.
Auto.
Cut t'=(recty f).
Intros.
Cut ~t'=(recty f).
Intros.
Rewrite H7 in H8.
Absurd ~(recty f)=(recty f).
Auto.
Auto.
Apply heap_lookup_not_recty with H HT l.
Auto.
Auto.
Apply subtype_recty with (recty f).
Auto.
Auto.
Split with w0.
Split with (recty f).
Auto.
Save.


Lemma can_word_forms_int
      	    : (H:Heap; HT:HeapTy; w:WordVal; t:Omega)
	      (WFHeap H HT) ->
	      (WFWordVal HT w t) ->
	      (t=intty) ->
	      (EX i | w=(wi i)).
Proof.
Intros.
Rewrite H2 in H1.
Inversion H1.
Split with i; Auto.
Cut t'=intty.
Intros.
Cut ~t'=intty.
Intros.
Rewrite H8 in H9.
Absurd ~intty=intty.
Auto.
Auto.
Apply (heap_lookup_not_intty H HT l t').
Auto.
Auto.
Apply subtype_intty with intty.
Auto.
Auto.
Save.




Lemma can_reg_forms_any : (HT:HeapTy; R:RegFile; G:(Map Reg Omega);
                       r:Reg; t:Omega) 
                       (WFRegFile HT R G) -> 
		       (mlookup Reg Omega G r t) ->
		       (EX w | (mlookup Reg WordVal R r w)).
Proof.
Intros.
Inversion H.
Cut (EX w:WordVal | (mlookup Reg WordVal R r w)/\(WFWordVal HT w t)).
Intros.
Elim H6.
Intros.
Split with x.
Inversion H7.
Auto.
Apply (H1 r t H0).
Save.


Lemma can_reg_forms_tuple
    : (HT:HeapTy; R:RegFile; G:RegFileTy; 
       r:Reg; tl:(list Omega); il:(list initflag))
      (WFRegFile HT R G) ->
      (mlookup Reg Omega G r (tupty tl il)) ->
      (EX l | (mlookup ? ? R r (wl l))).
Proof.
Intros.
Inversion H.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w (tupty tl il))).
Intros.
Inversion_clear H6.
Inversion_clear H7.
Inversion H8.
Split with l.
Rewrite <- H11 in H6.
Auto.
Apply H1.
Auto.
Save.


Lemma tupty_lists_length_eq
  : (H:Heap; HT:HeapTy; R:RegFile; G:RegFileTy; 
       r:Reg; l:label; tl:(list Omega); il:(list initflag); V:(list WordVal))
    (WFHeap H HT) ->
    (WFRegFile HT R G) ->
    (mlookup Reg Omega G r (tupty tl il)) ->
    (mlookup ? ? R r (wl l)) ->
    (hlookup (hunwrap H) l (tuple V)) ->
    (length V)=(length tl).
Proof.
   Intros.
Cut (WFWordVal HT (wl l) (tupty tl il)).
Intros.
Inversion H5.
Cut (EX il':? | t'=(tupty tl il')).
Intros.
Elim H11; Clear H11; Intros il'; Intros.
Rewrite H11 in H7.
Move H4 after H7.
Cut (WFHeapVal HT (tuple V) (tupty tl il')).
Intros.
Inversion H12.
Apply (wfwordvalinitlist_length_eq_wl_tl HT V tl il').
Auto.

Inversion H0.
Inversion_clear H13.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H21.
Inversion_clear H22.
Apply H21 with l.
Auto.
Auto.

Apply subtype_tupty with il.
Auto.

Inversion H1.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w (tupty tl il))).
Intros.
Inversion_clear H10.
Inversion_clear H11.
Replace (wl l) with x.
Auto.

Apply (mlookup_eq ? ? r x (wl l) R).
Apply eqorneq_reg.
Auto.
Auto.

Apply H5.
Auto.
Save.


Lemma can_heap_forms_reg_tuple
    : (H:Heap; HT:HeapTy; R:RegFile; G:RegFileTy; 
       r:Reg; l:label; tl:(list Omega); il:(list initflag))
      (WFHeap H HT) ->
      (WFRegFile HT R G) ->
      (mlookup ? ? G r (tupty tl il)) ->
      (mlookup ? ? R r (wl l)) ->
      (EX V | (hlookup (hunwrap H) l (tuple V))).
Proof.
Intros.
Induction H.
Induction HT.
Unfold unwrapMap.
Inversion H0.
Inversion_clear H4.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Unfold unwrapMap in H4; Unfold unwrapMap in H7; Unfold unwrapMap 
in H8; Unfold unwrapMap in H9; Unfold unwrapMap in H10;
Unfold unwrapMap in H11; Unfold unwrapMap in H13.
Cut (EX h:? | (mlookup ? ? m l h)).
Cut (WFWordVal (wfmap ? ? m1 m2) (wl l) (tupty tl il)).
Intros.
Inversion H12.
Cut (EX il':? | t'=(tupty tl il')).
Intros.
Elim H20; Clear H20; Intros il'.
Intros.
Unfold unwrapMap in H16.
Rewrite H20 in H16.
Elim H14; Clear H14; Intros h; Intros.
Cut (WFHeapVal (wfmap ? ? m1 m2) h (tupty tl il')).
Intros.
Inversion H21.
Split with wl0.
Rewrite <- H24 in H14.
Auto.

Inversion H13.
Apply H21 with l.
Auto.
Auto.

Apply subtype_tupty with il.
Auto.

Inversion H1.
Cut (EX w:? |
       (mlookup ? ? R r w)
       /\(WFWordVal (wfmap ? ? m1 m2) w (tupty tl il))).
Intros.
Inversion_clear H18.
Inversion_clear H19.
Replace (wl l) with x0.
Auto.

Apply (mlookup_eq ? ? r x0 (wl l) R).
Apply eqorneq_reg.
Auto.
Auto.

Apply H12.
Auto.

Apply H10.
Inversion H1.
Cut (WFWordVal (wfmap ? ? m1 m2) (wl l) (tupty tl il)).
Intros.
Inversion H18.
Unfold unwrapMap in H20.
Apply (H9 l t').
Auto.

Cut (EX w:? |
       (mlookup ? ? R r w)
       /\(WFWordVal (wfmap ? ? m1 m2) w (tupty tl il))).
Intros.
Inversion_clear H18.
Inversion_clear H19.
Replace (wl l) with x0.
Auto.
Apply (mlookup_eq ? ? r x0 (wl l) R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H12.
Auto.
Save.



Lemma can_heap_forms_code
    : (HT:HeapTy; H:Heap; l:label; G:(Map Reg Omega); T:APTy)
      (WFHeap H HT) ->
      (mlookup nat Omega (unwrapMap nat Omega HT) l (codety G T)) ->
      (EX I | (mlookup ? ? (unwrapMap ? ? H) l (code G T I))).
Proof.
Intros.
Inversion H0.
Inversion_clear H3.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Cut (EX h:HeapVal | (mlookup nat HeapVal (unwrapMap nat HeapVal H) l h)).
Intros.
Inversion_clear H11.
Cut (WFHeapVal HT x0 (codety G T)).
Intros.
Inversion H11.
Split with I.
Rewrite <- H14.
Rewrite H18 in H16.
Rewrite H16.
Auto.
Inversion H12.
Apply (H11 l x0 (codety G T) H13 H1).
Apply (H9 l (H8 l (codety G T) H1)).
Save.



Lemma can_heap_forms_code_I
      	    : (HT:HeapTy; h:HeapVal; t:Omega; G:RegFileTy; I:InstrSeq; T:APTy)
	      (WFHeapVal HT h t) ->
	      t=(codety G T) ->
	      h=(code G T I) ->
	      (WFInstrSeq HT G T I).
Intros.
Rewrite H0 in H.
Rewrite H1 in H.
Inversion H.
Auto.
Save.



Lemma can_word_forms_code
      	    : (H:Heap; HT:HeapTy; w:WordVal; G:RegFileTy; T:APTy)
	      (WFHeap H HT) ->
	      (WFWordVal HT w (codety G T)) ->
	      (EX l | w=(wl l) /\
      	       	  (EX I | (mlookup ? ? (unwrapMap ? ? H) l (code G T I)))).
Proof.
Intros.
Inversion H1.
Split with l.
Split.
Auto.
Apply can_heap_forms_code with HT.
Auto.
Replace (codety G T) with t'.
Auto.
Apply subtype_codety with (codety G T).
Auto.
Auto.
Save.


Lemma wfwordval_in_reg
    : (HT:HeapTy; R:RegFile; G:RegFileTy)
      (r:Reg; w:WordVal; t:Omega)
      (WFRegFile HT R G) ->
      (mlookup ? ? G r t) ->
      (mlookup ? ? R r w) ->
      (WFWordVal HT w t).
Proof.
Intros.
Inversion H.
Cut (EX w:WordVal | (mlookup Reg WordVal R r w)/\(WFWordVal HT w t)).
Intros.
Inversion_clear H7.
Inversion_clear H8.
Replace w with x.
Auto.
Apply (mlookup_eq Reg WordVal r x w R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H2.
Auto.
Save.



Lemma wfheapval_in_heap 
    : (H:Heap; HT:HeapTy; l:label; h:HeapVal; t:Omega)
      (WFHeap H HT) ->
      (hlookup (hunwrap H) l h) ->
      (htlookup (htunwrap HT) l t) ->
      (WFHeapVal HT h t).
Proof.
Intros.
Inversion H0.
Inversion_clear H4.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H13.
Apply (H12 l h t H1 H2).
Save.


Lemma wfheap_code_eq_aux
    : (H:Heap; HT:HeapTy; l:label; G,G':RegFileTy; T,T':APTy; I:InstrSeq)
      (WFHeap H HT) ->
      (hlookup (hunwrap H) l (code G T I)) ->
      (htlookup (htunwrap HT) l (codety G' T')) ->
      (WFHeapVal HT (code G T I) (codety G' T')) ->
      (G=G') /\ (T=T').
Intros.
Inversion H3.
Split.
Auto.
Auto.
Save.


Lemma wfheap_code_eq
    : (H:Heap; HT:HeapTy; l:label; G,G':RegFileTy; T,T':APTy; I:InstrSeq)
      (WFHeap H HT) ->
      (hlookup (hunwrap H) l (code G T I)) ->
      (htlookup (htunwrap HT) l (codety G' T')) ->
      (G=G') /\ (T=T').
Intros.
Apply (wfheap_code_eq_aux H HT l G G' T T' I H0 H1 H2
        (wfheapval_in_heap H HT l (code G T I) (codety G' T') 
      	        H0 H1 H2)).
Save.


(* lemmas for register subtyping *)
Lemma regfilesubtype_id : (G:RegFileTy) (RegFileSubtype G G).
Intros.
Constructor.
Intros.
Auto.
Save.

Lemma regsub_mlookup
    : (G,G':RegFileTy; r:Reg; t:Omega)
      (RegFileSubtype G G') ->
      (mlookup ? ? G' r t) ->
      (mlookup ? ? G r t).
Intros.
Inversion H.
Apply (H1 r t H0).
Save.

Lemma regfile_weakening
    : (HT:HeapTy; R:RegFile; G,G':RegFileTy)
      (RegFileSubtype G G') ->
      (WFRegFile HT R G) ->
      (WFRegFile HT R G').
Intros.
Inversion H.
Inversion H0.
Apply regfile_wf.
Auto.
Auto.
Save.


(* some lemmas about our encoding *)
Lemma wfheaplookup_label_lt
    : (H:Heap; HT:HeapTy; l:label; h:HeapVal; s:nat)
      (WFHeap H HT) ->
      (mlookup nat HeapVal (unwrapMap nat HeapVal H) l h) ->
      (msize nat HeapVal (unwrapMap nat HeapVal H) s) ->
      (lt l s).
Proof.
Intros.
Inversion H0.
Inversion_clear H4.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Clear H10.
Replace s with x.
Apply H8 with h.
Auto.
Apply msize_eq with m:=(unwrapMap nat HeapVal H).
Auto.
Auto.
Save.


Lemma size_notin_wfheap   : (H:Heap; HT:HeapTy; 
      	       	       	     h:(Map label HeapVal); wfh:(mWF ? ? h);
      	       	       	     l:label)
			    H=(wfmap label HeapVal h wfh) ->
		            (WFHeap H HT) ->
		            (msize ? ? h l) ->
			    (mnotindom ? ? h l).
Proof.
Intros.
Inversion H1.
Inversion_clear H4.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Cut x=l.
Intros.
Rewrite <- H12.
Cut (mindom ? ? h x)\/(mnotindom ? ? h x).
Intros.
Inversion_clear H14.
Absurd (mindom ? ? h x).
Unfold not.
Intros.
Cut (EX t:? | (mlookup ? ? (unwrapMap ? ? H) x t)).
Intros.
Inversion_clear H16.
Cut (lt x x).
Unfold lt.
Apply le_Sn_n.
Apply (H8 x x0 H17).
Rewrite H0.
Unfold unwrapMap.
Apply mindom_imp_mlookup.
Auto.
Auto.
Auto.
Apply min_or_not_nat_wf.
Auto.
Rewrite H0 in H4.
Unfold unwrapMap in H4.
Move H2 after H4.
Apply (msize_eq ? ? h x l).
Auto.
Auto.
Save.


Lemma size_notin_wfheapty : (H:Heap; HT:HeapTy; 
      	       	       	     ht:(Map label Omega); wfht:(mWF ? ? ht);
      	       	       	     l:label)
			    HT=(wfmap label Omega ht wfht) ->
		            (WFHeap H HT) ->
		            (msize ? ? ht l) ->
			    (mnotindom ? ? ht l).
Proof.
Intros.
Inversion H1.
Inversion_clear H4.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Cut x=l.
Intros.
Rewrite <- H12.
Cut (mindom nat Omega ht x)\/(mnotindom nat Omega ht x).
Intros.
Inversion_clear H14.
Absurd (mindom nat Omega ht x).
Unfold not.
Intros.
Cut (EX t:? | (mlookup nat Omega (unwrapMap nat Omega HT) x t)).
Intros.
Inversion_clear H16.
Cut (lt x x).
Unfold lt.
Apply le_Sn_n.
Apply (H9 x x0 H17).
Rewrite H0.
Unfold unwrapMap.
Apply mindom_imp_mlookup.
Auto.
Auto.
Auto.
Apply min_or_not_nat_wf.
Auto.
Apply (msize_eq nat Omega) with m:=ht.
Rewrite H0 in H7.
Unfold unwrapMap in H7.
Auto.
Auto.
Save.


Lemma wfh_ht_size_eq
  : (h:(Map label HeapVal); wfh:(mWF ? ? h);
     ht:(Map label Omega); wfht:(mWF ? ? ht); l:nat)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (htsize ht l).    
Proof.
Intros.
Inversion H.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H6.
Replace l with x.
Unfold unwrapMap in H5.
Auto.
Apply (msize_eq nat HeapVal h).
Auto.
Auto.
Save.


Lemma wfh_h_size_eq
  : (h:(Map label HeapVal); wfh:(mWF ? ? h);
     ht:(Map label Omega); wfht:(mWF ? ? ht); l:nat)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htsize ht l) ->
    (hsize h l).
Proof.
Intros.
Inversion H.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H6.
Replace l with x.
Unfold unwrapMap in H2.
Auto.
Unfold unwrapMap in H5.
Apply (msize_eq nat Omega ht).
Auto.
Auto.
Save.


(* this does not hold with current encoding  ------------
Lemma WFHeap_next : (l:label; h:HeapVal; t:Omega; 
      	       	     H:(Map label HeapVal);
      	       	     HT:(Map label Omega);
		     mwfh:(mWF ? ? (melem ? ? l h H));
		     mwfht:(mWF ? ? (melem ? ? l t HT)))
		    (WFHeap (wfmap ? ? (melem ? ? l h H) mwfh)
		            (wfmap ? ? (melem ? ? l t HT) mwfht)) ->
		    (WFHeap (wfmap ? ? H (mwf_next ? ? l h H mwfh))
      	       	       	    (wfmap ? ? HT (mwf_next ? ? l t HT mwfht))).
*)


Lemma makeUninitTupty_len_eq 
  : (n:nat; V:(list Omega))
    n=(length V) ->
    n=(length (makeUninitTupty V)).
Proof.
Induction n.
Intros.
Induction V.
Rewrite H.
Unfold makeUninitTupty.
Unfold length.
Auto.
Absurd O=(length (cons a V)).
Unfold length.
Apply O_S.
Auto.
Intros.
Induction V.
Absurd (length (nil Omega))=(S n0).
Unfold length.
Apply O_S.
Apply sym_eq.
Auto.
Cut n0=(length (makeUninitTupty V)).
Intros.
Unfold makeUninitTupty.
Unfold length.
Fold makeUninitTupty.
Unfold length in H1.
Apply eq_S.
Auto.
Apply H.
Unfold length in H0.
Apply eq_add_S.
Unfold length.
Auto.
Save.



Lemma makeUninitWF : (HT:HeapTy; V:(list Omega))
      	       	     (WFHeapVal HT (tuple (makeUninitTup V))
		                   (tupty V (makeUninitTupty V))).
Proof.
Intros.
Constructor.
Induction V.
Unfold makeUninitTup; Unfold makeUninitTupty.
Constructor.
Unfold makeUninitTup; Unfold makeUninitTupty.
Constructor.
Apply uninit.
Unfold makeUninitTup in HrecV; Unfold makeUninitTupty in HrecV.
Auto.
Save.



Lemma reg_tuple_ex_heaptype
  : (ht:(Map label Omega); wfht:(mWF ? ? ht);
     R:RegFile; G:RegFileTy; 
     r:Reg; tl:(list Omega); il:(list initflag); l:label)
    (WFRegFile (wfmap ? ? ht wfht) R G) ->
    (mlookup ? ? G r (tupty tl il)) ->
    (mlookup ? ? R r (wl l)) ->
    (EX s:Omega | (mlookup nat Omega ht l s) /\ (Subtype s (tupty tl il))).
Proof.
Intros.
Inversion H.
Cut (EX w:WordVal |
       (mlookup Reg WordVal R r w)
       /\(WFWordVal (wfmap nat Omega ht wfht) w (tupty tl il))).
Intros.
Inversion_clear H7.
Inversion_clear H8.
Cut x=(wl l).
Intros.
Rewrite H8 in H9.
Inversion H9.
Split with t'.
Split.
Unfold unwrapMap in H11.
Auto.
Auto.
Apply (mlookup_eq Reg WordVal r x (wl l) R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H2.
Auto.
Save.



(* Heap extension lemmas *)
Lemma heap_ext_6
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t:Omega; l:label)
    (wv:WordVal; t':Omega)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (htextend ht l t ht') ->
    (WFWordVal (wfmap ? ? ht wfht) wv t') ->
    (WFWordVal (wfmap ? ? ht' wfht') wv t').
Proof.
Induction wv.
Intros.
Inversion H2.
Apply label_wval with t'0.
Unfold unwrapMap.
Unfold unwrapMap in H4.
Cut ~l=n.
Intros.
Apply (mextend_preserv nat Omega ht ht' n l t'0 t).
Auto.
Auto.
Auto.
Inversion H.
Inversion_clear H9.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Clear H16.
Unfold unwrapMap in H14.
Cut (lt n l).
Intros.
Cut ~n=l.
Intros.
Auto.
Apply lt_neq.
Auto.
Unfold unwrapMap in H12.
Replace l with x.
Apply H14 with t'0.
Auto.
Apply msize_eq with m:=h.
Apply (wfh_h_size_eq h wfh ht wfht).
Auto.
Auto.
Auto.
Auto.
Intros.
Inversion H2.
Constructor.
Intros.
Inversion H2.
Intros.
Inversion H3.
Apply fold_word_wval with t'0.
Auto.
Apply H.
Auto.
Auto.
Auto.
Auto.
Save.


Lemma heap_ext_7
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t:Omega; l:label)
    (wv:WordVal; t':Omega; b:initflag)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (htextend ht l t ht') ->
    (WFWordValinit (wfmap ? ? ht wfht) wv t' b) ->
    (WFWordValinit (wfmap ? ? ht' wfht') wv t' b).
Proof.
Intros.
Induction b.
Constructor.
Inversion H2.
Apply (heap_ext_6 h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
Inversion H2.
Constructor.
Apply (heap_ext_6 h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
Apply uninit.
Save.


(* in this lemma, I has to be first so that induction on it works *)
Lemma heap_ext_4
  : (I:InstrSeq)
    (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t:Omega; l:label)
    (R:RegFileTy; A:APTy)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (htextend ht l t ht') ->
    (WFInstrSeq (wfmap ? ? ht wfht) R A I) ->
    (WFInstrSeq (wfmap ? ? ht' wfht') R A I).
Proof.
Induction I.
Intros.
Inversion H3.
Apply s_add with G':=G'.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* begin s_addi case *)
Apply s_addi with G'.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_addi case *)
(* begin s_alloc case *)
Apply s_alloc with n:=n G':=G'.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_alloc case *)
Apply s_bgt with G'.
Auto.
Auto.
Unfold unwrapMap.
Unfold unwrapMap in H8.
Cut l=l0\/~l=l0.
Intros.
Inversion_clear H14.
Rewrite <- H15 in H8.
Absurd (mindom ? ? ht l)/\(mnotindom ? ? ht l).
Apply not_min_and_not_wf.
Auto.
Split.
Apply mlookup_imp_mindom with (codety G' A).
Auto.
Inversion H2.
Auto.
Apply (mextend_preserv ? ? ht ht' l0 l (codety G' A) t).
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* begin s_bump case *)
Apply s_bump.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_bump case *)
(* begin s_fold case *)
Apply s_fold with G' rst.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_fold case *)
(* begin s_ld case *)
Apply s_ld with G' ol il t0.
Auto.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_ld case *)
(* begin s_mov and s_movi cases *)
Apply s_mov with G' t0.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
Apply s_movi with G'.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_mov and s_movi cases *)
(* begin s_movl case *)
Cut l=l0\/~l=l0.
Intros.
Inversion_clear H13.
Rewrite H14 in H2.
Absurd (mindom ? ? ht l0)/\(mnotindom ? ? ht l0).
Apply not_min_and_not_wf.
Auto.
Split.
Apply mlookup_imp_mindom with t0.
Auto.
Inversion H2.
Auto.
Apply s_movl with G' t0 t'.
Apply mextend_preserv with ht l t.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
Apply eqorneq_nat.
(* end s_movl case *)
(* begin s_st case *)
Apply s_st with Ts:=Ts V:=V V':=V' t:=t0 G':=G'.
Auto.
Auto.
Auto.
Move H after H13.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_st case *)
(* begin s_unfold case *)
Apply s_unfold with G' f unf rst.
Auto.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
(* end s_unfold case *)
Intros.
Inversion H2.
Apply s_jd with G':=G'.
Unfold unwrapMap.
Unfold unwrapMap in H4.
Inversion H1.
Move H4 after H9.
Cut (mindom nat Omega ht n).
Intros.
Cut l=n\/~l=n.
Intros.
Inversion_clear H15.
Rewrite H16 in H9.
Absurd (mindom nat Omega ht n)/\(mnotindom nat Omega ht n).
Apply not_min_and_not_wf.
Auto.
Split.
Auto.
Auto.
Apply mlooknext.
Auto.
Auto.
Apply eqorneq_nat.
Apply mlookup_imp_mindom with (codety G' A).
Auto.
Auto.
Intros.
Inversion H2.
Apply s_jmp with G'.
Auto.
Auto.
Save.



Lemma heap_ext_5_aux
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (wl:(list WordVal); tl:(list Omega); il:(list initflag))
    (t:Omega; l:label)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (htextend ht l t ht') ->
    (WFWordValinitList (wfmap ? ? ht wfht) wl tl il) ->
    (WFWordValinitList (wfmap ? ? ht' wfht') wl tl il).
Proof.
Induction wl.
Intros.
Inversion H2.
Constructor.
Intros.
Inversion H3.
Constructor.
Apply (heap_ext_7 h h' wfh wfh' ht ht' wfht wfht' t l0 a t0 i).
Auto.
Auto.
Auto.
Auto.
Apply (H tl0 il0 t l0).
Auto.
Auto.
Auto.
Auto.
Save.




Lemma heap_ext_3
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (R,R':RegFile; G,G':(Map Reg Omega))
    (rd:Reg; l:nat; tl:(list Omega); il:(list initflag))
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (WFRegFile (wfmap ? ? ht wfht) R G) -> 
    (mupdext Reg WordVal R rd (wl l) R') ->
    (mupdext Reg Omega G rd (tupty tl il) G') -> 
    (mextend nat Omega ht l (tupty tl il) ht') ->
    (WFRegFile (wfmap ? ? ht' wfht') R' G').
Proof.
Intros.
Constructor.
Intros.
Cut r=rd\/~r=rd.
Intros.
Inversion_clear H6.
Split with (wl l).
Split.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Rewrite H7.
Auto.
Replace t with (tupty tl il).
Inversion H4.
Apply label_wval with (tupty tl il).
Unfold unwrapMap.
Apply mextend_imp_mlookup with ht.
Apply eqorneq_nat.
Auto.
Constructor.
Cut (mlookup Reg Omega G' rd (tupty tl il)).
Intros.
Apply (mlookup_eq Reg Omega rd (tupty tl il) t G').
Apply eqorneq_reg.
Auto.
Rewrite <- H7.
Auto.
Apply mupdext_imp_mlookup with G.
Apply eqorneq_reg.
Auto.
Inversion H1.
Cut (EX w:WordVal |
       (mlookup Reg WordVal R r w)
       /\(WFWordVal (wfmap nat Omega ht wfht) w t)).
Intros.
Inversion_clear H12.
Inversion_clear H13.
Split with x.
Split.
Apply (mupdext_preserv Reg WordVal R R' r rd x (wl l)).
Auto.
Auto.
Auto.
Apply (heap_ext_6 h h' wfh wfh' ht ht' wfht wfht' (tupty tl il) l).
Auto.
Auto.
Auto.
Auto.
Cut (mlookup Reg Omega G r t).
Intros.
Apply H6.
Auto.
Apply (mupdext_preserv_rev Reg Omega G G' r rd t (tupty tl il)).
Auto.
Auto.
Auto.
Apply eqorneq_reg.
Intros.
Cut r=rd\/~r=rd.
Intros.
Inversion_clear H8.
Cut v=(wl l).
Intros.
Rewrite H8 in H6.
Unfold stripWV in H6.
Inversion H6.
Rewrite <- H11.
Replace n with (S l).
Constructor.
Inversion H4.
Unfold unwrapMap in H7.
Rewrite <- H15 in H7.
Inversion H7.
Replace n0 with l.
Auto.
Inversion H.
Inversion_clear H22.
Inversion_clear H25.
Inversion_clear H26.
Clear H27.
Unfold unwrapMap in H22.
Unfold unwrapMap in H25.
Replace n0 with x.
Apply (msize_eq ? ? h l x).
Auto.
Auto.
Apply (msize_eq ? ? ht x n0).
Auto.
Auto.
Apply (mlookup_eq ? ? r v (wl l) R').
Apply eqorneq_reg.
Auto.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Rewrite H9.
Auto.
Inversion H1.
Unfold unwrapMap in H10.
Unfold unwrapMap in H7.
Inversion H4.
Rewrite <- H18 in H7.
Inversion H7.
Cut (lt l0 n0).
Intros.
Apply lt_S.
Auto.
Apply H10 with r v.
Apply mupdext_preserv_rev with R' rd (wl l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply eqorneq_reg.
Save.




Lemma heap_ext_3ap
  : (HT,HT':HeapTy)
    (a:AP; tl:(list Omega); il:(list initflag); n:nat)
    (WFap HT a fresh) ->
    n=(length tl) ->
    (mextend nat Omega (unwrapMap nat Omega HT) a (tupty tl il)
      	       	       (unwrapMap nat Omega HT')) ->
    (WFap HT' a (used n)).
Proof.
Intros.
Induction HT.
Induction HT'.
Unfold unwrapMap in H1.
Apply used_wf with tl il.
Auto.
Unfold unwrapMap.
Inversion H1.
Constructor.
Inversion H.
Unfold unwrapMap in H7.
Auto.
Apply label_wval with (tupty tl il).
Unfold unwrapMap.
Apply mextend_imp_mlookup with m.
Apply eqorneq_nat.
Auto.
Constructor.
Save.



Lemma heap_ext_5
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (hv:HeapVal; t,t':Omega; l:label)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (htextend ht l t ht') ->
    (WFHeapVal (wfmap ? ? ht wfht) hv t') ->
    (WFHeapVal (wfmap ? ? ht' wfht') hv t').
Proof.
Intros.
Induction hv.
Inversion H2.
Constructor.
Apply (heap_ext_5_aux h h' wfh wfh' ht ht' wfht wfht' l0 tl il t l).
Auto.
Auto.
Auto.
Auto.
Inversion H2.
Constructor.
Apply (heap_ext_4 i h h' wfh wfh' ht ht' wfht wfht' t l).
Auto.
Auto.
Auto.
Auto.
Save.



Lemma heap_ext_2 
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (hv:HeapVal; t:Omega; l:label)
(*    (n:nat; V:(vector InitOmega n))
    t=(tupty n V) -> *)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (hsize h l) ->
    (htextend ht l t ht') ->
    (hextend h l hv h') ->
    (WFHeapVal (wfmap ? ? ht' wfht') hv t) ->
    (WFHeap (wfmap ? ? h' wfh') (wfmap ? ? ht' wfht')).
Proof.
Intros.
Constructor.
Exists (S l).
Split.
Unfold unwrapMap.
Inversion H2.
Constructor.
Auto.
Split.
Unfold unwrapMap.
Inversion H1.
Constructor.      
Apply wfh_ht_size_eq with h:=h ht:=ht wfh:=wfh wfht:=wfht.
Auto.
Auto.

Split.
Intros.
Unfold unwrapMap in H4.
Inversion H2.
Rewrite <- H9 in H4.
Cut (l=n \/ ~l=n).
Intros.
Inversion_clear H10.
Replace n with l.
Auto.
Inversion H4.
Absurd ~n=n.
Auto.
Rewrite H10 in H11.
Auto.
Inversion H.
Inversion_clear H19.
Inversion_clear H22.
Inversion_clear H23.
Inversion_clear H24.
Unfold unwrapMap in H23.
Unfold unwrapMap in H19.
Cut (lt n l).
Intros.
Auto.
Replace l with x.
Apply (H23 n h0 H16).
Apply (msize_eq nat HeapVal h).
Auto.
Auto.
Apply eqorneq_nat.

Split.
Intros.
Unfold unwrapMap in H4.
Inversion H1.
Rewrite <- H9 in H4.
Cut (l=n \/ ~l=n).
Intros.
Inversion_clear H10.
Replace n with l.
Auto.
Inversion H4.
Absurd ~n=n.
Auto.
Rewrite H10 in H11.
Auto.
Inversion H.
Inversion_clear H19.
Inversion_clear H22.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H25.
Unfold unwrapMap in H22.
Unfold unwrapMap in H24.
Cut (lt n l).
Intros.
Auto.
Replace l with x.
Apply (H24 n t0 H16).
Apply (msize_eq nat Omega ht).
Auto.
Apply wfh_ht_size_eq with h:=h ht:=ht wfh:=wfh wfht:=wfht.
Auto.
Auto.
Apply eqorneq_nat.

Split.
Intros.
Unfold unwrapMap.
Inversion H2.
Cut (l=n \/ ~l=n).
Intros.
Inversion_clear H10.
Exists hv.
Replace l with n.
Constructor.
Inversion H.
Inversion_clear H12.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Unfold unwrapMap in H18.
Cut (EX h0:HeapVal | (mlookup nat HeapVal h n h0)).
Intros.
Inversion_clear H19.
Exists x0.
Constructor.
Auto.
Auto.
Cut (lt n x).
Intros.
Apply (H18 n H19).
Replace x with l.
Apply lt_S_neq.
Auto.
Auto.
Apply (msize_eq nat HeapVal h).
Auto.
Auto.
Apply eqorneq_nat.

Split.
Intros.
Unfold unwrapMap.
Inversion H1.
Cut (l=n \/ ~l=n).
Intros.
Inversion_clear H10.
Exists t.
Replace l with n.
Constructor.
Inversion H.
Inversion_clear H12.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H20.
Unfold unwrapMap in H19.
Cut (EX t:Omega | (mlookup nat Omega ht n t)).
Intros.
Inversion_clear H20.
Exists x0.
Constructor.
Auto.
Auto.
Cut (lt n x).
Intros.
Apply (H19 n H20).
Replace x with l.
Apply lt_S_neq.
Auto.
Auto.
Apply (msize_eq nat HeapVal h).
Auto.
Auto.
Apply eqorneq_nat.

Split.
Intros.
Unfold unwrapMap in H4.
Unfold unwrapMap in H5.
Cut (l=n \/ ~l=n).
Intros.
Inversion_clear H6.
Inversion H1.
Inversion H2.
Rewrite <- H16 in H4.
Rewrite <- H11 in H5.
Rewrite H7 in H4.
Rewrite H7 in H5.
Replace h0 with hv.
Replace t0 with t.
Auto.
Inversion H5.
Auto.
Absurd ~n=n.
Auto.
Auto.
Inversion H4.
Auto.
Absurd ~n=n.
Auto.
Auto.

Apply (heap_ext_5 h h' wfh wfh' ht ht' wfht wfht' h0 t t0 l).
Auto.
Auto.
Auto.
Inversion H.
Inversion_clear H8.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Inversion H1. 
Inversion H2.
Rewrite <- H26 in H4.
Rewrite <- H21 in H5.
Inversion H4.
Absurd ~n=n.
Rewrite H27 in H7.
Auto.
Auto.
Inversion H5.
Absurd ~n=n.
Auto.
Rewrite H34 in H7.
Auto.
Unfold unwrapMap in H17.
Inversion_clear H17.
Apply (H41 n h0 t0 H32 H39).
Apply eqorneq_nat.

Intros.
Unfold unwrapMap.
Inversion H2.
Constructor.
Auto.

Inversion H.
Inversion_clear H10.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Auto.
Save.


Lemma heap_ext_1 : (H:Heap; HT:HeapTy; 
      	       	       	ht:(Map label Omega); wfht:(mWF ? ? ht);
      	       	       	l:label; t:Omega)
		       HT=(wfmap label Omega ht wfht) ->
		       (WFHeap H HT) ->
		       (msize ? ? ht l) ->
      	       	       (EX ht' | (EXT wfht' | (EX HT':HeapTy | 
		         HT'=(wfmap label Omega ht' wfht') /\
			 (mextend ? ? ht l t ht')))).
Proof.
Intros.
Split with (melem ? ? l t ht).
Cut (mWF nat Omega (melem nat Omega l t ht)).
Intros.
Split with H3.
Split with (wfmap nat Omega (melem nat Omega l t ht) H3).
Split.
Auto.
Constructor.
Apply (size_notin_wfheapty H HT ht wfht l H0 H1 H2).
Constructor.
Apply (size_notin_wfheapty H HT ht wfht l H0 H1 H2).
Auto.
Save.



(* heap update lemmas *)
Lemma subtype_update
  : (i:int; il1,il1',il0,il0':(list initflag))
    (SubtypeBList il1 il0) ->
    (updatetupty il1 i il1') ->
    (updatetupty il0 i il0') ->
    (SubtypeBList il1' il0').    
Proof.
Induction i.
Intros.
Inversion H0.
Inversion H1.
Constructor.
Constructor.
Rewrite <- H2 in H.
Rewrite <- H3 in H.
Inversion H.
Auto.
Intros.
Inversion H1.
Inversion H2.
Rewrite <- H4 in H0.
Rewrite <- H8 in H0.
Inversion H0.
Constructor.
Auto.
Apply (H V V' V0 V'0).
Auto.
Auto.
Auto.
Save.


Lemma subtypeb_trans
  : (b,a,c:initflag)
    (SubtypeB a b) ->
    (SubtypeB b c) ->
    (SubtypeB a c).
Proof.
Induction b.
Intros.
Inversion H.
Inversion H0.
Constructor.
Constructor.
Intros.
Inversion H.
Inversion H0.
Constructor.
Inversion H0.
Constructor.
Save.


Lemma subtypeblist_trans
  : (il2,il1,il3:(list initflag))
    (SubtypeBList il1 il2) ->
    (SubtypeBList il2 il3) ->
    (SubtypeBList il1 il3).
Proof.
Induction il2.
Intros.
Inversion H.
Inversion H0.
Constructor.
Intros.
Inversion H0.
Inversion H1.
Constructor.
Apply (subtypeb_trans a i1 i3).
Auto.
Auto.
Apply H.
Auto.
Auto.
Save.


Lemma subtype_trans
  : (t1,t2,t3:Omega)
    (Subtype t1 t2) ->
    (Subtype t2 t3) ->
    (Subtype t1 t3).
Proof.
Intros.
Inversion H.
Auto.
Rewrite <- H3 in H0.
Inversion H0.
Constructor.
Auto.
Constructor.
Apply subtypeblist_trans with il2.
Auto.
Auto.
Save.


Lemma heap_upd_6
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (wv:WordVal; t':Omega)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (WFWordVal (wfmap ? ? ht wfht) wv t') ->
    (WFWordVal (wfmap ? ? ht' wfht') wv t').
Proof.
Induction wv.
Intros.
Cut l=n\/~l=n.
Intros.
Inversion_clear H4.
Rewrite <- H5 in H3.
Replace n with l.
Inversion H3.
Apply label_wval with t.
Unfold unwrapMap.
Apply mupdate_imp_mlookup with ht.
Apply eqorneq_nat.
Auto.
Cut t'0=s.
Intros.
Rewrite H10 in H8.
Apply subtype_trans with s.
Auto.
Auto.
Unfold unwrapMap in H6.
Apply (mlookup_eq ? ? l t'0 s ht).
Apply eqorneq_nat.
Auto.
Auto.
Inversion H3.
Apply label_wval with t'0.
Unfold unwrapMap.
Unfold unwrapMap in H6.
Apply (mupdate_preserv ? ? ht ht' n l t'0 t).
Auto.
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Intros.
Inversion H3.
Constructor.
Intros.
Inversion H3.
Intros.
Inversion H4.
Apply fold_word_wval with t'0.
Auto.
Apply H.
Auto.
Auto.
Auto.
Auto.
Auto.
Save.



Lemma heap_upd_7
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (wv:WordVal; t':Omega; b:initflag)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (WFWordValinit (wfmap ? ? ht wfht) wv t' b) ->
    (WFWordValinit (wfmap ? ? ht' wfht') wv t' b).
Proof.
Induction b.
Intros.
Constructor.
Inversion H3.
Apply (heap_upd_6 h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Intros.
Inversion H3.
Constructor.
Apply (heap_upd_6 h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply uninit.
Save.






(* I has to be first so that induction on it works *)
Lemma heap_upd_4
  : (I:InstrSeq)
    (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (R:RegFileTy; A:APTy)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (WFInstrSeq (wfmap ? ? ht wfht) R A I) ->
    (WFInstrSeq (wfmap ? ? ht' wfht') R A I).
Proof.
Induction I.
Intros.
Inversion H4.
Apply s_add with G':=G'.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_addi with G'.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_alloc with n:=n G':=G'.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_bgt with G'.
Auto.
Auto.
Unfold unwrapMap.
Unfold unwrapMap in H9.
Cut l=l0\/~l=l0.
Intros.
Inversion_clear H15.
Rewrite <- H16 in H9.
Rewrite <- H16.
Replace (codety G' A) with t.
Apply mupdate_imp_mlookup with ht.
Apply eqorneq_nat.
Auto.
Cut s=(codety G' A).
Intros.
Rewrite H15 in H2.
Apply subtype_codety with (codety G' A).
Auto.
Auto.
Apply (mlookup_eq ? ? l s (codety G' A) ht).
Apply eqorneq_nat.
Auto.
Auto.
Apply (mupdate_preserv ? ? ht ht' l0 l (codety G' A) t).
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_bump.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_fold with G' rst.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_ld with G' ol il t0.
Auto.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_mov with G' t0.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_movi with G'.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Unfold unwrapMap in H7.
Cut l=l0\/~l=l0.
Intros.
Inversion_clear H14.
Rewrite <- H15 in H7.
Move H1 after H7.
Cut t0=s.
Intros.
Apply s_movl with G' t t'.
Unfold unwrapMap.
Apply mupdate_imp_mlookup with ht.
Apply eqorneq_nat.
Rewrite <- H15.
Auto.
Apply subtype_trans with s.
Auto.
Rewrite <- H14.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (mlookup_eq ? ? l t0 s ht).
Apply eqorneq_nat.
Auto.
Auto.
Apply s_movl with G' t0 t'.
Unfold unwrapMap.
Apply mupdate_preserv with ht l t.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Apply s_st with Ts:=Ts V:=V t:=t0 V':=V' G':=G'.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply s_unfold with G' f unf rst.
Auto.
Auto.
Auto.
Auto.
Apply (H h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Intros.
Inversion H3.
Unfold unwrapMap in H5.
Apply s_jd with G':=G'.
Unfold unwrapMap.
Cut l=n\/~l=n.
Intros.
Inversion_clear H10.
Apply mupdate_imp_mlookup with ht.
Apply eqorneq_nat.
Cut t=(codety G' A).
Intros.
Rewrite <- H10.
Rewrite <- H11.
Auto.
Cut s=(codety G' A).
Intros.
Apply (subtype_codety t s G' A).
Auto.
Apply (mlookup_eq nat Omega n s (codety G' A) ht).
Apply eqorneq_nat.
Rewrite <- H11.
Auto.
Auto.
Apply (mlookup_eq nat Omega n s (codety G' A) ht).
Apply eqorneq_nat.
Rewrite <- H11.
Auto.
Auto.
Apply (mupdate_preserv nat Omega ht ht' n l (codety G' A) t).
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Auto.
Intros.
Inversion H3.
Apply s_jmp with G'.
Auto.
Auto.
Save.


Lemma heap_upd_5_aux
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (wl:(list WordVal); tl:(list Omega); il:(list initflag))
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (WFWordValinitList (wfmap ? ? ht wfht) wl tl il) ->
    (WFWordValinitList (wfmap ? ? ht' wfht') wl tl il).
Proof.
Induction wl.
Intros.
Inversion H3.
Constructor.
Intros.
Inversion H4.
Constructor.
Apply (heap_upd_7 h h' wfh wfh' ht ht' wfht wfht' t s l a t0 i).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H tl0).
Auto.
Auto.
Auto.
Auto.
Auto.
Save.


Lemma heap_upd_3
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (R:RegFile; G:(Map Reg Omega))
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (WFRegFile (wfmap ? ? ht wfht) R G) -> 
    (WFRegFile (wfmap ? ? ht' wfht') R G).
Proof.
Intros.
Constructor.
Intros.
Inversion H3.
Cut (EX w:WordVal |
       (mlookup Reg WordVal R r w)
       /\(WFWordVal (wfmap nat Omega ht wfht) w t0)).
Intros.
Inversion_clear H10.
Inversion_clear H11.
Split with x.
Split.
Auto.
Apply (heap_upd_6 h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply H5.
Auto.
Inversion H3.
Intros.
Apply H5 with r v.
Auto.
Auto.
Unfold unwrapMap.
Unfold unwrapMap in H10.
Apply mupdate_size_same_rev with ht' l t.
Apply eqorneq_nat.
Auto.
Auto.
Save.



Lemma heap_upd_3ap
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (a:AP; at:APTy)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (WFap (wfmap ? ? ht wfht) a at) ->
    (WFap (wfmap ? ? ht' wfht') a at).
Proof.
Intros.
Induction at.
Inversion H3.
Unfold unwrapMap in H4.
Constructor.
Unfold unwrapMap.
Apply (mupdate_size_same nat Omega ht ht' l t a).
Apply eqorneq_nat.
Auto.
Auto.
Inversion H3.
Unfold unwrapMap in H6.
Apply used_wf with tl ol.
Auto.
Unfold unwrapMap.
Apply (mupdate_size_same nat Omega ht ht' l t (S a)).
Apply eqorneq_nat.
Auto.
Auto.
Apply (heap_upd_6 h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Save.


Lemma heap_upd_5
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (hv:HeapVal; t':Omega)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (WFHeapVal (wfmap ? ? ht wfht) hv t') ->
    (WFHeapVal (wfmap ? ? ht' wfht') hv t').
Proof.
Intros.
Induction hv.
Inversion H3.
Constructor.
Apply (heap_upd_5_aux h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Inversion H3.
Constructor.
Apply (heap_upd_4 i h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Save.



Lemma ord_updateheap_ord
  : (h,h':HeapMap; l:label; hv:HeapVal)
    (mupdate ? ? h l hv h') ->
    (OrdHeap h) ->
    (OrdHeap h').
Proof.
Induction h.
Intros.
Inversion H.
Intros.
Inversion H0.
Inversion H1.
Constructor.
Rewrite <- H2.
Auto.
Auto.
Constructor.
Inversion H1.
Apply mupdate_size_same with m l hv.
Apply eqorneq_nat.
Auto.
Auto.
Apply H with l hv.
Auto.
Inversion H1.
Auto.
Save.


Lemma heap_upd_2
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h');
     ht,ht':(Map label Omega); wfht:(mWF ? ? ht); wfht':(mWF ? ? ht'))
    (t,s:Omega; l:label)
    (hv:HeapVal)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htlookup ht l s) ->
    (Subtype t s) ->
    (htupdate ht l t ht') ->
    (hupdate h l hv h') ->
    (WFHeapVal (wfmap ? ? ht' wfht') hv t) ->
    (WFHeap (wfmap ? ? h' wfh') (wfmap ? ? ht' wfht')).
Proof.
Intros.
Constructor.
Inversion H.
Inversion_clear H6.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Split with x.
Split.
Unfold unwrapMap.
Unfold unwrapMap in H6.
Apply mupdate_size_same with m:=h m':=h' a:=l b:=hv.
Apply eqorneq_nat.
Auto.
Auto.
Split.
Unfold unwrapMap.
Unfold unwrapMap in H9.
Apply mupdate_size_same with m:=ht m':=ht' a:=l b:=t.
Apply eqorneq_nat.
Auto.
Auto.
Split.
Unfold unwrapMap.
Move H10 after H15.
Unfold unwrapMap in H10.
Intros.
Cut (EX c:? | (mlookup ? ? h n c)).
Intros.
Inversion_clear H16.
Apply H10 with x0.
Auto.
Apply mupdate_preserv_rev_ex with a':=l b':=hv m':=h' b:=h0.
Auto.
Auto.
Apply eqorneq_nat.
Auto.
Auto.
Split.
Intros.
Unfold unwrapMap in H14.
Unfold unwrapMap in H11.
Cut (EX c:? | (mlookup ? ? ht n c)).
Intros.
Inversion_clear H16.
Apply H11 with x0.
Auto.
Apply mupdate_preserv_rev_ex with a':=l b':=t m':=ht' b:=t0.
Auto.
Auto.
Apply eqorneq_nat.
Auto.
Auto.
Split.
Intros.
Unfold unwrapMap.
Unfold unwrapMap in H12.
Move H12 after H14.
Cut n=l\/~n=l.
Intros.
Inversion_clear H16.
Replace n with l.
Split with hv.
Apply mupdate_imp_mlookup with h.
Apply eqorneq_nat.
Auto.
Cut (EX h0:HeapVal | (mlookup nat HeapVal h n h0)).
Intros.
Inversion_clear H16.
Split with x0.
Apply (mupdate_preserv nat HeapVal h h' n l x0 hv).
Auto.
Auto.
Auto.
Apply H12.
Auto.
Apply eqorneq_nat.
Split.
Intros.
Unfold unwrapMap.
Unfold unwrapMap in H13.
Move H13 after H14.
Cut n=l\/~n=l.
Intros.
Inversion_clear H16.
Replace n with l.
Split with t.
Apply mupdate_imp_mlookup with ht.
Apply eqorneq_nat.
Auto.
Cut (EX t:? | (mlookup nat Omega ht n t)).
Intros.
Inversion_clear H16.
Split with x0.
Apply (mupdate_preserv nat Omega ht ht' n l x0 t).
Auto.
Auto.
Auto.
Apply H13.
Auto.
Apply eqorneq_nat.
Unfold unwrapMap.

Split.
Intros.
Unfold unwrapMap in H15.
Clear H13.
Clear H12.
Clear H11.
Clear H10.
Cut n=l\/~n=l.
Intros.
Inversion_clear H10.
Replace h0 with hv.
Replace t0 with t.
Auto.
Cut (mlookup ? ? ht' n t).
Intros.
Apply mlookup_eq with m:=ht' a:=n.
Apply eqorneq_nat.
Auto.
Auto.
Apply mupdate_imp_mlookup with ht.
Apply eqorneq_nat.
Replace n with l.
Auto.
Cut (mlookup ? ? h' n hv).
Intros.
Apply mlookup_eq with m:=h' a:=n.
Apply eqorneq_nat.
Auto.
Auto.
Apply mupdate_imp_mlookup with h.
Apply eqorneq_nat.
Replace n with l.
Auto.
Cut (WFHeapVal (wfmap ? ? ht wfht) h0 t0).
Intros.
Apply (heap_upd_5 h h' wfh wfh' ht ht' wfht wfht' t s l).
Auto.
Auto.
Auto.
Auto.
Auto.
Inversion_clear H15.
Apply (H10 n h0 t0).
Apply (mupdate_preserv_rev nat HeapVal h h' n l h0 hv).
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Auto.
Auto.
Auto.
Apply (mupdate_preserv_rev nat Omega ht ht' n l t0 t).
Apply eqorneq_nat.
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Intros.
Apply ord_updateheap_ord with h l hv.
Auto.
Inversion_clear H15.
Auto.
Save.




(* ----------------------------- *)


Lemma inv_reg_wval
  : (HT:HeapTy; R:RegFile; G:RegFileTy)
    (r:Reg; w:WordVal; t:Omega)
    (WFRegFile HT R G) ->
    (mlookup ? ? G r t) ->
    (mlookup ? ? R r w) ->
    (WFWordVal HT w t).
Proof.
Intros.
Inversion H.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w t)).
Intros.
Inversion_clear H7.
Inversion_clear H8.
Replace w with x.
Auto.
Apply (mlookup_eq ? ? r x w R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H2.
Auto.
Save.


Lemma inv_reg_wval_tupty
  : (HT:HeapTy; R:RegFile; G:RegFileTy)
    (r:Reg; l:label; tl:(list Omega); il:(list initflag))
    (WFRegFile HT R G) ->
    (mlookup ? ? G r (tupty tl il)) ->
    (mlookup ? ? R r (wl l)) ->
    (WFWordVal HT (wl l) (tupty tl il)).
Proof.
Intros.
Inversion H.
Cut (EX w:? | (mlookup Reg WordVal R r w)/\(WFWordVal HT w (tupty tl il))).
Intros.
Inversion_clear H7.
Inversion_clear H8.
Replace (wl l) with x.
Auto.
Apply (mlookup_eq ? ? r x (wl l) R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H2.
Auto.
Save.


Lemma inv_reg_intty
  : (H:Heap; HT:HeapTy; R:RegFile; G:RegFileTy; r:Reg)
    (WFHeap H HT) ->
    (WFRegFile HT R G) ->
    (mlookup ? ? G r intty) ->
    (EX i | (mlookup ? ? R r (wi i))).
Proof.
Intros.
Inversion H1.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w intty)).
Intros.
Inversion_clear H8.
Inversion_clear H9.
Inversion H10.
Split with i.
Replace (wi i) with x.
Auto.
Cut t'=intty.
Intros.
Cut ~t'=intty.
Intros.
Rewrite H15 in H16.
Absurd ~intty=intty.
Auto.
Auto.
Apply heap_lookup_not_intty with H HT l.
Auto.
Auto.
Apply subtype_intty with intty.
Auto.
Auto.
Apply H3.
Auto.
Save.


Lemma subtypeblist_refl
  : (V:(list initflag))
    (SubtypeBList V V).
Proof.
Induction V.
Constructor.
Intros.
Constructor.
Constructor.
Auto.
Save.


Lemma updatetupty_subtypeblist
  : (i:nat; V,V':(list initflag)) (updatetupty V i V') -> (SubtypeBList V' V).
Proof.
Induction i.
Intros.
Inversion H.
Constructor.
Induction b.
Constructor.
Constructor.
Apply subtypeblist_refl.
Intros.
Inversion H0.
Constructor.
Constructor.
Apply H.
Auto.
Save.


Lemma updatetupty_subtype
  : (i:nat; tl:(list Omega); il,il':(list initflag))
    (updatetupty il' i il) ->
    (Subtype (tupty tl il) (tupty tl il')).
Proof.
Induction i.
Intros.
Inversion H.
Constructor.
Constructor.
Induction b.
Constructor.
Constructor.
Apply subtypeblist_refl.
Intros.
Inversion H0.
Constructor.
Constructor.
Constructor.
Apply updatetupty_subtypeblist with n.
Auto.
Save.


Lemma inv_label_lookup_subtype
  : (HT:HeapTy; R:RegFile; G:RegFileTy; r:Reg; l:label; t:Omega)
    (WFRegFile HT R G) ->
    (mlookup ? ? R r (wl l)) ->
    (mlookup ? ? G r t) ->
    (EX s | (htlookup (htunwrap HT) l s) /\ (Subtype s t)).
Proof.
Intros.
Inversion H.
Induction HT.
Unfold unwrapMap.
Cut (WFWordVal (wfmap ? ? m m0) (wl l) t).
Intros.
Inversion H7.
Split with t'.
Unfold unwrapMap in H9.
Split.
Auto.
Auto.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal (wfmap ? ? m m0) w t)).
Intros.
Inversion_clear H7.
Inversion_clear H8.
Replace (wl l) with x.
Auto.
Apply (mlookup_eq ? ? r x (wl l) R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H2.
Auto.
Save.


Lemma updatetuple_wfheapval
  : (HT:HeapTy; i:int; V,V':(list WordVal); w:WordVal; t:Omega;
     Ts:(list Omega); il,il':(list initflag))
    (WFHeapVal HT (tuple V) (tupty Ts il)) ->
    (WFWordVal HT w t) ->
    (ListNth ? Ts i t) ->
    (updatetuple V i w V') ->
    (updatetupty il i il') ->
    (WFHeapVal HT (tuple V') (tupty Ts il')).
Induction i.
Intros.
Inversion H2.
Rewrite <- H4 in H.
Inversion H3.
Rewrite <- H5 in H.
Inversion H.
Inversion H12.
Constructor.
Constructor.
Apply init.
Rewrite <- H18 in H1.
Inversion H1.
Auto.
Auto.
Intros.
Inversion H3.
Inversion H4.
Inversion H2.
Cut (WFHeapVal HT (tuple V'0) (tupty V2 V'1)).
Intros.
Inversion H18.
Constructor.
Constructor.
Rewrite <- H6 in H0.
Rewrite <- H15 in H0.
Rewrite <- H11 in H0.
Inversion H0.
Inversion H27.
Auto.
Auto.
Apply (H V0 V'0 w t V2 V1 V'1).
Constructor.
Rewrite <- H6 in H0.
Rewrite <- H11 in H0.
Rewrite <- H15 in H0.
Inversion H0.
Inversion H21.
Auto.
Auto.
Auto.
Auto.
Auto.
Save.







(* cases of the progress proof *)
Lemma progress_add : (H:Heap; R:RegFile; I:InstrSeq; a:AP;
      	       	      rd,rs,rs':Reg)
		     (WFProgram (H, (R, (a, (iseq (add rd rs rs') I))))) ->
		     (EX P' | 
      	       	       	 (Eval (H, (R, (a, (iseq (add rd rs rs') I)))) P')).
Proof.
Intros; Inversion H0.
Inversion H9.
Cut (EX rsval:? | (mlookup ? ? R rs (wi rsval))); Intros.
Inversion_clear H22.
Cut (EX rsval':? | (mlookup ? ? R rs' (wi rsval'))); Intros.
Inversion_clear H22.
Cut (EX R':? | (mupdext ? ? R rd (wi (plus x x0)) R')); Intros.
Inversion_clear H22.
Split with (H,(x1,(a,I))).
Apply ev_add with rsval:=x rsval':=x0 R':=x1.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Inversion H7.
Cut (EX w:? | (mlookup ? ? R rs' w)/\(WFWordVal HT w intty)).
Intros.
Inversion_clear H28.
Inversion_clear H29.
Inversion H30.
Split with i.
Rewrite <- H31 in H28.
Auto.
Cut t'=intty.
Intros.
Cut ~t'=intty.
Intros.
Rewrite H35 in H36.
Absurd ~intty=intty.
Auto.
Auto.
Apply heap_lookup_not_intty with H:=H HT:=HT l:=l0 t:=t'.
Auto.
Auto.
Apply subtype_intty with intty.
Auto.
Auto.
Apply (H22 rs' intty).
Auto.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal HT w intty)).
Intros.
Inversion_clear H22.
Inversion_clear H23.
Inversion H24.
Split with i.
Rewrite <- H25 in H22.
Auto.
Cut t'=intty.
Intros.
Cut ~t'=intty.
Intros.
Rewrite H29 in H30.
Absurd ~intty=intty.
Auto.
Auto.
Apply heap_lookup_not_intty with H:=H HT:=HT l:=l0 t:=t'.
Auto.
Auto.
Apply subtype_intty with intty.
Auto.
Auto.
Inversion H7.
Apply (H22 rs intty).
Auto.
Save.


Lemma progress_addi : (H:Heap; R:RegFile; I:InstrSeq; a:AP;
      	       	       rd,rs:Reg; i:int)
		      (WFProgram (H, (R, (a, (iseq (addi rd rs i) I))))) ->
		      (EX P' | 
      	       	       	 (Eval (H, (R, (a, (iseq (addi rd rs i) I)))) P')).
Proof.
Intros; Inversion H0. Clear H10.
Inversion H9.
Cut (EX rsval:? | (mlookup ? ? R rs (wi rsval))); Intros.
Inversion_clear H20.
Cut (EX R':? | (mupdext ? ? R rd (wi (plus x i)) R')); Intros.
Inversion_clear H20.
Split with (H,(x0,(a,I))).
Apply ev_addi with x.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Inversion H7.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal HT w intty)).
Intros.
Inversion_clear H25.
Inversion_clear H26.
Inversion H27.
Split with i1.
Rewrite H28.
Auto.
Cut t'=intty.
Intros.
Cut ~t'=intty.
Intros.
Rewrite H32 in H33.
Absurd ~intty=intty; Auto.
Apply heap_lookup_not_intty with H HT l0.
Auto.
Auto.
Apply subtype_intty with intty.
Auto.
Auto.
Apply H20.
Auto.
Save.


Lemma progress_jmp : (H:Heap; R:RegFile; a:AP; r:Reg)
		     (WFProgram (H, (R, (a, (jmp r))))) ->
		     (EX P' | 
      	       	       	 (Eval (H, (R, (a, (jmp r)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w (codety G' t))).
Intros.
Elim H16; Clear H16; Intros w H16.
Inversion_clear H16.
Cut (EX l:? |
       w=(wl l)
       /\(EX I:? | (mlookup ? ? (unwrapMap ? ? H) l (code G' t I)))).
Intros.
Elim H16; Clear H16; Intros l0 H16.
Inversion_clear H16.
Elim H20; Clear H20; Intros I0 H20.
Split with (H,(R,(a,I0))).
Apply ev_jmp with l0 G' t.
Rewrite <- H19.
Auto.
Auto.
Apply can_word_forms_code with HT.
Auto.
Auto.
Inversion H7.
Apply H16.
Auto.
Save.


Lemma progress_jd : (H:Heap; R:RegFile; a:AP; l:label)
		     (WFProgram (H, (R, (a, (jd l))))) ->
		     (EX P' | 
      	       	       	 (Eval (H, (R, (a, (jd l)))) P')).
Proof.
Intros.
Inversion H0. Clear H10.
Inversion H9.
Cut (EX I:? | (mlookup ? ? (unwrapMap ? ? H) l (code G' t I))).
Intros.
Inversion_clear H16.
Split with (H,(R,(a,x))).
Apply ev_jd with G:=G' T:=t.
Auto.
Apply can_heap_forms_code with HT:=HT.
Auto.
Auto.
Save.



Lemma progress_ld : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	    (rd,rs:Reg; i:int)
		    (WFProgram (H, (R, (a, (iseq (ld rd rs i) I))))) ->
		    (EX P' | 
		       (Eval (H, (R, (a, (iseq (ld rd rs i) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX l:? | (mlookup ? ? R rs (wl l))).
Intros.
Elim H22; Clear H22; Intros l0 H22.
Cut (EX V:? | (hlookup (hunwrap H) l0 (tuple V))).
Intros.
Elim H23; Clear H23; Intros V H23.
Cut (EX v:? | (ListNth ? V i v)).
Intros.
Elim H24; Clear H24; Intros v H24.
Cut (EX R':? | (mupdext ? ? R rd v R')).
Intros.
Elim H25; Clear H25; Intros R' H25.
Split with (H,(R',(a,I))).
Apply ev_ld with l0 V v.
Auto.
Auto.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Apply lt_imp_listnth.
Replace (length V) with (length ol).
Apply listnth_lt_length with t0.
Auto.
Apply sym_eq.
Apply (tupty_lists_length_eq H HT R G rs l0 ol il V).
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (can_heap_forms_reg_tuple H HT R G rs l0 ol il).
Auto.
Auto.
Auto.
Auto.
Apply (can_reg_forms_tuple HT R G rs ol il).
Auto.
Auto.
Save.



Lemma progress_mov : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	     (rd,rs:Reg)
		     (WFProgram (H,(R,(a,(iseq (mov rd rs) I))))) ->
		     (EX P' |
		         (Eval (H,(R,(a,(iseq (mov rd rs) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX v | (mlookup ? ? R rs v) /\ (WFWordVal HT v t0)).
Intros.
Inversion_clear H19.
Inversion_clear H20.
Cut (EX R':? | (mupdext ? ? R rd x R')).
Intros.
Inversion_clear H20.
Split with (H,(x0,(a,I))).
Apply ev_mov with x.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Inversion H7.
Apply H19.
Auto.
Save.


Lemma progress_movi : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	      (rd:Reg; i:int)
		      (WFProgram (H,(R,(a,(iseq (movi rd i) I))))) ->
		      (EX P' |
		         (Eval (H,(R,(a,(iseq (movi rd i) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX R':? | (mupdext ? ? R rd (wi i) R')).
Intros.
Inversion_clear H18.
Split with (H,(x,(a,I))).
Constructor.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Save.


Lemma progress_movl : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	      (rd:Reg; l:label)
		      (WFProgram (H,(R,(a,(iseq (movl rd l) I))))) ->
		      (EX P' |
		         (Eval (H,(R,(a,(iseq (movl rd l) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX R':? | (mupdext ? ? R rd (wl l) R')).
Intros.
Inversion_clear H20.
Split with (H,(x,(a,I))).
Constructor.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Save.



Lemma progress_unfold : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	       (rd,rs:Reg)
		       (WFProgram (H, (R, (a, (iseq (unfold rd rs) I))))) ->
		       (EX P' | 
		          (Eval (H, (R, (a, (iseq (unfold rd rs) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX rsval:? | (mlookup ? ? R rs rsval)); Intros.
Elim H21; Clear H21; Intros rsval; Intros.
Cut (EX v:? | (EX t:? | rsval=(wfold v t))).
Intros.
Elim H22; Clear H22; Intros v H22.
Elim H22; Clear H22; Intros t' H22.
Cut (EX R':? | (mupdext ? ? R rd v R')).
Intros.
Elim H23; Clear H23; Intros R' H23.
Split with (H,(R',(a,I))).
Apply (ev_unfold H R R' a I rd rs v t').
Rewrite <- H22.
Auto.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Apply can_word_forms_rec with H HT f.
Auto.
Rewrite H14 in H13. 
Apply inv_reg_wval with R G rs.
Auto.
Auto.
Auto.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal HT w rst)).
Intros.
Inversion_clear H21.
Inversion_clear H22.
Split with x.
Auto.
Inversion H7.
Apply H21.
Auto.
Save.



Lemma progress_fold  : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	       (rd,rs:Reg; t:Omega)
		       (WFProgram (H, (R, (a, (iseq (fold rd t rs) I))))) ->
		       (EX P' | 
		          (Eval (H, (R, (a, (iseq (fold rd t rs) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX rsval:? | (mlookup ? ? R rs rsval)); Intros.
Elim H21; Clear H21; Intros rsval; Intros.
Cut (EX R':? | (mupdext ? ? R rd (wfold rsval t) R')); Intros.
Elim H22; Clear H22; Intros R'; Intros.
Split with (H,(R',(a,I))).
Induction H.
Apply (ev_fold (wfmap nat HeapVal m m0) R R' a I rd rs
        (recty f) rsval).
Auto.
Auto.
Rewrite H11.
Auto.
Apply mupdext_always.
Apply eqorneq_reg.
Inversion H7.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal HT w rst)).
Intros.
Inversion_clear H26.
Inversion_clear H27.
Split with x.
Auto.
Apply H21.
Auto.
Save.


Lemma progress_bgt   : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	       (rs,rt:Reg; l:label)
		       (WFProgram (H, (R, (a, (iseq (bgt rs rt l) I))))) ->
		       (EX P' | 
		          (Eval (H, (R, (a, (iseq (bgt rs rt l) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX rsval:? | (mlookup ? ? R rs (wi rsval))); Intros.
Cut (EX rtval:? | (mlookup ? ? R rt (wi rtval))); Intros.
Elim H22; Clear H22; Intros rsval H22; Elim H23; Clear H23; Intros
  rtval H23.
Cut (le rsval rtval)\/(gt rsval rtval).
Intros.
Inversion_clear H24.
Split with (H,(R,(a,I))).
Apply ev_bgt0 with rsval rtval.
Auto.
Auto.
Auto.
Cut (EX I'':? | (hlookup (hunwrap H) l (code G' t I''))).
Intros.
Inversion_clear H24.
Split with (H,(R,(a,x))).
Apply ev_bgt1 with G' t rsval rtval.
Auto.
Auto.
Auto.
Auto.
Apply can_heap_forms_code with HT.
Auto.
Auto.
Apply le_or_gt.
Apply inv_reg_intty with H HT G.
Auto.
Auto.
Auto.
Apply inv_reg_intty with H HT G.
Auto.
Auto.
Auto.
Save.


Lemma progress_bump  : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	       (n:nat)
		       (WFProgram (H, (R, (a, (iseq (bump n) I))))) ->
		       (EX P' | 
		          (Eval (H, (R, (a, (iseq (bump n) I)))) P')).
Proof.
Intros.
Inversion H0. Clear H10. Inversion H9.
Cut (EX l:? | (hsize (hunwrap H) l)).
Intros.
Inversion_clear H16.
Split with (H,(R,(x,I))).
Apply ev_bump.
Auto.
Apply msize_always.
Save.



Lemma progress_alloc : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	       ( rd:Reg; V:(list Omega))
      	       	       (WFProgram (H, (R, (a, (iseq (alloc rd V) I))))) ->
		       (EX P' | 
		         (Eval (H, (R, (a, (iseq (alloc rd V) I)))) P')).
Intros; Inversion H0. Clear H10. Inversion H9.
Induction H.
Induction HT.
Cut (EX R' | (mupdext ? ? R rd (wl a) R')).
Cut (EX h' | (EXT wfh':(mWF ? ? h') | (mextend ? ? m a (tuple (makeUninitTup V)) h'))).
Intros.
Inversion_clear H.
Inversion_clear H20.
Inversion_clear H19.
Exists ( (wfmap ? ? x x0), (x1, (a, I))).
Apply (ev_alloc (wfmap nat HeapVal m m0) (wfmap nat HeapVal x x0) R x1 a I m x m0 x0).
Auto.
Auto.
Auto.
Auto.
(* prove heap ext *)
Cut (mnotindom ? ? m a).
Intros.
Exists (melem ? ? a (tuple (makeUninitTup V)) m).
Cut (mWF ? ? (melem ? ? a (tuple (makeUninitTup V)) m)).
Intros.
Exists H19.
Constructor.
Auto.
Apply mwf1.
Auto.
Auto.
(* prove mnotindom ? ? m a *)
Apply size_notin_wfheap with l:=a wfh:=m0 H:=(wfmap ? ? m m0) HT:=(wfmap ? ? m1 m2).
Auto.
Auto.
Rewrite <- H15 in H8.
Inversion H8.
Unfold unwrapMap in H.
Inversion H6.
Inversion_clear H22.
Inversion_clear H25.
Inversion_clear H26.
Unfold unwrapMap in H23.
Unfold unwrapMap in H25.
Replace a with x.
Auto.
Apply (msize_eq nat Omega) with m1.
Auto.
Auto.
(* prove register ext *)
Apply mupdext_always.
Apply eqorneq_reg.
Save.


Lemma progress_st : (H:Heap; R:RegFile; a:AP; I:InstrSeq)
      	       	    (rd,rs:Reg; i:int)
      	       	    (WFProgram (H, (R, (a, (iseq (st rd i rs) I))))) ->
		    (EX P' | 
		        (Eval (H, (R, (a, (iseq (st rd i rs) I)))) P')).
Proof.
Intros; Inversion H0. Clear H10. Inversion H9.
Induction H.
Induction HT.
Cut (EX l:? | (mlookup ? ? R rd (wl l))).
Cut (EX w:? | (mlookup ? ? R rs w)).
Intros.
Elim H; Clear H; Intros w; Intros.
Elim H23; Clear H23; Intros l0; Intros.
Cut (EX W:? | (hlookup m l0 (tuple W))); Intros.
Elim H24; Clear H24; Intros W; Intros.
Cut (EX W':? | (updatetuple W i w W')); Intros.
Elim H25; Clear H25; Intros W'; Intros.
Cut (EX h':? | (hupdate m l0 (tuple W') h')/\(mWF ? ? h')); Intros.
Elim H26; Clear H26; Intros h'; Intros.
Elim H26; Clear H26; Intros H26 wfh'; Intros.
Split with ((wfmap ? ? h' wfh'),(R,(a,I))).
Apply (ev_store (wfmap ? ? m m0) (wfmap ? ? h' wfh') R a I m h' m0 wfh'
        rd rs i l0 W W' w).
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.

Apply mupdate_always_wf with (tuple W).
Auto.
Apply eqorneq_nat.
Auto.
Apply lt_imp_updatetuple.
Replace (length W) with (length Ts).
Apply listnth_lt_length with t0.
Auto.

Apply sym_eq.
Apply (tupty_lists_length_eq (wfmap ? ? m m0) (wfmap ? ? m1 m2) R G rd
        l0 Ts V W).
Auto.
Auto.
Auto.
Auto.
Auto.

Apply (can_heap_forms_reg_tuple (wfmap ? ? m m0) (wfmap ? ? m1 m2) R G
        rd l0 Ts V).
Auto.
Auto.
Auto.
Auto.

Apply (can_reg_forms_any (wfmap ? ? m1 m2) R G rs t0).
Auto.
Auto.

Apply (can_reg_forms_tuple (wfmap ? ? m1 m2) R G rd Ts V).
Auto.
Auto.
Save.




(* ------------------------------- *)
Lemma instrseq_subdepth_S
  : (IS',IS:InstrSeq; i:Instr; n:nat)
    (ISubDepth (iseq i IS) IS' n) ->
    (ISubDepth IS IS' (S n)).
Proof.
Induction IS'.
Intros.
Constructor.
Inversion H0.
Constructor.
Apply H with i1.
Auto.
Intros.
Inversion H.
Intros.
Inversion H.
Save.


Lemma instrseq_next_step
  : (H:Heap; i:Instr; I:InstrSeq)
    (EX l | (EX G | (EX T | (EX I' | (EX n |
      	(hlookup (hunwrap H) l (code G T I')) /\
	(ISubDepth (iseq i I) I' n)))))) ->
    (EX l | (EX G | (EX T | (EX I' | (EX n |
      	(hlookup (hunwrap H) l (code G T I')) /\
	(ISubDepth I I' n)))))).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H1.
Split with x.
Split with x0.
Split with x1.
Split with x2.
Split with (S x3).
Split.
Auto.
Apply instrseq_subdepth_S with i.
Auto.
Save.



(* cases of the preservation proof *)
Lemma preserv_add : (H:Heap; R:RegFile; I:InstrSeq; a:AP;
                     rd,rs,rs':Reg; P':Program)
      	       	    (WFProgram (H, (R, (a, (iseq (add rd rs rs') I))))) ->
		    (Eval (H, (R, (a, (iseq (add rd rs rs') I)))) P') ->
		    (WFProgram P').
Intros; Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R' G' a t I).
Auto.
Apply (regfile_upd_wf H HT R R' G G' rd (wi (plus rsval rsval')) intty).
Auto.
Auto.
Constructor.
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (add rd rs rs').
Auto.
Save.



Lemma preserv_addi : (H:Heap; R:RegFile; I:InstrSeq; a:AP;
                     rd,rs:Reg; i:int; P':Program)
      	       	    (WFProgram (H, (R, (a, (iseq (addi rd rs i) I))))) ->
		    (Eval (H, (R, (a, (iseq (addi rd rs i) I)))) P') ->
		    (WFProgram P').
Intros; Inversion H0; Inversion H1; Inversion H10; Apply (program_wf H HT R' G' a t I).
Auto.
Apply (regfile_upd_wf H HT R R' G G' rd (wi (plus rsval i)) intty).
Auto.
Auto.
Constructor.
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (addi rd rs i).
Auto.
Save.



Lemma preserv_alloc : (H:Heap; R:RegFile; I:InstrSeq; a:AP;
      	       	       rd:Reg; V:(list Omega); P':Program)
		      (WFProgram (H, (R, (a, (iseq (alloc rd V) I))))) ->
		      (Eval (H, (R, (a, (iseq (alloc rd V) I)))) P') ->
		      (WFProgram P').
Proof.
Intros.
Inversion H0.
Inversion H1.
Inversion H10.
Induction HT.
Cut (EX ht':? |
       (EXT wfht':? |
            (EX HT':HeapTy |
               HT'=(wfmap label Omega ht' wfht')
               /\(mextend ? ? m a (tupty V (makeUninitTupty V)) ht')))).
Intros.
Inversion_clear H33.
Inversion_clear H34.
Inversion_clear H33.
Inversion_clear H34.
Apply (program_wf H' x1 R' G' a (used n) I).
Rewrite H21; Rewrite H33.
Apply (heap_ext_2 h h' wfh wfh' m x m0 x0 (tuple (makeUninitTup V))
        (tupty V (makeUninitTupty V)) a).
Rewrite <- H20.
Auto.
Apply (wfh_h_size_eq h wfh m m0).
Rewrite <- H20.
Auto.
Rewrite <- H29 in H9.
Inversion H9.
Unfold unwrapMap in H34.
Auto.
Auto.
Auto.
Apply (heap_ext_5 h h' wfh wfh' m x m0 x0 (tuple (makeUninitTup V))
        (tupty V (makeUninitTupty V)) (tupty V (makeUninitTupty V)) a).
Rewrite <- H20.
Auto.
Apply (wfh_h_size_eq h wfh m m0).
Rewrite <- H20.
Auto.
Rewrite <- H29 in H9.
Inversion H9.
Unfold unwrapMap in H34.
Auto.
Auto.
Apply makeUninitWF.
Rewrite H33.
Apply (heap_ext_3 h h' wfh wfh' m x m0 x0 R R' G G' rd a V
        (makeUninitTupty V)).
Rewrite <- H20.
Auto.
Rewrite <- H29 in H9.
Inversion H9.
Unfold unwrapMap in H34.
Apply (wfh_h_size_eq h wfh m m0 a).
Rewrite <- H20.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply heap_ext_3ap with (wfmap nat Omega m m0) V (makeUninitTupty V).
Rewrite H29.
Auto.
Auto.
Auto.
Unfold unwrapMap.
Rewrite H33.
Auto.
Rewrite H33.
Inversion H9.
Unfold unwrapMap in H34.
Apply (heap_ext_4 I h h' wfh wfh' m x m0 x0
        (tupty V (makeUninitTupty V)) a).
Rewrite <- H20.
Auto.
Apply (wfh_h_size_eq h wfh m m0).
Rewrite <- H20.
Auto.
Auto.
Auto.
Auto.
Rewrite <- H29 in H40.
Absurd (used n0)=fresh.
Discriminate.
Auto.
Inversion H11.
Inversion_clear H34.
Inversion_clear H36.
Inversion_clear H34.
Inversion_clear H36.
Inversion_clear H34.
Split with x2.
Split with x3.
Split with x4.
Split with x5.
Split with (S x6).
Split.
Apply mextend_preserv with h a (tuple (makeUninitTup V)).
Rewrite H21.
Auto.
Rewrite H20 in H36.
Auto.
Compare x2 a.
Intros.
Absurd (mindom ? ? h a)/\(mnotindom ? ? h a).
Apply not_min_and_not_wf.
Auto.
Split.
Rewrite <- e.
Apply mlookup_imp_mindom with (code x3 x4 x5).
Rewrite H20 in H36.
Auto.
Inversion H23.
Auto.
Auto.
Apply instrseq_subdepth_S with (alloc rd V).
Auto.
Rewrite <- H29 in H9.
Inversion H9.
Unfold unwrapMap in H33.
Apply heap_ext_1 with l:=a ht:=m wfht:=m0 HT:=HT H:=H.
Auto.
Rewrite H34.
Auto.
Auto.
Save.



Lemma preserv_bgt   : (H:Heap; R:RegFile; I:InstrSeq; a:AP; P':Program)
      	       	      (rs,rt:Reg; l:label)
		      (WFProgram (H, (R, (a, (iseq (bgt rs rt l) I))))) ->
		      (Eval (H, (R, (a, (iseq (bgt rs rt l) I)))) P') ->
		      (WFProgram P').
Proof.
Intros.
Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R G a t I).
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (bgt rs rt l).
Auto.

Apply (program_wf H HT R G' a t I'').
Auto.
Apply (regfile_weakening HT R G G').
Auto.
Auto.
Auto.
Apply (can_heap_forms_code_I HT (code G0 T I'') (codety G' t) G' I'').
Apply (wfheapval_in_heap H HT l (code G0 T I'') (codety G' t)).
Auto.
Auto.
Auto.
Auto.
Replace G0 with G'.
Replace T with t.
Auto.
Apply sym_eq.
Apply (right_and ? ? (wfheap_code_eq H HT l G0 G' T t I'' H7 H24 H34)).
Apply sym_eq.
Apply (left_and ? ? (wfheap_code_eq H HT l G0 G' T t I'' H7 H24 H34)).
Split with l.
Split with G0.
Split with T.
Split with I''.
Split with O.
Split.
Auto.
Constructor.
Save.



Lemma preserv_bump  : (H:Heap; R:RegFile; I:InstrSeq; a:AP; n:nat; P':Program)
      	       	      (WFProgram (H, (R, (a, (iseq (bump n) I))))) ->
		      (Eval (H, (R, (a, (iseq (bump n) I)))) P') ->
		      (WFProgram P').
Proof.
Intros.
Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R G l0 fresh I).
Auto.
Auto.
Induction H.
Unfold unwrapMap in H19.
Move H19 after H24.
Constructor.
Induction HT.
Unfold unwrapMap.
Apply (wfh_ht_size_eq m m0 m1 m2 l0).
Auto.
Auto.
Auto.
Apply instrseq_next_step with (bump n).
Auto.
Save.


Lemma preserv_fold : (H:Heap; R:RegFile; I:InstrSeq; a:AP; P':Program)
      	       	     (rd,rs:Reg; t:Omega)
      	       	     (WFProgram (H, (R, (a, (iseq (fold rd t rs) I))))) ->
		     (Eval (H, (R, (a, (iseq (fold rd t rs) I)))) P') ->
		     (WFProgram P').
Proof.
Intros.
Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R' G' a t0 I).
Auto.
Apply (regfile_upd_wf H HT R R' G G' rd (wfold rsval t) (recty f)).
Auto.
Auto.
Rewrite <- H24.
Apply fold_word_wval with rst.
Auto.
Apply (inv_reg_wval HT R G rs rsval rst).
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.	
Auto.
Apply instrseq_next_step with (fold rd t rs).
Auto.
Save.



Lemma preserv_jmp : (H:Heap; R:RegFile; a:AP; r:Reg; P':Program)
      	       	   (WFProgram (H, (R, (a, (jmp r))))) ->
		   (Eval (H, (R, (a, (jmp r)))) P') ->
		   (WFProgram P').
Proof.
Intros; Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R G' a t I').
Auto.
Apply (regfile_weakening HT R G G' H25 H8).
Auto.
Apply (can_heap_forms_code_I HT (code G0 T I') (codety G' t) G' I').
Apply (wfheapval_in_heap H HT l0 (code G0 T I') (codety G' t)).
Auto.
Auto.
Inversion H8.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w (codety G' t))).
Intros.
Inversion_clear H31.
Inversion_clear H32.
Cut x=(wl l0).
Intros.
Rewrite H32 in H33.
Inversion H33.
Cut t'=(codety G' t).
Intros.
Rewrite H39 in H35.
Auto.
Apply subtype_codety with (codety G' t).
Auto.
Auto.
Apply (mlookup_eq ? ? r x (wl l0) R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H26.
Auto.
Auto.
Replace G0 with G'.
Replace T with t.
Auto.
Apply sym_eq.
Cut (mlookup ? ? (unwrapMap ? ? HT) l0 (codety G' t)).
Intros.
Apply (right_and ? ? (wfheap_code_eq H HT l0 G0 G' T t I' H7 H19 H26)).
Cut (WFWordVal HT (wl l0) (codety G' t)).
Intros.
Inversion H26.
Replace (codety G' t) with t'.
Auto.
Apply subtype_codety with (codety G' t).
Auto.
Auto.
Inversion H8.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w (codety G' t))).
Intros.
Inversion_clear H31.
Inversion_clear H32.
Replace (wl l0) with x.
Auto.
Apply (mlookup_eq ? ? r x (wl l0) R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H26; Auto.
Cut (mlookup ? ? (unwrapMap ? ? HT) l0 (codety G' t)).
Intros.
Apply sym_eq.
Apply (left_and ? ? (wfheap_code_eq H HT l0 G0 G' T t I' H7 H19 H26)).
Cut (WFWordVal HT (wl l0) (codety G' t)); Intros.
Inversion H26.
Replace (codety G' t) with t'; Auto.
Apply subtype_codety with (codety G' t); Auto; Auto.
Cut (EX w:? | (mlookup ? ? R r w)/\(WFWordVal HT w (codety G' t))).
Intros.
Inversion_clear H26.
Inversion_clear H27.
Replace (wl l0) with x.
Auto.
Apply (mlookup_eq ? ? r x (wl l0) R).
Apply eqorneq_reg.
Auto.
Auto.
Inversion H8.
Apply H26; Auto.
Split with l0.
Split with G0.
Split with T.
Split with I'.
Split with O.
Split.
Auto.
Constructor.
Save.



Lemma preserv_jd : (H:Heap; R:RegFile; a:AP; l:label; P':Program)
      	       	   (WFProgram (H, (R, (a, (jd l))))) ->
		   (Eval (H, (R, (a, (jd l)))) P') ->
		   (WFProgram P').
Intros; Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R G' a t I').
Auto.
Apply (regfile_weakening HT R G G' H24 H8).
Auto.
Apply (can_heap_forms_code_I HT (code G0 T I') (codety G' t)
        G' I').
Apply (wfheapval_in_heap H HT l (code G0 T I') (codety G' t)
        H7 H18 H20).
Auto.
Replace G0 with G'.
Replace T with t.
Auto.
Apply sym_eq.
Apply (right_and ? ? (wfheap_code_eq H HT l G0 G' T t I' H7 H18 H20)).
Apply sym_eq.
Apply (left_and ? ? (wfheap_code_eq H HT l G0 G' T t I' H7 H18 H20)).
Split with l.
Split with G0.
Split with T.
Split with I'.
Split with O.
Split.
Auto.
Constructor.
Save.



Lemma wfwordvalinitlist_imp_wfwordval
  : (HT:HeapTy; V:(list WordVal); ol:(list Omega); il:(list initflag))
    (WFWordValinitList HT V ol il) ->
    (i:nat; v:WordVal; t:Omega)
    (ListNth ? V i v) ->
    (ListNth ? ol i t) ->
    (ListNth ? il i true) ->
    (WFWordVal HT v t).
Proof.
Intros HT V ol il H.
Induction H.
Intros.
Inversion H.
Intros.
Induction i0.
Inversion H1.
Inversion H2.
Inversion H3.
Rewrite H7 in H.
Rewrite H10 in H.
Rewrite H4 in H.
Inversion H.
Auto.
Apply HrecH with i0.
Inversion H1.
Auto.
Inversion H2.
Auto.
Inversion H3.
Auto.
Save.

Lemma subtypeblist_init_true
  : (il,il':(list initflag))
    (SubtypeBList il' il) ->
    (i:nat)
    (ListNth ? il i true) ->
    (ListNth ? il' i true).
Proof.
Induction il.
Intros.
Inversion H0.
Intros a l H il' H0.
Inversion H0.
Rewrite <- H3 in H0.
Induction i.
Intros.
Inversion H6.
Rewrite H7 in H4.
Inversion H4.
Constructor.
Intros.
Constructor.
Apply H.
Auto.
Inversion H7.
Auto.
Save.



Lemma preserv_ld : (H:Heap; R:RegFile; I:InstrSeq; a:AP; P':Program)
      	       	   (rd,rs:Reg; i:int)
      	       	   (WFProgram (H, (R, (a, (iseq (ld rd rs i) I))))) ->
		   (Eval (H, (R, (a, (iseq (ld rd rs i) I)))) P') ->
		   (WFProgram P').
Proof.
Intros; Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R' G' a t I).
Auto.
Apply regfile_upd_wf with H R G rd v t0.
Auto.
Auto.
Cut (WFWordVal HT (wl l0) (tupty ol il)).
Intros.
Inversion H37.
Cut (EX il':? | t'=(tupty ol il')).
Intros.
Elim H43; Clear H43; Intros il' H43.
Rewrite H43 in H39.
Cut (WFHeapVal HT (tuple V) (tupty ol il')).
Intros.
Inversion H44.
Apply wfwordvalinitlist_imp_wfwordval with V ol il' i.
Auto.
Auto.
Auto.
Rewrite H43 in H41.
Inversion H41.
Auto.
Apply subtypeblist_init_true with il.
Auto.
Auto.
Inversion H7.
Inversion_clear H45.
Inversion_clear H48.
Inversion_clear H49.
Inversion_clear H50.
Inversion_clear H51.
Inversion_clear H52.
Inversion_clear H53.
Inversion_clear H54.
Clear H55.
Apply H53 with l0.
Auto.
Auto.
Apply subtype_tupty with il.
Auto.
Inversion H8.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal HT w (tupty ol il))).
Intros.
Inversion_clear H42.
Inversion_clear H43.
Replace (wl l0) with x.
Auto.
Apply (mlookup_eq ? ? rs x (wl l0) R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H37.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (ld rd rs i).
Auto.
Save.



Lemma preserv_mov : (H:Heap; R:RegFile; a:AP; I:InstrSeq; P':Program)
      	       	     (rd,rs:Reg)
		     (WFProgram (H,(R,(a,(iseq (mov rd rs) I))))) ->
		     (Eval (H,(R,(a,(iseq (mov rd rs) I)))) P') ->
		     (WFProgram P').
Proof.
Intros; Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R' G' a t I).
Auto.
Apply regfile_upd_wf with H R G rd v t0.
Auto.
Auto.
Inversion H8.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal HT w t0)).
Intros.
Inversion_clear H36.
Inversion_clear H37.
Replace v with x.
Auto.
Apply (mlookup_eq ? ? rs x v R).
Apply eqorneq_reg.
Auto.
Auto.
Apply H31.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (mov rd rs).
Auto.
Save.


Lemma preserv_movi : (H:Heap; R:RegFile; a:AP; I:InstrSeq; P':Program)
      	       	     (rd:Reg; i:int)
		     (WFProgram (H,(R,(a,(iseq (movi rd i) I))))) ->
		     (Eval (H,(R,(a,(iseq (movi rd i) I)))) P') ->
		     (WFProgram P').
Proof.
Intros; Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R' G' a t I).
Auto.
Apply regfile_upd_wf with H R G rd (wi i) intty.
Auto.
Auto.
Constructor.
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (movi rd i).
Auto.
Save.



Lemma preserv_movl : (H:Heap; R:RegFile; a:AP; I:InstrSeq; P':Program)
      	       	     (rd:Reg; l:label)
		     (WFProgram (H,(R,(a,(iseq (movl rd l) I))))) ->
		     (Eval (H,(R,(a,(iseq (movl rd l) I)))) P') ->
		     (WFProgram P').
Proof.
Intros; Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R' G' a t I).
Auto.
Apply regfile_upd_wf with H R G rd (wl l) t'.
Auto.
Auto.
Apply label_wval with t0.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (movl rd l).
Auto.
Save.


Lemma preserv_st_aux2
     	       : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h'))
		 (ht:(Map label Omega); wfht:(mWF ? ? ht))
		 (R:RegFile; G:RegFileTy)
      	       	 (rd,rs:Reg; i:int; l:label; w:WordVal; t:Omega)
		 (V,V':(list WordVal); Ts:(list Omega); 
      	       	  il,V0,V'0:(list initflag))
		 (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
		 (WFRegFile (wfmap ? ? ht wfht) R G) ->
		 (mlookup ? ? R rd (wl l)) ->
		 (mlookup ? ? R rs w) ->
		 (mlookup ? ? h l (tuple V)) ->
		 (updatetuple V i w V') ->
		 (mupdate ? ? h l (tuple V') h') ->
		 (mlookup ? ? G rd (tupty Ts V0)) ->
		 (mlookup ? ? G rs t) ->
		 (ListNth ? Ts i t) ->
		 (updatetupty V0 i V'0) ->
		 (mlookup ? ? ht l (tupty Ts il)) ->
		 (Subtype (tupty Ts il) (tupty Ts V0)) ->
     	       	 (EX ht':(Map nat Omega) |
      	       	    (EXT wfht':(mWF nat Omega ht') |
      	       	    (EX il' |
      	              (mupdate nat Omega ht l (tupty Ts il') ht') /\
      	       	      (updatetupty il i il') /\
      	       	      (Subtype (tupty Ts il') (tupty Ts V'0)) /\ 
              	      (Subtype (tupty Ts il') (tupty Ts il )) /\ 
      	       	      (WFHeapVal (wfmap ? ? ht' wfht') (tuple V') 
      	       	       	       	       	       	       (tupty Ts il'))))).
Proof.
Intros.
Cut (EX il':? | (updatetupty il i il')).
Intros.
Elim H12; Clear H12; Intro il'; Intros.
Cut (EX ht':(Map label Omega) |
       (EXT wfht':(mWF ? ? ht') | (mupdate ? ? ht l (tupty Ts il') ht'))).
Intros.
Elim H13; Clear H13; Intro ht'; Intros.
Elim H13; Clear H13; Intro wfht'; Intros.
Split with ht'.
Split with wfht'.
Split with il'.
Split.
Auto.
Split.
Auto.

Split.
Constructor.
Apply subtype_update with i il V0.
Inversion H11.
Apply subtypeblist_refl.
Auto.
Auto.
Auto.

Split.
Constructor.
Apply updatetupty_subtypeblist with i.
Auto.

Cut (WFHeapVal (wfmap ? ? ht wfht) (tuple V') (tupty Ts il')).
Intros.
Apply (heap_upd_5 h h' wfh wfh' ht ht' wfht wfht' (tupty Ts il')
        (tupty Ts il) l).
Auto.
Auto.

Apply updatetupty_subtype with i.
Auto.
Auto.
Auto.

Apply (updatetuple_wfheapval (wfmap ? ? ht wfht) i V V' w t Ts il il').
Inversion H.
Inversion_clear H15.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H21.
Inversion_clear H22.
Inversion_clear H23.
Unfold unwrapMap in H24.
Inversion_clear H24.
Apply H23 with l.
Auto.
Auto.

Inversion H0.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal (wfmap ? ? ht wfht) w t)).
Intros.
Inversion_clear H19.
Inversion_clear H20.
Replace w with x.
Auto.

Apply (mlookup_eq ? ? rs x w R).
Apply eqorneq_reg.
Auto.
Auto.

Apply H14.
Auto.
Auto.
Auto.
Auto.

Cut (EX ht':(Map label Omega) |
       (mupdate ? ? ht l (tupty Ts il') ht')/\(mWF ? ? ht')).
Intros.
Inversion_clear H13.
Inversion_clear H14.
Split with x.
Split with H15.
Auto.

Apply mupdate_always_wf with (tupty Ts il).
Auto.

Apply eqorneq_nat.
Auto.

Cut (lt i (length il)).
Intros.
Apply lt_imp_updatetupty.
Auto.

Cut (WFWordValinitList (wfmap ? ? ht wfht) V Ts il).
Intros.
Cut (length il)=(length Ts).
Intros.
Rewrite H13.
Apply listnth_lt_length with t.
Auto.

Apply wfwordvalinitlist_length_eq_il_tl with (wfmap nat Omega ht wfht) V.
Auto.

Inversion H.
Inversion_clear H13.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H21.
Unfold unwrapMap in H22.
Cut (WFHeapVal (wfmap nat Omega ht wfht) (tuple V) (tupty Ts il)).
Intros.
Inversion H21.
Auto.

Inversion H22.
Apply H21 with l.
Auto.

Auto.
Save.



Lemma preserv_st : (H:Heap; R:RegFile; I:InstrSeq; a:AP;	
      	       	    rd,rs:Reg; i:int; P':Program)
		   (WFProgram (H, (R, (a, (iseq (st rd i rs) I))))) ->
		   (Eval (H, (R, (a, (iseq (st rd i rs) I)))) P') ->
		   (WFProgram P').
Proof.
Intros; Inversion H0; Inversion H1; Inversion H10.
Induction HT.
Cut (EX il:? |
       (htlookup m l0 (tupty Ts il))
       /\(Subtype (tupty Ts il) (tupty Ts V0))); Intros.
Elim H41; Clear H41; Intros il; Intros.
Inversion_clear H41.
Cut (EX ht':(Map nat Omega) |
       (EXT wfht':(mWF nat Omega ht') |
            (EX il':? |
               (mupdate nat Omega m l0 (tupty Ts il') ht')
               /\(updatetupty il i il')
                 /\(Subtype (tupty Ts il') (tupty Ts V'0))
                   /\(Subtype (tupty Ts il') (tupty Ts il))
                     /\(WFHeapVal (wfmap ? ? ht' wfht') (tuple V')
                         (tupty Ts il'))))); Intros.
Elim H41; Intro ht'; Clear H41; Intros.
Elim H41; Intro wfht'; Clear H41; Intros.
Elim H41; Intro il'; Clear H41; Intros.
Inversion_clear H41.
Inversion_clear H45.
Inversion_clear H46.
Inversion_clear H47.
Apply (program_wf H' (wfmap ? ? ht' wfht') R G' a t I).
Rewrite H22.
Apply (heap_upd_2 h h' wfh wfh' m ht' m0 wfht' (tupty Ts il')
        (tupty Ts il) l0 (tuple V')).
Rewrite <- H20.
Auto.

Auto.
Auto.
Auto.
Auto.
Auto.

Cut (WFRegFile (wfmap ? ? ht' wfht') R G).
Intros.
Apply (regfile_upd_wf H' (wfmap nat Omega ht' wfht') R R G G' rd (
       wl l0) (tupty Ts V'0)).
Rewrite H22.
Apply (heap_upd_2 h h' wfh wfh' m ht' m0 wfht' (tupty Ts il')
        (tupty Ts il) l0 (tuple V')).
Rewrite <- H20.
Auto.

Auto.
Auto.
Auto.
Auto.
Auto.
Auto.

Apply label_wval with (tupty Ts il').
Unfold unwrapMap.
Apply mupdate_imp_mlookup with m.
Apply eqorneq_nat.
Auto.
Auto.

Cut (mupdate ? ? R rd (wl l0) R).
Intros.
Apply mupdate_imp_mupdext.
Apply eqorneq_reg.
Auto.

Apply mupdate_refl.
Apply eqorneq_reg.
Auto.

Apply mupdate_imp_mupdext.
Apply eqorneq_reg.
Auto.

Apply (heap_upd_3 h h' wfh wfh' m ht' m0 wfht' (tupty Ts il')
        (tupty Ts il) l0).
Rewrite <- H20.
Auto.
Auto.
Auto.
Auto.
Auto.

Apply (heap_upd_3ap h h' wfh wfh' m ht' m0 wfht' (tupty Ts il')
        (tupty Ts il) l0).
Rewrite <- H20.
Auto.
Auto.
Auto.
Auto.
Auto.

Apply (heap_upd_4 I h h' wfh wfh' m ht' m0 wfht' (tupty Ts il')
        (tupty Ts il) l0).
Rewrite <- H20.
Auto.
Auto.
Auto.
Auto.
Auto.

Inversion H11.
Inversion_clear H47.
Inversion_clear H49.
Inversion_clear H47.
Inversion_clear H49.
Inversion_clear H47.
Split with x.
Split with x0.
Split with x1.
Split with x2.
Split with (S x3).
Split.
Apply mupdate_preserv with h l0 (tuple V').
Rewrite H22.
Auto.
Rewrite H20 in H49.
Auto.
Compare x l0.
Intros.
Absurd (code x0 x1 x2)=(tuple V).
Discriminate.
Apply (mlookup_eq ? ? x (code x0 x1 x2) (tuple V) h).
Apply eqorneq_nat.
Rewrite H20 in H49.
Auto.
Rewrite e.
Auto.
Auto.
Apply instrseq_subdepth_S with (st rd i rs).
Auto.

Apply (preserv_st_aux2 h h' wfh wfh' m m0 R G rd rs i l0 w t0 V V' Ts
        il V0).
Rewrite <- H20.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.

Cut (EX s:? | (mlookup ? ? m l0 s)/\(Subtype s (tupty Ts V0))).
Intros.
Inversion_clear H41.
Inversion_clear H42.
Inversion H43.
Split with V0.
Split.
Rewrite H44 in H41.
Auto.

Constructor.

Split with il1.
Split.
Rewrite H42 in H44.
Rewrite <- H44 in H41.
Auto.

Rewrite H42 in H44.
Rewrite <- H44 in H43.
Auto.

Replace m with (unwrapMap ? ? (wfmap ? ? m m0)).
Apply inv_label_lookup_subtype with R G rd.
Auto.
Auto.
Auto.

Unfold unwrapMap.
Auto.
Save.



Lemma inv_wval_fold : (HT:HeapTy; v:WordVal; f:OmegaR;
      	       	       unf,t:Omega) 
      	       	      (WFWordVal HT (wfold v t) (recty f)) -> 
		      (RUnlift (RUnfold f))=unf ->
      	       	      (WFWordVal HT v unf).
Proof.
Intros.
Inversion H.
Replace unf with t'.
Auto.
Rewrite <- H0.
Rewrite <- H5.
Auto.
Save.



Lemma preserv_unfold : (H:Heap; R:RegFile; I:InstrSeq; a:AP; P':Program)
      	       	     (rd,rs:Reg)
      	       	     (WFProgram (H, (R, (a, (iseq (unfold rd rs) I))))) ->
		     (Eval (H, (R, (a, (iseq (unfold rd rs) I)))) P') ->
		     (WFProgram P').
Proof.
Intros.
Inversion H0; Inversion H1; Inversion H10.
Apply (program_wf H HT R' G' a t I).
Auto.
Apply (regfile_upd_wf H HT R R' G G' rd v unf).
Auto.
Auto.
Cut (WFWordVal HT (wfold v t0) (recty f)).
Intros.
Apply inv_wval_fold with f t0; Auto.
Inversion H8.
Cut (EX w:? | (mlookup ? ? R rs w)/\(WFWordVal HT w (recty f))).
Intros.
Inversion_clear H38.
Inversion_clear H39.
Replace (wfold v t0) with x.
Auto.
Apply (mlookup_eq ? ? rs x (wfold v t0) R eqorneq_reg).
Auto.
Auto.
Apply H33.
Rewrite <- H26.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply instrseq_next_step with (unfold rd rs).
Auto.
Save.

