(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)
(* Thanks to Yves Bertot, Clement Renard at Inria for help with Coq *)

(***************************************************************************)
(* Definition of mapping structure and related utility functions *)
(***************************************************************************)

Require natutil.

(***************************************************************************)

(* Map structure ------------------------ *)
Inductive Map [A,B:Set] : Set
  := mempty : (Map A B)
  |  melem  : A -> B -> (Map A B) -> (Map A B).

Definition is_mempty
  := [A,B:Set; m:(Map A B)] (Cases m of mempty => True | _ => False end).


(* size of mapping - probably only makes sense if mapping is well-formed *)
Inductive msize [A,B:Set] : (Map A B) -> nat -> Prop
  := msize0 : (msize A B (mempty A B) O)
  |  msize1 : (a:A; b:B; m:(Map A B); n:nat) 
              (msize A B m n) -> (msize A B (melem A B a b m) (S n)).


(* element is in the domain of the mapping *)
Inductive mindom [A,B:Set] : (Map A B) -> A -> Prop
  := minfound : (a:A; b:B; m:(Map A B)) (mindom A B (melem A B a b m) a)
  |  minnext  : (a,a':A; b:B; m:(Map A B)) (~a=a') ->
      	       	(mindom A B m a') -> (mindom A B (melem A B a b m) a').


(* an element is not in the domain of the map *)
Inductive mnotindom [A,B:Set] : (Map A B) -> A -> Prop
  :=  mnotindom0 : (a':A) (mnotindom A B (mempty A B) a')
  |   mnotindom1 : (a,a':A; b:B; m:(Map A B)) (~a=a') ->
		   (mnotindom A B m a') ->
		   (mnotindom A B (melem A B a b m) a').


(* well-formedness means all distinct elements unequal *)
Inductive mWF [A,B:Set] : (Map A B) -> Prop
  := mwf0 : (mWF A B (mempty A B))
  |  mwf1 : (a:A; b:B; m:(Map A B))
      	    (mnotindom A B m a) ->
	    (mWF A B m) ->
	    (mWF A B (melem A B a b m)).


(* a well-formed map wrapper *)
Inductive WFMap [A,B:Set] : Set
  := wfmap : (m:(Map A B)) (mWF A B m) -> (WFMap A B).

(* utilities for working with well-formed maps *)
Definition unwrapMap : (A,B:Set) (WFMap A B) -> (Map A B)
  := [A,B][m] Cases m of (wfmap m' mwf) => m' end.
Definition emptyWFMap : (A,B:Set) (WFMap A B)
  := [A,B] (wfmap ? ? (mempty ? ?) (mwf0 ? ?)).


(* Used to prove that a pair is in this mapping *)
Inductive mlookup [A,B:Set] : (Map A B) -> A -> B -> Prop 
  := mlookfound : (a:A; b:B; m:(Map A B)) (mlookup A B (melem A B a b m) a b)
  |  mlooknext  : (a,a':A; b,b':B; m:(Map A B)) 
      	       	  (mlookup A B m a' b') -> ~a=a' ->
      	       	  (mlookup A B (melem A B a b m) a' b').


(* updates the mapping of an element that is already in the domain *)
Inductive mupdate [A,B:Set] : (Map A B) -> A -> B -> (Map A B) -> Prop
  := mupdfound : (a:A; b,b':B; m:(Map A B))
      	       	 (mupdate A B (melem A B a b' m) a b (melem A B a b m))
  |  mupdnext  : (a,a':A; b,b':B; m,m':(Map A B))
      	       	 (mupdate A B m a b m') ->
      	       	 (~a=a') ->
		 (mupdate A B (melem A B a' b' m) a b (melem A B a' b' m')).


(* extends the mapping, adding a new element to the domain *)
Inductive mextend [A,B:Set] : (Map A B) -> A -> B -> (Map A B) -> Prop
  := mext : (a:A; b:B; m:(Map A B))
   	    (mnotindom A B m a) ->
	    (mextend A B m a b (melem A B a b m)).


(* sometimes we don't care exactly what happens to the mapping *)
Inductive mupdext [A,B:Set] : (Map A B) -> A -> B -> (Map A B) -> Prop
  := mupdext0 : (a:A; b:B)
      	       	(mupdext A B (mempty A B) a b (melem A B a b (mempty A B)))
  |  mupdext1 : (a,a':A; b,b':B; m,m':(Map A B))
      	       	(mupdext A B m a b m') ->
		(~a=a') ->
		(mupdext A B (melem A B a' b' m) a b (melem A B a' b' m'))
  |  mupdext2 : (a:A; b,b':B; m:(Map A B))
      	       	(mupdext A B (melem A B a b' m) a b (melem A B a b m)).



(* -------------------------------------------------------------------- *)
(* THEOREMS ABOUT MAPS *)
(*
msize_always
msize_eq
eq_map_wf
mwf_next
mwf_elem_neq
mwf_anyelem_neq
mindom_next
mindom_imp_mlookup
mindom_imp_not_mnotindom
min_or_not_nat_wf
not_min_and_not_wf
mnotindom_imp_not_mindom 
mnotindom_imp_mextend 
mnotindom_mextend_imp_mnotindom
mindom_mnotindom_neq
mlookup_next 
mlookup_imp_mindom 
mlookup_eq
mlookup_imp_mupdate
mupdate_refl
mupdate_imp_mlookup_ex
mupdate_imp_mlookup
mupdate_imp_mupdext
mupdate_preserv
mupdate_preserv_rev
mupdate_preserv_rev_ex
mupdate_size_same
mupdate_mnotindom_preserv
mupdate_always_wf
mupdate_same_eq
mextend_imp_mlookup
mextend_wf 
mextend_preserv
mextend_preserv_rev
mextend_size_Sn
mupdext_always
mupdext_preserv
mupdext_preserv_rev
mupdext_preserv_rev_mindom
mupdext_imp_mlookup
mupdext_same_eq

ext_mupdext_two_preserv
ext_map_oneoff_updext_same_oneoff
ext_map_eq_updext_eq
ext_map_lookup_eq



*)




(* a map always has a size *)
Lemma msize_always : (A,B:Set; m:(Map A B))
      	       	     (EX l | (msize A B m l)).
Proof.
Intros.
Induction m.
Split with O.
Constructor.

Inversion_clear Hrecm.
Split with (S x).
Constructor.
Auto.
Save.


(* a map has a unique size *)
Lemma msize_eq : (A,B:Set; m:(Map A B); a,a':nat)
      	       	 (msize A B m a) ->
		 (msize A B m a') ->
		 a=a'.
Proof.
Induction m.
Intros.
Inversion H.
Inversion H0.
Auto.
Intros.
Inversion H0.
Inversion H1.
Cut n=n0.
Intros.
Auto.
Apply (H n n0 H6 H11).
Save.


(* well-formedness of equal maps *)
Lemma eq_map_wf
  : (A,B:Set; m,m':(Map A B); a:A; b:B)
    (mWF A B m) ->
    m=(melem A B a b m') ->
    (mWF A B (melem A B a b m')).
Proof.
Induction m.
Intros.
Absurd (mempty A B)=(melem A B a b m').
Discriminate; Auto.
Auto.
Intros.
Constructor.
Inversion H0.
Replace m' with m0.
Replace a0 with a.
Auto.
Inversion H1.
Auto.
Inversion H1.
Auto.
Inversion H0.
Replace m' with m0.
Auto.
Injection H1; Auto.
Save.



(* subsets of a well-formed map are well-formed *)
Lemma mwf_next : (A,B:Set; a:A; b:B; m:(Map A B))
      	       	 (mWF ? ? (melem ? ? a b m)) -> (mWF ? ? m).
Intros.
Inversion H.
Auto.
Defined.


(* subsequent elements in a well-formed map are not equal *)
Lemma mwf_elem_neq : (A,B:Set; a,a':A; b,b':B; m:(Map A B))
      	       	     (mWF A B (melem A B a b (melem A B a' b' m))) ->
		     (~a=a').
Proof.
Intros.
Inversion H.
Inversion H2.
Auto.
Save.


(* any two distinct elements in a well-formed map are not equal *)
Lemma mwf_anyelem_neq : (A,B:Set; a,a':A; b:B; m:(Map A B))
      	       	       	(mWF A B (melem A B a b m)) ->
			(mindom A B m a') ->
			(~a=a').
Proof.
Intros.
Induction m.
Inversion H0.
Inversion H0.
Cut ~a=a0.
Intros.
Rewrite <- H1.
Auto.
Apply (mwf_elem_neq A B a a0 b b0 m H).
Cut (mWF A B (melem A B a b m)).
Intros.
Apply (Hrecm H7 H6).
Clear H1.
Clear a1.
Clear H3.
Clear a'0.
Clear H4.
Clear m0.
Clear H2.
Clear b1.
Apply mwf1.
Inversion H.
Inversion H3.
Auto.
Cut (mWF A B (melem A B a0 b0 m)).
Intros.
Apply (mwf_next A B a0 b0 m H1).
Apply (mwf_next A B a b (melem A B a0 b0 m) H).
Save.


(* if a' is in the domain of the map and a not-equal a' then a' must be
   in the domain of the remainder *)
Lemma mindom_next 
      : (A,B:Set; m:(Map A B); a,a':A; b:B)
        (mindom A B (melem A B a b m) a') ->
	(~a=a') ->
	(mindom A B m a').
Proof.
Intros.
Inversion H.
Tauto.
Auto.
Save.


(* if a is in the domain of the map, it must be mapped to something *)
Lemma mindom_imp_mlookup : (A,B:Set; m:(Map A B); a:A)
      	       	       	   (mindom A B m a) ->
			   (EX b | (mlookup A B m a b)).
Proof.
Intros.
Induction m.
Inversion H.
Inversion H.
Split with b.
Constructor.
Cut (EX b:B | (mlookup A B m a b)).
Intros.
Inversion_clear H6.
Split with x.
Constructor.
Auto.
Auto.
Apply (Hrecm H5).
Save.


(* if l is in the domain, it can't not be in the domain *)
Lemma mindom_imp_not_mnotindom 
  : (B:Set; m:(Map nat B); l:nat)
    (mWF ? ? m) ->
    (mindom nat B m l) -> ~(mnotindom nat B m l).
Proof.
Intros.
Induction m.
Inversion H0.
Cut l=a\/~l=a.
Intros.
Inversion_clear H1.
Unfold not.
Intros.
Rewrite H2 in H1.
Inversion H1.
Absurd ~a=a.
Auto.
Auto.
Inversion H0.
Rewrite H1 in H2.
Absurd ~l=l.
Auto.
Auto.
Unfold not.
Intros.
Inversion H8.
Apply Hrecm.
Inversion H.
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Save.


(* an element is either in or not in a (well-formed) map *)
Lemma min_or_not_nat_wf
  : (B:Set; m:(Map nat B); l:nat)
    (mWF ? ? m) ->
    (mindom nat B m l) \/ (mnotindom nat B m l).
Proof.
Intros.
Induction m.
Right.
Constructor.
Inversion H.
Cut (mindom nat B m l)\/(mnotindom nat B m l).
Intros.
Inversion H5.
Left.
Apply minnext.
Apply (mwf_anyelem_neq nat B a l b m H H6).
Auto.
Cut a=l\/~a=l.
Intros.
Inversion_clear H7.
Left.
Rewrite H8.
Constructor.
Right.
Apply mnotindom1.
Auto.
Auto.
Apply eqorneq_nat.
Apply (Hrecm H4).
Save.


(* an element is not both in and not in a (well-formed) map *)
Lemma not_min_and_not_wf
  : (B:Set; m:(Map nat B); l:nat)
    (mWF ? ? m) ->
    ~((mindom nat B m l) /\ (mnotindom nat B m l)).
Intros.
Induction m.
Unfold not.
Intros.
Inversion H0.
Inversion H1.
Unfold not.
Intros.
Inversion H0.
Cut ~(mnotindom nat B (melem nat B a b m) l).
Intros.
Apply H3.
Auto.
Apply mindom_imp_not_mnotindom.
Auto.
Auto.
Save.


(* an element not in the map implies it is not in the map (duh...) *)
Lemma mnotindom_imp_not_mindom 
  : (B:Set; m:(Map nat B); l:nat)
    (mWF ? ? m) ->
    (mnotindom nat B m l) -> ~(mindom nat B m l).
Proof.
Intros.
Induction m.
Unfold not.
Intros.
Inversion H1.
Cut l=a\/~l=a.
Intros; Inversion_clear H1.
Unfold not.
Intros.
Rewrite H2 in H0.
Inversion H0.
Absurd ~a=a.
Auto.
Auto.
Unfold not.
Intros.
Inversion H1.
Rewrite H3 in H2.
Absurd ~l=l.
Auto.
Auto.
Inversion H0.
Inversion H.
Apply Hrecm.
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Save.


(* an element not in the map means the map can be extended with it *)
Lemma mnotindom_imp_mextend 
  : (A,B:Set; m:(Map A B); a:A; b:B)
    (mnotindom ? ? m a) ->
    (EX m' | (mextend ? ? m a b m')).
Proof.
Intros.
Split with (melem ? ? a b m).
Constructor.
Auto.
Save.


(* extending the map with a different element means the element is still
   not in the map *)
Lemma mnotindom_mextend_imp_mnotindom
  : (A,B:Set; m,m':(Map A B); a,a':A; b:B)
    (mnotindom A B m a') ->
    (mextend A B m a b m') ->
    (~a=a') ->
    (mnotindom A B m' a').
Proof.
Intros.
Inversion H0.
Constructor.
Auto.
Auto.
Save.



Lemma mindom_mnotindom_neq
  : (A,B:Set; m:(Map A B); wfm:(mWF ? ? m); a,b:A)
    ((x,x':A) x=x' \/ ~x=x') ->
    (mindom ? ? m a) ->
    (mnotindom ? ? m b) ->
    ~a=b.
Proof.
Induction m.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H3.
Rewrite H4 in H2.
Inversion H2.
Auto.
Apply H.
Inversion wfm.
Auto.
Auto.
Inversion H1.
Rewrite H3 in H4.
Absurd ~a0=a0; Auto.
Auto.
Inversion H2.
Auto.
Auto.
Save.



(* if the current element is not equal, then must be in the remaining *)
Lemma mlookup_next 
      : (A,B:Set; m:(Map A B); a,a':A; b,b':B)
        (mlookup A B (melem A B a b m) a' b') ->
	(~a=a') ->
	(mlookup A B m a' b').
Intros.
Inversion H.
Tauto.
Auto.
Save.


(* if you can look up an element, it must be in the domain *)
Lemma mlookup_imp_mindom : (A,B:Set; m:(Map A B); a:A; b:B)
      	       	       (mlookup A B m a b) -> (mindom A B m a).
Proof.
Intros.
Elim H.
Intros.
Constructor.
Intros.
Constructor.
Auto.
Auto.
Defined.


(* lookup is a function (dependent on the charctrstcs of the domain type) *)
Lemma mlookup_eq : (A,B:Set; a:A; b,b':B; m:(Map A B))
      	       	   ( (x,x':A)x=x'\/~x=x' ) ->
		   (mlookup A B m a b) ->
		   (mlookup A B m a b') ->
		   (b=b').
Proof.
Intros A B a b b' m P H H0.
Induction m.
Inversion H.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H1.
Rewrite <- H2 in H.
Rewrite <- H2 in H0.
Cut b=b0/\b'=b0.
Intros.
Inversion H1.
Rewrite H3.
Rewrite H4.
Auto.
Inversion H.
Inversion H0.
Split.
Auto.
Rewrite <- H11.
Auto.
Split.
Auto.
Absurd ~a=a.
Auto.
Auto.
Absurd ~a=a.
Auto.
Auto.
Cut (mlookup A B m a b').
Cut (mlookup A B m a b).
Auto.
Apply mlookup_next with a:=a0 b:=b0.
Auto.
Auto.
Apply mlookup_next with a:=a0 b:=b0.
Auto.
Auto.
Apply P.
Save.


(* mlookup is injective *)
Lemma mlookup_neq_dom_neq
      	       	     : (A,B:Set; m:(Map A B); a,a':A; b,b':B)
		       ( (x,x':A) x=x' \/ ~x=x' ) ->
		       (mlookup ? ? m a b) ->
		       (mlookup ? ? m a' b') ->
		       ~b=b' ->
		       ~a=a'.
Proof.
Induction m.
Intros.
Inversion H0.
Intros.
Cut a0=a'\/~a0=a'.
Intros.
Inversion_clear H4.
Rewrite H5 in H1.
Move H2 after H1.
Inversion H1.
Inversion H2.
Rewrite H14 in H9.
Rewrite H9 in H3.
Absurd ~b0=b0.
Auto.
Auto.
Rewrite H4 in H16.
Absurd ~a'=a'.
Auto.
Auto.
Inversion H2.
Rewrite H12 in H11.
Absurd ~a'=a'.
Auto.
Auto.
Rewrite H5.
Apply H with b0 b'.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Save.


(* if an element is in the map, it can be updated *)
Lemma mlookup_imp_mupdate : (A,B:Set; m:(Map A B); a:A; b,b':B)
      	       	       	    (mlookup A B m a b) ->
			    (EX m' | (mupdate A B m a b' m')).
Proof.
Intros.
Induction m.
Inversion H.
Inversion H.
Split with (melem A B a b' m).
Apply mupdfound.
Cut (mlookup A B m a b).
Intro.
Cut (EX m':(Map A B) | (mupdate A B m a b' m')).
Intros.
Elim H8.
Intros.
Split with (melem A B a0 b0 x).
Apply mupdnext.
Auto.
Auto.
Apply (Hrecm H7).
Auto.
Save.


(* updating with the same thing maintains the same map *)
Lemma mupdate_refl
  : (A,B:Set; m:(Map A B); a:A; b:B)
    ( (x,x':A)x=x' \/ ~x=x' ) ->
    (mlookup A B m a b) ->
    (mupdate A B m a b m).
Proof.
Induction m.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H2.
Rewrite H3.
Rewrite H3 in H1.
Inversion H1.
Constructor.
Absurd ~a0=a0.
Auto.
Auto.
Inversion H1.
Rewrite H2 in H3.
Absurd ~a0=a0.
Auto.
Auto.
Apply mupdnext.
Apply H.
Auto.
Auto.
Auto.
Auto.
Save.



(* if an element has been updated, it must have had a previous value *)
Lemma mupdate_imp_mlookup_ex : (A,B:Set; m,m':(Map A B); a:A; b:B)
      	       	       	    (P : ((x,x':A) x=x' \/ ~x=x'))
      	       	       	    (mupdate A B m a b m') ->
			    (EX b' | (mlookup A B m a b')).
Proof.
Induction m.
Induction m'.
Intros.
Inversion H.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H1.
Split with b.
Rewrite H2.
Constructor.
Cut (EX b':B | (mlookup A B m0 a0 b')).
Intros.
Inversion_clear H1.
Split with x.
Apply mlooknext.
Auto.
Auto.
Inversion H0.
Rewrite H1 in H2.
Absurd ~a0=a0.
Auto.
Auto.
Apply (H m'0 a0 b0).
Auto.
Auto.
Auto.
Save.


(* if an element is updated, it must be lookup-able exactly the same way
   in the new map *)
Lemma mupdate_imp_mlookup : (A,B:Set; m,m':(Map A B); a:A; b:B)
      	       	       	    (P : ((x,x':A) x=x' \/ ~x=x'))
      	       	       	    (mupdate A B m a b m') ->
			    (mlookup A B m' a b).
Proof.
Induction m.
Induction m'.
Intros.
Inversion H.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H1.
Rewrite H2 in H0.
Inversion H0.
Apply mlookfound.
Absurd ~a0=a0.
Auto.
Auto.
Inversion H0.
Rewrite H1 in H2.
Absurd ~a0=a0.
Auto.
Auto.
Apply mlooknext.
Apply (H m'0 a0 b0).
Auto.
Auto.
Auto.
Auto.
Save.



(* mupdate implies mupdext *)
Lemma mupdate_imp_mupdext : (A,B:Set; m,m':(Map A B); a:A; b:B)
      	       	       	    (P:((x,x':A) x=x' \/ ~x=x'))
      	       	       	    (mupdate A B m a b m') ->
			    (mupdext A B m a b m').
Proof.
Induction m.
Induction m'.
Intros.
Inversion H.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H1.
Rewrite H2 in H0.
Inversion H0.
Rewrite H2.
Constructor.
Absurd ~a0=a0.
Auto.
Auto.
Inversion H0.
Rewrite H1 in H2.
Absurd ~a0=a0.
Auto.
Auto.
Apply mupdext1.
Auto.
Auto.
Auto.
Save.



Lemma mupdate_preserv
  : (A,B:Set; m,m':(Map A B); a,a':A; b,b':B)
    (mupdate A B m a' b' m') ->
    (mlookup A B m a b) ->
    ~a=a' ->
    (mlookup A B m' a b).
Proof.
Intros.
Induction H.
Apply mlooknext.
Inversion H0.
Rewrite H5 in H1.
Absurd ~a=a.
Auto.
Auto.
Auto.
Auto.
Inversion H0.
Constructor.
Apply mlooknext.
Apply HrecH.
Auto.
Auto.
Auto.
Save.


Lemma mupdate_preserv_rev
  : (A,B:Set; m,m':(Map A B); a,a':A; b,b':B)
    (P: ((x,x':A)x=x' \/ ~x=x') )
    (mupdate A B m a' b' m') ->
    (mlookup A B m' a b) ->
    ~a=a' ->
    (mlookup A B m a b).
Proof.
Intros.
Induction H.
Inversion H0.
Rewrite H5 in H1.
Absurd ~a=a.
Auto.
Auto.
Apply mlooknext.
Auto.
Auto.
Cut a'=a\/~a'=a.
Intros.
Inversion_clear H3.
Rewrite H4 in H0.
Inversion H0.
Replace a' with a.
Constructor.
Absurd ~a=a.
Auto.
Auto.
Apply mlooknext.
Inversion H0.
Rewrite H3 in H4.
Absurd ~a=a.
Auto.
Auto.
Apply HrecH.
Auto.
Auto.
Auto.
Auto.
Save.


Lemma mupdate_preserv_rev_ex
  : (A,B:Set; m,m':(Map A B); a,a':A; b,b':B)
    (P: ((x,x':A)x=x' \/ ~x=x') )
    (mupdate A B m a' b' m') ->
    (mlookup A B m' a b) ->
    (EX c | (mlookup A B m a c)).
Proof.
Induction m.
Induction m'.
Intros; Inversion H.
Intros; Inversion H0.
Induction m'.
Intros.
Inversion H0.
Intros.
Cut a=a1\/~a=a1.
Intros.
Inversion_clear H3.
Replace a with a1.
Split with b.
Constructor.
Cut (EX c:? | (mlookup ? ? m0 a1 c)).
Intros.
Inversion_clear H3.
Split with x.
Apply mlooknext.
Auto.
Auto.
Cut a=a'\/~a=a'.
Intros.
Inversion_clear H3.
Rewrite H5 in H1.
Inversion H1.
Rewrite <- H11 in H2.
Rewrite <- H5 in H2.
Move H4 after H2.
Inversion H2.
Rewrite H10 in H4.
Absurd ~a1=a1.
Auto.
Auto.
Split with b1.
Auto.
Absurd ~a0=a0.
Auto.
Auto.
Inversion H1.
Rewrite H3 in H5.
Absurd ~a'=a'.
Auto.
Auto.
Move H after H14.
Inversion H2.
Rewrite <- H11 in H15.
Rewrite H15 in H4.
Absurd ~a1=a1.
Auto.
Auto.
Apply (H m1 a1 a' b1 b').
Auto.
Auto.
Auto.
Auto.
Apply P.
Save.


Lemma mupdate_size_same : (A,B:Set; m,m':(Map A B); a:A; b:B; n:nat)
      	       	       	  (P:((x,x':A)x=x'\/~x=x'))
      	       	       	  (mupdate A B m a b m') ->
			  (msize A B m n) ->
			  (msize A B m' n).
Proof.
Induction m.
Intros.
Inversion H.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H2.
Rewrite H3 in H0.
Inversion H0.
Inversion H1.
Apply msize1.
Auto.
Absurd ~a0=a0.
Auto.
Auto.
Inversion H0.
Rewrite H2 in H3.
Absurd ~a0=a0.
Auto.
Auto.
Move H after H10.
Inversion H1.
Apply msize1.
Apply (H m'0 a0 b0 n0).
Auto.
Auto.
Auto.
Auto.
Save.


Lemma mupdate_size_same_rev : (A,B:Set; m,m':(Map A B); a:A; b:B; n:nat)
      	       	       	  (P:((x,x':A)x=x'\/~x=x'))
      	       	       	  (mupdate A B m a b m') ->
			  (msize A B m' n) ->
			  (msize A B m n).
Proof.
Induction m.
Intros.
Inversion H.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H2.
Rewrite H3 in H0.
Inversion H0.
Rewrite <- H8 in H1.
Inversion H1.
Constructor.
Auto.
Absurd ~a0=a0; Auto.
Inversion H0.
Rewrite <- H8 in H1.
Inversion H1.
Constructor.
Auto.
Rewrite <- H8 in H1.
Inversion H1.
Constructor.
Apply (H m'0 a0 b0 n0).
Auto.
Auto.
Auto.
Auto.
Save.



Lemma mupdate_mnotindom_preserv
      	       	       	: (A,B:Set; m,m':(Map A B); a,a':A; b:B)
     	       	       	  ( (x,x':A)x=x'\/~x=x' ) ->
			  (mupdate A B m a b m') ->
			  (mnotindom A B m a') ->
			  (mnotindom A B m' a').
Proof.
Induction m.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H3.
Rewrite H4 in H1.
Inversion H1.
Rewrite <- H4.
Inversion H2.
Apply mnotindom1.
Auto.
Auto.
Absurd ~a0=a0.
Auto.
Auto.
Inversion H1.
Rewrite H3 in H4.
Absurd ~a0=a0.
Auto.
Auto.
Inversion H2.
Apply mnotindom1.
Auto.
Apply H with a0 b0.
Auto.
Auto.
Auto.
Auto.
Save.


(* updating an existing mapping can always be done and maintains 
   wellformed-ness *)
Lemma mupdate_always_wf : (A,B:Set; m:(Map A B); wfm:(mWF ? ? m); a:A; b,b':B)
     	       	       	  ( (x,x':A)x=x'\/~x=x' ) ->
      	       	       	  (mlookup A B m a b') ->
			  (EX m' |
			     (mupdate A B m a b m') /\
			     (mWF A B m')).
Proof.
Induction m.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H2.
Rewrite H3.
Split with (melem A B a0 b0 m0).
Split.
Constructor.
Rewrite <- H3.
Inversion wfm.
Apply mwf1.
Auto.
Auto.
Cut (EX m':? | (mupdate A B m0 a0 b0 m')/\(mWF A B m')).
Intros.
Elim H2.
Intro m''.
Intros.
Inversion_clear H4.
Clear H2.
Split with (melem A B a b m'').
Split.
Apply mupdnext.
Auto.
Auto.
Apply mwf1.
Inversion wfm.
Apply mupdate_mnotindom_preserv with m0 a0 b0.
Auto.
Auto.
Auto.
Auto.
Apply H with b'.
Inversion wfm.
Auto.
Auto.
Inversion H1.
Rewrite H2 in H3.
Absurd ~a0=a0.
Auto.
Auto.
Auto.
Auto.
Save.


(* updating the map in the same way produces equivalent map *)
Lemma mupdate_same_eq : (A,B:Set; m,m',m'':(Map A B); a:A; b:B)
      	       	       	( (x,x':A) x=x' \/ ~x=x' ) ->
      	       	       	(mupdate A B m a b m') ->
			(mupdate A B m a b m'') ->
			m'=m''.
Proof.
Induction m.
Intros.
Inversion H0.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H3.
Rewrite H4 in H1.
Rewrite H4 in H2.
Inversion H1.
Inversion H2.
Auto.
Absurd ~a0=a0.
Auto.
Auto.
Auto.
Absurd ~a0=a0; Auto.
Inversion H1.
Rewrite H3 in H4.
Absurd ~a0=a0; Auto.
Inversion H2.
Rewrite H12 in H4.
Absurd ~a0=a0; Auto.
Replace m'1 with m'0.
Auto.
Apply H with a0 b0.
Auto.
Auto.
Auto.
Auto.
Save.


(* if you extend, then it must be in the map *)
Lemma mextend_imp_mlookup : (A,B:Set; m,m':(Map A B); a:A; b:B)
      	       	       	    ( (x,x':A)x=x'\/~x=x' ) ->
      	       	       	    (mextend A B m a b m') ->
			    (mlookup A B m' a b).
Proof.
Intros.
Induction m.
Inversion H0.
Constructor.
Inversion H0.
Constructor.
Save.


(* if you extend a well-formed map correctly, it is still well-formed *)
Lemma mextend_wf 
  : (A,B:Set; m,m':(Map A B); a:A; b:B)
    (mWF ? ? m) ->
    (mnotindom ? ? m a) ->
    (mextend ? ? m a b m') ->
    (mWF ? ? m').
Proof.
Intros.
Inversion H1.
Constructor.
Auto.
Auto.
Save.




Lemma mextend_preserv
  : (A,B:Set; m,m':(Map A B); a,a':A; b,b':B)
    (mextend A B m a' b' m') ->
    (mlookup A B m a b) ->
    ~a=a' ->
    (mlookup A B m' a b).
Proof.
Intros.
Induction H.
Apply mlooknext.
Auto.
Auto.
Save.



(* ... in the reverse order *)
Lemma mextend_preserv_rev
  : (A,B:Set; m,m':(Map A B); a,a':A; b,b':B)
    (mextend A B m a' b' m') ->
    (mlookup A B m' a b) ->
    ~a=a' ->
    (mlookup A B m a b).
Proof.
Induction m.
Intros.
Inversion H.
Rewrite <- H6 in H0.
Inversion H0.
Rewrite H7 in H1.
Absurd ~a=a.
Auto.
Auto.
Inversion H12.
Intros.
Inversion H0.
Rewrite <- H7 in H1.
Move H2 after H1.
Inversion H1.
Rewrite H8 in H2.
Absurd ~a0=a0.
Auto.
Auto.
Auto.
Save.

(* extended map is one longer *)
Lemma mextend_size_Sn
  : (A,B:Set; m,m':(Map A B); a:A; b:B; s:nat)
    (mextend ? ? m a b m') ->
    (msize ? ? m s) ->
    (msize ? ? m' (S s)).
Proof.
Induction m.
Intros.
Inversion H.
Inversion H0.
Constructor.
Constructor.
Intros.
Inversion H0.
Constructor.
Auto.
Save.


(* you can always update or extend a map *)
Lemma mupdext_always : (A,B:Set; m:(Map A B); a:A; b:B)
      	       	       (P:((x,x':A) x=x' \/ ~x=x'))
      	       	       (EX m' | (mupdext A B m a b m')).  
Proof.
Intros.
Induction m.
Split with (melem ? ? a b (mempty ? ?)).
Constructor.
Inversion_clear Hrecm.
Cut a=a0\/~a=a0.
Intros.
Induction H0.
Split with (melem ? ? a0 b m).
Replace a0 with a.
Constructor.
Split with (melem ? ? a0 b0 x).
Apply mupdext1.
Auto.
Auto.
Apply P.
Save.


(* updext-ing a map preserves the old mapping values for non-affect elements *)
Lemma mupdext_preserv
  : (A,B:Set; m,m':(Map A B); a,a':A; b,b':B)
    (mupdext A B m a' b' m') ->
    (mlookup A B m a b) ->
    ~a=a' ->
    (mlookup A B m' a b).
Proof.
Intros.
Induction H.
Inversion H0.
Inversion H0.
Constructor.
Apply mlooknext.
Apply HrecH.
Auto.
Auto.
Auto.
Inversion H0.
Rewrite H5 in H1.
Absurd ~a=a.
Auto.
Auto.
Apply mlooknext.
Auto.
Auto.
Save.


(* ... in the reverse order *)
Lemma mupdext_preserv_rev
  : (A,B:Set; m,m':(Map A B); a,a':A; b,b':B)
    (mupdext A B m a' b' m') ->
    (mlookup A B m' a b) ->
    ~a=a' ->
    (mlookup A B m a b).
Proof.
Intros.
Induction H.
Inversion H0.
Rewrite H5 in H1.
Absurd ~a=a.
Auto.
Auto.
Auto.
Inversion H0.
Constructor.
Apply mlooknext.
Auto.
Auto.
Inversion H0.
Rewrite H5 in H1.
Absurd ~a=a.
Auto.
Auto.
Apply mlooknext.
Auto.
Auto.
Save.



Lemma mupdext_preserv_rev_mindom
  : (A,B:Set; m,m':(Map A B); a,a':A; b':B)
    ( (x,x':A)x=x'\/~x=x' ) ->
    (mupdext A B m a' b' m') ->
    (mindom A B m' a) ->
    ~a=a' ->
    (mindom A B m a).
Proof.
Induction m.
Intros.
Inversion H0.
Rewrite <- H6 in H1.
Inversion H1.
Rewrite H3 in H2.
Absurd ~a=a; Auto.
Auto.
Intros.
Inversion H1.
Rewrite <- H9 in H2.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H12.
Rewrite H13.
Constructor.
Apply minnext.
Auto.
Apply H with m'0 a' b'.
Auto.
Auto.
Inversion H2.
Rewrite H12 in H13.
Absurd ~a0=a0.
Auto.
Auto.
Auto.
Auto.
Auto.
Rewrite <- H9 in H2.
Inversion H2.
Rewrite H10 in H3.
Absurd ~a0=a0; Auto.
Apply minnext.
Auto.
Auto.
Save.



(* having updated or extended a map, the element is now in it *)
Lemma mupdext_imp_mlookup : (A,B:Set; m,m':(Map A B); a:A; b:B)
      	       	       	    ( (x,x':A)x=x' \/ ~x=x' ) ->
      	       	       	    (mupdext A B m a b m') ->
			    (mlookup A B m' a b).
Proof.
Induction m.
Intros m' a b P H.
Inversion H.
Constructor.
Intros a b m0 H m' a0 b0 P.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H1.
Rewrite H2 in H0.
Inversion H0.
Absurd ~a0=a0.
Auto.
Auto.
Constructor.
Inversion H0.
Constructor.
Apply (H m'0 a0 b0).
Auto.
Auto.
Auto.
Constructor.
Apply P.
Save.




Lemma mupdext_same_eq : (A,B:Set; m,m',m'':(Map A B); a:A; b:B)
      	       	       	( (x,x':A) x=x' \/ ~x=x' ) ->
      	       	       	(mupdext A B m a b m') ->
			(mupdext A B m a b m'') ->
			m'=m''.
Proof.
Induction m.
Intros.
Inversion H0.
Inversion H1.
Auto.
Intros.
Cut a=a0\/~a=a0.
Intros.
Inversion_clear H3.
Rewrite <- H4 in H1.
Rewrite <- H4 in H2.
Inversion H1.
Absurd ~a=a.
Auto.
Auto.
Inversion H2.
Absurd ~a=a; Auto.
Auto.
Inversion H1.
Inversion H2.
Replace m'1 with m'0.
Auto.
Apply H with a0 b0.
Auto.
Auto.
Auto.
Rewrite H12 in H11.
Absurd ~a0=a0; Auto.
Rewrite H3 in H4.
Absurd ~a0=a0; Auto.
Auto.
Save.


(*
Lemma mupdate_imp_mindom : (A,B:Set; m,m':(Map A B); a:A; b:B)
      	       	       	   (mupdate A B m a b m') ->
			   (mindom A B m a).
Proof.
Intros.
Induction H.
Constructor.
Constructor.
Auto.
Auto.
Save.

Lemma mindom_empty_false : (A,B:Set; a:A) (mindom A B (mempty A B) a) -> False.
Proof.
Intros.
Inversion H.
Save.

Fixpoint mupdatef [A,B:Set; m:(Map A B); a:A] 
  :  B -> (mindom A B m a) -> (Map A B)
  := [b]
     <[m:(Map A B)](mindom A B m a) -> (Map A B)>
     Cases m of (mempty) => [min:(mindom A B (mempty A B) a)]
                           (False_rec (Map A B) (mindom_empty_false A B a min))
              | (melem a' b' m') 
                  => [min:(mindom A B (melem A B a' b' m') a)] 
      	       	     Cases min of 
		        (minfound a b m') => m'
                      | (minnext a' a b' m' neqa min') => m'
              	     end
     end.
*)



Lemma ext_mupdext_two_preserv
      	       	     : (A,B:Set; m,m1,m2:(Map A B); a,a':A; b1,b2,b':B)
		       ( (x,x':A) x=x' \/ ~x=x' ) ->
		       (mupdext A B m a b1 m1) ->
		       (mupdext A B m a b2 m2) ->
		       ~a=a' ->
		       (mlookup A B m1 a' b') ->
		       (mlookup A B m2 a' b').
Proof.
Intros.
Cut (mlookup ? ? m a' b').
Intros.
Apply mupdext_preserv with m a b2.
Auto.
Auto.
Auto.
Apply mupdext_preserv_rev with m1 a b1.
Auto.
Auto.
Auto.
Save.


Lemma ext_map_oneoff_updext_same_oneoff
      	       	     : (A,B:Set; m1,m2,m1',m2':(Map A B); a,a0:A; b0,b1,b2:B)
		       ( (x,x':A) x=x' \/ ~x=x' ) ->
		       ( (a':A; b':B) ~a=a' ->
		                      (mlookup ? ? m1 a' b') ->
				      (mlookup ? ? m2 a' b') ) ->
		       (mupdext ? ? m1 a0 b0 m1') ->
		       (mupdext ? ? m2 a0 b0 m2') ->
		       ( (a':A; b':B) ~a=a' ->
		                      (mlookup ? ? m1' a' b') ->
				      (mlookup ? ? m2' a' b') ).
Proof.
Intros.
Cut a'=a0\/~a'=a0.
Intros.
Inversion_clear H5.
Rewrite H6.
Rewrite H6 in H4.
Replace b' with b0.
Apply mupdext_imp_mlookup with m2.
Auto.
Auto.
Apply (mlookup_eq ? ? a0 b0 b' m1').
Auto.
Apply mupdext_imp_mlookup with m1; Auto.
Auto.
Cut (mlookup ? ? m2 a' b').
Intros.
Apply mupdext_preserv with m2 a0 b0.
Auto.
Auto.
Auto.
Apply H0.
Auto.
Apply mupdext_preserv_rev with m1' a0 b0.
Auto.
Auto.
Auto.
Auto.
Save.


Lemma ext_map_eq_updext_eq
      	       	     : (A,B:Set; m1,m2,m1',m2':(Map A B); a:A; b:B)
		       ( (x,x':A) x=x' \/ ~x=x' ) ->
		       ( (a':A; b':B) (mlookup ? ? m1 a' b') ->
		                      (mlookup ? ? m2 a' b') ) ->
		       (mupdext ? ? m1 a b m1') ->
		       (mupdext ? ? m2 a b m2') ->
		       ( (a':A; b':B) (mlookup ? ? m1' a' b') ->
		                      (mlookup ? ? m2' a' b') ).
Proof.
Induction m1.
Induction m2.
Intros.
Inversion H1.
Inversion H2.
Rewrite <- H7 in H3.
Inversion H3.
Constructor.
Inversion H15.
Intros.
Inversion H2.
Rewrite <- H8 in H4.
Inversion H4.
Inversion H3.
Replace a' with a0.
Replace b' with b0.
Apply mlooknext.
Apply mupdext_imp_mlookup with m.
Auto.
Auto.
Auto.
Inversion H4.
Constructor.
Rewrite H5 in H25.
Absurd ~a'=a'; Auto.
Inversion H13.
Induction m2.
Intros.
Absurd (a':A; b':B)
        (mlookup A B (melem A B a b m) a' b')
        ->(mlookup A B (mempty A B) a' b').
Unfold not; Intros.
Cut (mlookup ? ? (mempty ? ?) a b).
Intros.
Inversion H6.
Apply H5.
Constructor.
Auto.
Intros.
Cut a'=a1\/~a'=a1.
Intros.
Inversion_clear H6.
Replace a' with a1.
Replace b' with b1.
Apply mupdext_imp_mlookup with (melem A B a0 b0 m0).
Auto.
Auto.
Cut (mlookup ? ? m1' a' b1).
Intros.
Apply (mlookup_eq ? ? a' b1 b' m1').
Auto.
Auto.
Auto.
Replace a' with a1.
Apply mupdext_imp_mlookup with (melem ? ? a b m).
Auto.
Auto.
Cut (mlookup ? ? (melem A B a0 b0 m0) a' b').
Intros.
Apply mupdext_preserv with (melem ? ? a0 b0 m0) a1 b1.
Auto.
Auto.
Auto.
Apply H2.
Apply mupdext_preserv_rev with m1' a1 b1.
Auto.
Auto.
Auto.
Auto.
Save.


Lemma ext_map_lookup_eq
      	       	     : (A,B:Set; m1,m2:(Map A B); a:A; b:B)
		       ( (x,x':A) x=x' \/ ~x=x' ) ->
		       ( (a':A; b':B) ~a=a' -> (mlookup ? ? m1 a' b') ->
		                               (mlookup ? ? m2 a' b') ) ->
		       (mlookup ? ? m1 a b) ->
		       (mlookup ? ? m2 a b) ->
		       ( (a':A; b':B) (mlookup ? ? m1 a' b') ->
		                      (mlookup ? ? m2 a' b') ).
Proof.
Intros.
Cut a=a'\/~a=a'.
Intros.
Inversion_clear H4.
Rewrite <- H5 in H3.
Replace a' with a; Replace b' with b.
Auto.
Apply (mlookup_eq ? ? a b b' m1).
Auto.
Auto.
Auto.
Apply H0.
Auto.
Auto.
Auto.
Save.
