(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.0 - 7.3)
 *)


(***************************************************************************)
(* Utilities for manipulating natural numbers *)
(***************************************************************************)

Require Div2.
Require EqNat.
Require Lt.
Require Le.



(***************************************************************************)
Fixpoint blt_nat [n:nat] : nat -> bool
  := [m:nat]
     Cases n m of 
       	       O O => false
	     | O (S _) => true
	     | (S _) O => false
	     | (S n1) (S m1) => (blt_nat n1 m1)
	end.


Lemma blt_S_false : (a:nat) (blt_nat (S a) a) = false.
Proof.
Induction a.
Unfold blt_nat.
Auto.
Intros.
Unfold blt_nat.
Auto.
Save.

Lemma blt_S_eq : (a,b:nat; t:bool) (blt_nat a b)=t -> (blt_nat (S a) (S b))=t.
Proof.
Intros.
Unfold blt_nat.
Fold blt_nat.
Auto.
Save.

Lemma blt_lt_true : (a,b:nat) (lt a b) -> (blt_nat a b)=true.
Proof.
Induction a.
Intros.
Inversion H.
Unfold blt_nat.
Auto.
Unfold blt_nat.
Auto.
Intros.
Inversion H0.
Apply blt_S_eq.
Apply H.
Apply lt_n_Sn.
Apply blt_S_eq.
Apply H.
Rewrite <- H2 in H0.
Apply lt_S_n.
Auto.
Save.


Lemma blt_Sm_false : (a,b:nat) (blt_nat a b)=false -> (blt_nat (S a) b)=false.
Proof.
Double Induction 1 2.
Intros.
Unfold blt_nat.
Auto.
Intros.
Unfold blt_nat in H0.
Absurd true=false.
Discriminate.
Auto.
Intros.
Unfold blt_nat.
Auto.
Intros.
Unfold blt_nat in H1.
Fold blt_nat in H1.
Cut (blt_nat (S n0) n)=false.
Intros.
Apply blt_S_eq.
Auto.
Apply H0.
Auto.
Save.

Lemma blt_false : (a,b:nat) (lt a b) -> (blt_nat b a)=false.
Proof.
Intros.
Induction H.
Apply blt_S_false.
Apply blt_Sm_false.
Auto.
Save.


Lemma blt_eq_false : (a:nat) (blt_nat a a)=false.
Proof.
Induction a.
Unfold blt_nat.
Auto.
Intros.
Unfold blt_nat.
Auto.
Save.


Lemma le_blt_false : (a,b:nat) (le a b) -> (blt_nat b a)=false.
Proof.
Intros.
Induction H.
Apply blt_eq_false.
Apply blt_false.
Unfold lt.
Apply le_n_S.
Auto.
Save.




(***************************************************************************)


Lemma beq_neq_false : (a,b:nat) ~a=b -> false=(beq_nat a b).
Proof.
Double Induction 1 2.
Intros.
Absurd ~O=O; Auto.
Intros.
Unfold beq_nat.
Auto.
Intros.
Unfold beq_nat.
Auto.
Intros.
Unfold beq_nat.
Fold beq_nat.
Apply H0.
Auto.
Save.


Lemma eq_neq_Sl : (a,b:nat) a=b -> ~(S a)=b.
Proof.
Double Induction 1 2.
Intros.
Auto.
Intros.
Absurd O=(S n).
Discriminate.
Auto.
Intros.
Absurd O=(S n).
Discriminate.
Auto.
Intros.
Apply not_eq_S.
Auto.
Save.


(***************************************************************************)
(* definitions of numbers *)
Definition N0 := O.
Definition N1 := (S N0).
Definition N2 := (S N1).
Definition N3 := (S N2).
Definition N4 := (S N3).
Definition N5 := (S N4).
Definition N6 := (S N5).
Definition N7 := (S N6).
Definition N8 := (S N7).
Definition N9 := (S N8).
Definition N10 := (S N9).
Definition N11 := (S N10).
Definition N12 := (S N11).
Definition N13 := (S N12).
Definition N14 := (S N13).
Definition N15 := (S N14).
Definition N16 := (S N15).
Definition N17 := (S N16).
Definition N18 := (S N17).
Definition N19 := (S N18).
Definition N20 := (S N19).
Definition N21 := (S N20).
Definition N22 := (S N21).
Definition N23 := (S N22).
Definition N24 := (S N23).
Definition N25 := (S N24).
Definition N26 := (S N25).
Definition N27 := (S N26).
Definition N28 := (S N27).
Definition N29 := (S N28).
Definition N30 := (S N29).
Definition N31 := (S N30).
Definition N32 := (S N31).

(***************************************************************************)
(* utilities for decoding *)

(* to get  n % d  call this as (modaux n d d O), the second argument is
   decremented in recursive calls, the fourth is incremented *)
Fixpoint modaux [n:nat] : nat -> nat -> nat -> nat
  := [c,d,r:nat]
     Cases n of O => r
      	     | (S p) => Cases c of O => O 
      	       	       	       	| (S O) => (modaux p d d O)
	                        | (S c') => (modaux p c' d (S r)) end end.

Definition mod : nat -> nat -> nat
  := [n,d:nat] (modaux n d d O).

Definition div4 := [n:nat] (div2 (div2 n)).
Definition div8 := [n:nat] (div4 (div2 n)).
Definition div16 := [n:nat] (div2 (div8 n)).
Definition div32 := [n:nat] (div2 (div16 n)).


Definition mult4 := [n:nat] (mult N4 n).
Definition mult8 := [n:nat] (mult N8 n).
Definition mult16 := [n:nat] (mult N16 n).
Definition mult32 := [n:nat] (mult N32 n).


(***************************************************************************)
Theorem eqorneq_nat : (a,a':nat)a=a'\/~a=a'.
Proof.
Intros.
Compare a a'.
Auto.
Auto.
Save.


Theorem le_or_gt : (m,n:nat) (le n m) \/ (gt n m).
Proof.
Double Induction 1 2.
Left.
Auto.
Intros.
Right.
Red.
Red.
Apply le_n_S.
Apply le_O_n.
Intros.
Left.
Apply le_O_n.
Intros.
Cut (le n n1)\/(gt n n1).
Intros.
Inversion_clear H1.
Left.
Apply le_n_S.
Auto.
Right.
Unfold gt.
Unfold gt in H2.
Apply lt_n_S.
Auto.
Apply H0.
Save.


Lemma lt_neq : (m,n:nat) (lt n m) -> ~n=m.
Proof.
Induction m.
Intros.
Inversion H.
Intros.
Induction n0.
Auto.
Cut (lt n0 n).
Intros.
Cut ~n0=n.
Intros.
Auto.
Apply H.
Auto.
Apply lt_S_n.
Auto.
Save.


Lemma lt_S_neq : (n,x:nat) (lt n (S x)) -> ~x=n -> (lt n x).
Proof.
Double Induction 1 2.
Intros.
Absurd ~O=O.
Auto.
Auto.
Intros.
Apply lt_O_Sn.
Intros.
Unfold lt in H0.
Cut (le (S n0) O).
Intros.
Absurd (le (S n0) O).
Apply le_Sn_O.
Auto.
Apply le_S_n.
Auto.
Intros.
Apply lt_n_S.
Cut n1=n0\/~n1=n0.
Intros.
Inversion_clear H3.
Absurd ~(S n1)=(S n1).
Auto.
Rewrite <- H4 in H2.
Auto.
Cut (lt n1 (S n0)).
Intros.
Apply (H0 n0 H3).
Auto.
Apply lt_S_n.
Auto.
Apply eqorneq_nat.
Save.


Lemma lt_S_eq_or_lt : (n,x:nat) (lt n (S x)) -> (x=n \/ (lt n x)).
Proof.
Intros.
Cut x=n\/~x=n.
Intros.
Inversion_clear H0.
Left; Auto.
Right; Apply lt_S_neq.
Auto.
Auto.
Apply eqorneq_nat.
Save.


Lemma plus_n_O_n : (n:nat) (plus n O)=n.
Proof.
Induction n.
Unfold plus.
Auto.
Intros.
Unfold plus.
Fold plus.
Rewrite H.
Auto.
Save.


Lemma plus_S_neq : (n,m:nat) ~n=(plus n (S m)).
Proof.
Induction n.
Intros.
Unfold plus.
Auto.
Intros.
Unfold plus.
Fold plus.
Auto.
Save.


(***************************************************************************)
