(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.0)
 *)


(***************************************************************************)
(* Typed Instruction Set *)
(***************************************************************************)

Require EqNat.
Require natutil.

(***************************************************************************)
(* Syntax *)

Require tisreg.

Syntactic Definition Word    := nat.
Syntactic Definition Mem     := Word -> Word.
Syntactic Definition _RegFile := _Reg -> Word.
Syntactic Definition State   := (Mem * (_RegFile * Word)).

Inductive _Instr : Set
  := _add  : _Reg -> _Reg -> _Reg -> _Instr
  |  _addi : _Reg -> _Reg -> Word -> _Instr
  |  _movi : _Reg -> Word -> _Instr
  |  _bgt  : _Reg -> _Reg -> Word -> _Instr
  |  _jd   : Word -> _Instr
  |  _jmp  : _Reg -> _Instr
  |  _ld   : _Reg -> _Reg -> Word -> _Instr
  |  _st   : _Reg -> Word -> _Reg -> _Instr
  |  _ill  : _Instr
.


(***************************************************************************)

(* decode utilities *)
Definition icoded : Word -> nat := [w:Word] (mod w N8).
Definition r1argd : Word -> nat := [w:Word] (mod (div8 w) N32).
Definition r2argd : Word -> nat := [w:Word] (mod (div32 (div8 w)) N32).
Definition r3argd : Word -> nat := [w:Word] (mod (div32 (div32 (div8 w))) N32).
Definition w1argd : Word -> nat := [w:Word] (div8 w).
Definition w2argd : Word -> nat := [w:Word] (div32 (div8 w)).
Definition w3argd : Word -> nat := [w:Word] (div32 (div32 (div8 w))).


(* encode utilities *)
Definition icodee : nat -> Word := [n:nat] n.
Definition r1arge : nat -> Word := [n:nat] (mult8 n).
Definition r2arge : nat -> Word := [n:nat] (mult32 (mult8 n)).
Definition r3arge : nat -> Word := [n:nat] (mult32 (mult32 (mult8 n))).
Definition w1arge : nat -> Word := [n:nat] (mult8 n).
Definition w2arge : nat -> Word := [n:nat] (mult32 (mult8 n)).
Definition w3arge : nat -> Word := [n:nat] (mult32 (mult32 (mult8 n))).


(* Decode function *)
Definition Dc : Word -> _Instr
  := [w:Word] Cases (icoded w) of
      	O => 	(_add (nat2_reg (r1argd w)) (nat2_reg (r2argd w)) 
      	       	      (nat2_reg (r3argd w)))
      | (S O) => (_addi (nat2_reg (r1argd w)) (nat2_reg (r2argd w))
		      (w3argd w))
      | (S (S O)) => (_movi (nat2_reg (r1argd w)) (w2argd w))
      | (S (S (S O))) => (_bgt (nat2_reg (r1argd w)) (nat2_reg (r2argd w))
      	       	       	       	       	 (w3argd w))
      | (S (S (S (S O)))) => (_jd (w1argd w))
      | (S (S (S (S (S O))))) => (_jmp (nat2_reg (r1argd w)))
      | (S (S (S (S (S (S O)))))) => (_ld (nat2_reg (r1argd w))
      	       	       	       	       	  (nat2_reg (r2argd w))
					  (w3argd w))
      | (S (S (S (S (S (S (S O))))))) => (_st (nat2_reg (r1argd w))
      	       	       	       	       	      (w3argd w)
      	       	       	       	       	      (nat2_reg (r2argd w)))
      | _ =>    (_ill) end.


(* Encode function *)
Definition Ec : _Instr -> Word
  := [i:_Instr] Cases i of
      	(_add rd rs rs') => (plus (icodee O) 
	                    (plus (r1arge (_reg2nat rd))
			    (plus (r2arge (_reg2nat rs))
			          (r3arge (_reg2nat rs')))))
      | (_addi rd rs w) => (plus (icodee N1)
			   (plus (r1arge (_reg2nat rd))
			   (plus (r2arge (_reg2nat rs))
			         (w3arge w))))
      | (_movi rd w) => (plus (icodee N2)
      	       	       	(plus (r1arge (_reg2nat rd))
			      (w2arge w)))
      | (_bgt rd rs w) => (plus (icodee N3)
      	       	       	  (plus (r1arge (_reg2nat rd))
			  (plus (r2arge (_reg2nat rs))
			        (w3arge w))))
      | (_jd w) => (plus (icodee N4)
      	       	       	 (w1arge w))
      | (_jmp r) => (plus (icodee N5) (r1arge (_reg2nat r)))
      | (_ld rd rs w) => (plus (icodee N6)
      	       	       	 (plus (r1arge (_reg2nat rd))
			 (plus (r2arge (_reg2nat rs))
			       (w3arge w))))
      | (_st rd w rs) => (plus (icodee N7)
      	       	       	 (plus (r1arge (_reg2nat rd))
			 (plus (r2arge (_reg2nat rs))
			       (w3arge w))))
      | _ill => O
      end.


(***************************************************************************)
(* Machine semantics *)
Definition updatereg : _RegFile -> _Reg -> Word -> _RegFile
  := [R:_RegFile; rd:_Reg; v:Word]
     ([r:_Reg] if (beq_reg r rd) then v else (R r)).

Definition updatemem : Mem -> Word -> Word -> Mem
  := [M:Mem; a:Word; v:Word]
     ([w:Word] if (beq_nat w a) then v else (M w)).

Definition Step : State -> State
  := [St:State] Cases St of (M, (R, pc)) =>
    Cases (Dc (M pc)) of
        (_add rd rs rs') => (M, ((updatereg R rd (plus (R rs) (R rs'))),
	                         (S pc)))
      | (_addi rd rs w) => (M, ((updatereg R rd (plus (R rs) w)), (S pc)))
      | (_movi rd w) => (M, ((updatereg R rd w), (S pc)))
      | (_bgt rs rt w) => (if (blt_nat (R rt) (R rs))
      	       	       	   then (M, (R, w))
			   else (M, (R, (S pc))))
      | (_jd w) => (M, (R, w))
      | (_jmp r) => (M, (R, (R r)))
      | (_ld rd rs w) => (M, ((updatereg R rd (M (plus (R rs) w))), (S pc)))
      | (_st rd w rs) => ((updatemem M (plus (R rd) w) (R rs)), (R, (S pc)))
      | _ill => St
    end
  end.

Fixpoint MultiStep [n:nat] : State -> State
  := Cases n of O => [S:State] (Step S)
      	   | (S m) => [S:State] (Step (MultiStep m S))
     end.

(***************************************************************************)



