(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)


(***************************************************************************)
(* Translation from Mini-FTAL to machine state *)
(***************************************************************************)

Require Le.
Require Lt.
Require Gt.
Require Plus.

Require ftal.
Load tis.


(***************************************************************************)
(* the safety policy *)
Definition SP 
  := [S:State] 
     (let (M,T')=S in 
     (let (R,PC)=T' in
       ~(Dc (M PC))=_ill
     )).


Definition Safe
  := [S:State] (n:nat)(SP (MultiStep n S)).


(***************************************************************************)
(* ---------- the layout function ----------------- *)

Definition HV_size : HeapVal -> nat
  := [hv]
     Cases hv of (tuple V) => (length V)
       	       | (code G T I') => (lenInstrSeq I')
     end.


Fixpoint Layout_aux [H:(Map label HeapVal)] : label -> Word
  := Cases H of mempty 
      	       	   => ([l':label] O)
	      | (melem l h H')
	           => [l':label]
		      (if (blt_nat l l')
      	       	       	       	  then (plus (Layout_aux H' l') (HV_size h))
				  else (Layout_aux H' l'))
     end.


Definition Layout : Heap -> label -> Word
  := [H] (Layout_aux (hunwrap H)).


Syntactic Definition LayoutF := (label -> Word).


(***************************************************************************)
(* translations *)

(* register translations: *)
Require translate_ftal_aux.



Inductive TrWordVal : LayoutF -> WordVal -> Word -> Prop
  := trwv_label : (L:LayoutF; l:label) (TrWordVal L (wl l) (L l))
  |  trwv_int   : (L:LayoutF; i:int) (TrWordVal L (wi i) i)
  |  trwv_uninit : (L:LayoutF; t:Omega; l:label) (TrWordVal L (wuninit t) l)
  |  trwv_fold  : (L:LayoutF; v:WordVal; t:Omega; w:Word)
      	       	  (TrWordVal L v w) ->
		  (TrWordVal L (wfold v t) w)
.


Inductive TrWordValList : LayoutF -> (list WordVal) -> Mem -> Word -> Prop
  := trwvl_nil  : (L:LayoutF; M:Mem; l:Word)
      	       	  (TrWordValList L (nil ?) M l)
  |  trwvl_cons : (L:LayoutF; v:WordVal; V:(list WordVal); M:Mem; w,l:Word)
      	       	  (TrWordVal L v w) ->
		  (M l)=w ->
      	       	  (TrWordValList L V M (S l)) ->
		  (TrWordValList L (cons v V) M l)
.


Inductive TrInstr : LayoutF -> Instr -> _Instr -> Prop
  := tri_add    : (L:LayoutF; rd,rs,rt:Reg; _rd,_rs,_rt:_Reg)
      	       	  (TrRegF rd)=_rd ->
      	       	  (TrRegF rs)=_rs ->
      	       	  (TrRegF rt)=_rt ->
		  (TrInstr L (add rd rs rt) (_add _rd _rs _rt))
  |  tri_addi   : (L:LayoutF; rd,rs:Reg; _rd,_rs:_Reg; i:int)
      	       	  (TrRegF rd)=_rd ->
      	       	  (TrRegF rs)=_rs ->
		  (TrInstr L (addi rd rs i) (_addi _rd _rs i))
  |  tri_alloc  : (L:LayoutF; rd:Reg; _rd:_Reg; Ts:(list Omega))
                  (TrRegF rd)=_rd ->
		  (TrInstr L (alloc rd Ts) (_addi _rd _r31 O))
  |  tri_bgt    : (L:LayoutF; rs,rt:Reg; _rs,_rt:_Reg; l:label)
      	       	  (TrRegF rs)=_rs ->
      	       	  (TrRegF rt)=_rt ->
		  (TrInstr L (bgt rs rt l) (_bgt _rs _rt (L l)))
  |  tri_bump   : (L:LayoutF; i:int)
		  (TrInstr L (bump i) (_addi _r31 _r31 i))
  |  tri_fold   : (L:LayoutF; rd,rs:Reg; t:Omega; _rd,_rs:_Reg)
      	       	  (TrRegF rd)=_rd ->
      	       	  (TrRegF rs)=_rs ->
      	       	  (TrInstr L (fold rd t rs) (_addi _rd _rs O))
  |  tri_ld	: (L:LayoutF; rd,rs:Reg; i:int; _rd,_rs:_Reg)
      	       	  (TrRegF rd)=_rd ->
		  (TrRegF rs)=_rs ->
		  (TrInstr L (ld rd rs i) (_ld _rd _rs i))
  |  tri_mov    : (L:LayoutF; rd,rs:Reg; _rd,_rs:_Reg)
      	       	  (TrRegF rd)=_rd ->
		  (TrRegF rs)=_rs ->
		  (TrInstr L (mov rd rs) (_addi _rd _rs O))
  |  tri_movi   : (L:LayoutF; r:Reg; _r:_Reg; i:int)
      	       	  (TrRegF r)=_r ->
		  (TrInstr L (movi r i) (_movi _r i))
  |  tri_movl   : (L:LayoutF; r:Reg; _r:_Reg; l:label)
      	       	  (TrRegF r)=_r ->
      	       	  (TrInstr L (movl r l) (_movi _r (L l)))
  |  tri_st     : (L:LayoutF; rd,rs:Reg; i:int; _rd,_rs:_Reg)
      	       	  (TrRegF rd)=_rd ->
		  (TrRegF rs)=_rs ->
		  (TrInstr L (st rd i rs) (_st _rd i _rs))
  |  tri_unfold : (L:LayoutF; rd,rs:Reg; _rd,_rs:_Reg)
      	       	  (TrRegF rd)=_rd ->
		  (TrRegF rs)=_rs ->
		  (TrInstr L (unfold rd rs) (_addi _rd _rs O))
.


Inductive TrInstrSeq : LayoutF -> InstrSeq -> Mem -> Word -> Prop
  := tris_iseq  : (L:LayoutF; i:Instr; _i:_Instr; I:InstrSeq; M:Mem; l:Word)
      	       	  (TrInstr L i _i) ->
		  (Dc (M l)) = _i ->
		  (TrInstrSeq L I M (S l)) ->
		  (TrInstrSeq L (iseq i I) M l)
  |  tris_jd    : (L:LayoutF; l:label; M:Mem; w:Word)
      	       	  (Dc (M w)) = (_jd (L l)) ->
      	       	  (TrInstrSeq L (jd l) M w)
  |  tris_jmp   : (L:LayoutF; r:Reg; _r:_Reg; M:Mem; w:Word)
      	       	  (TrRegF r)=_r ->
      	       	  (Dc (M w)) = (_jmp _r) ->
		  (TrInstrSeq L (jmp r) M w)
.


Inductive TrHeapVal : LayoutF -> HeapVal -> Mem -> Word -> Prop
  := trhv_tuple : (L:LayoutF; V:(list WordVal); M:Mem; l:Word)
      	       	  (TrWordValList L V M l) ->
      	       	  (TrHeapVal L (tuple V) M l)
  |  trhv_code  : (L:LayoutF; G:RegFileTy; T:APTy; I:InstrSeq; M:Mem; l:Word)
      	       	  (TrInstrSeq L I M l) ->
      	       	  (TrHeapVal L (code G T I) M l)
.


Inductive TrHeap  : LayoutF -> Heap -> Mem -> Prop
  := trheap     : (L:LayoutF; H:Heap; s:nat; M:Mem)
      	       	  (hsize (hunwrap H) s) ->
		  ( (n:nat; h:HeapVal)
(*		    (lt n s) ->  *)
      	       	    (hlookup (hunwrap H) n h) ->
      	       	    (TrHeapVal L h M (L n)) ) ->
		  (TrHeap L H M)
.


Inductive TrRegFile : LayoutF -> RegFile -> _RegFile -> Prop
  := trregfile  : (L:LayoutF; R:RegFile; _R:_RegFile)
      	       	  ( (r:Reg; _r:_Reg; v:WordVal)
		    (mlookup ? ? R r v) ->
		    (TrRegF r)=_r ->
		    (TrWordVal L v (_R _r))) ->
(*		    (_R _r)=w ->
		    (TrWordVal L v w)) -> *)
      	       	  (TrRegFile L R _R)
.


Inductive TrAP : LayoutF -> AP -> _RegFile -> Prop
  := trap      : (L:LayoutF; A:AP; _R:_RegFile)
      	       	 (_R _r31)=(L A) ->
		 (TrAP L A _R)
.  


Inductive TrProgram : Program -> State -> Prop
  := trprogram   : (H:Heap; R:RegFile; A:AP; I:InstrSeq; 
      	       	    M:Mem; _R:_RegFile; pc:Word; L:LayoutF)
		   L=(Layout H) ->
		   (TrHeap L H M) ->
		   (TrRegFile L R _R) ->
		   (TrAP L A _R) ->
		   (EX l | (EX G | (EX T | (EX I' | (EX n | 
		       (hlookup (hunwrap H) l (code G T I')) /\
		       (ISubDepth I I' n) /\
		       (TrInstrSeq L I' M (L l)) /\
		       (plus (L l) n)=pc))))) -> 
(* 		   (TrInstrSeq L I M pc) ->  *)
      	       	   (TrProgram (H,(R,(A,I))) (M,(_R,pc)))
.


(***************************************************************************)


(***************************************************************************)
(* FPCC proofs *)


Definition Inv := [S:State] (EXT P:Program |
     	       	       	      (EXT D:(WFProgram P) |
 			         (TrProgram P S))).



(***************************************************************************)
(* utilities *)
Lemma tr_sub_instrseq
  : (IS,IS':InstrSeq; n:nat; L:LayoutF; M:Mem; l:Word)
    (ISubDepth IS' IS n) ->
    (TrInstrSeq L IS M l) ->
    (TrInstrSeq L IS' M (plus n l)).
Proof.
Induction IS.
Intros.
Inversion H0.
Auto.
Replace (plus (S n0) l) with (plus n0 (S l)).
Apply H.
Auto.
Inversion H1.
Auto.
Apply sym_eq.
Apply plus_Snm_nSm.
Intros.
Inversion H.
Replace (plus O l) with l.
Auto.
Auto.
Intros.
Inversion H.
Replace (plus O l) with l.
Auto.
Auto.
Save.


Lemma tr_pc_trinstrseq
  : (H:Heap; I:InstrSeq; L:LayoutF; M:Mem; pc:Word)
    (EX l | (EX G | (EX T | (EX I' | (EX n | 
      	(hlookup (hunwrap H) l (code G T I')) /\
	(ISubDepth I I' n) /\
	(TrInstrSeq L I' M (L l)) /\ (plus (L l) n)=pc))))) ->
    (TrInstrSeq L I M pc).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H3.
Rewrite <- H4.
Replace (plus (L x) x3) with (plus x3 (L x)).
Apply tr_sub_instrseq with x2.
Auto.
Auto.
Apply plus_sym.
Save.





Lemma tr_instrseq_subdepth_S
  : (IS',IS:InstrSeq; i:Instr; n:nat)
    (ISubDepth (iseq i IS) IS' n) ->
    (ISubDepth IS IS' (S n)).
Proof.
Induction IS'.
Intros.
Constructor.
Inversion H0.
Constructor.
Apply H with i1.
Auto.
Intros.
Inversion H.
Intros.
Inversion H.
Save.


Lemma tr_pc_Spc
  : (H:Heap; i:Instr; I:InstrSeq; L:LayoutF; M:Mem; pc:Word)
    (EX l | (EX G | (EX T | (EX I' | (EX n |
      	(hlookup (hunwrap H) l (code G T I')) /\
	(ISubDepth (iseq i I) I' n) /\
	(TrInstrSeq L I' M (L l)) /\ (plus (L l) n)=pc))))) ->
    (EX l | (EX G | (EX T | (EX I' | (EX n |
      	(hlookup (hunwrap H) l (code G T I')) /\
	(ISubDepth I I' n) /\
	(TrInstrSeq L I' M (L l)) /\ (plus (L l) n)=(S pc)))))).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H3.
Split with x.
Split with x0.
Split with x1.
Split with x2.
Split with (S x3).
Split.
Auto.
Split.
Apply tr_instrseq_subdepth_S with i.
Auto.
Split.
Auto.
Replace (plus (L x) (S x3)) with (plus (S x3) (L x)).
Unfold plus.
Apply eq_S.
Fold plus.
Replace (plus x3 (L x)) with (plus (L x) x3).
Auto.
Apply plus_sym.
Apply plus_sym.
Save.





(***************************************************************************)
(* Progress theorem *)

Theorem Progress : (S:State) (Inv S) -> (SP S).
Proof.
Intros.
Unfold Inv in H.
Elim H; Clear H; Intros P H.
Elim H; Clear H; Intros D H.
Inversion D.
Clear H5.
Rewrite <- H6 in H.
Inversion H.
Cut (TrInstrSeq L I M pc).
Intros.
Inversion H4.
Rewrite <- H25 in H17.
Inversion H17.
Red.
Rewrite H30.
Inversion H28.
Discriminate.

Rewrite <- H24 in H17.
Inversion H17.
Red.
Rewrite H29.
Inversion H27.
Discriminate.

Rewrite <- H24 in H17.
Inversion H17.
Red.
Rewrite H29.
Inversion H27.
Discriminate.
Auto.

Rewrite <- H26 in H17.
Red.
Inversion H17.
Rewrite H31.
Inversion H29.
Discriminate.

Rewrite <- H22 in H17.
Red.
Inversion H17.
Rewrite H27.
Inversion H25.
Discriminate.

Rewrite <- H25 in H17.
Inversion H17.
Red.
Rewrite H30.
Inversion H28.
Discriminate.

Rewrite <- H23 in H17.
Red.
Inversion H17.
Rewrite H26.
Discriminate.
Auto.

Rewrite <- H23 in H17.
Inversion H17.
Red.
Rewrite H27.
Discriminate.

Rewrite <- H26 in H17.
Inversion H17.
Red.
Rewrite H31.
Inversion H29.
Discriminate.
Auto.

Rewrite <- H24 in H17.
Inversion H17.
Red.
Rewrite H29.
Inversion H27.
Discriminate.
Auto.

Rewrite <- H23 in H17.
Inversion H17.
Red.
Rewrite H28.
Inversion H26.
Discriminate. Auto.

Rewrite <- H25 in H17.
Inversion H17.
Red.
Rewrite H30.
Inversion H28.
Discriminate. Auto.

Rewrite <- H27 in H17.
Inversion H17.
Red.
Rewrite H32.
Inversion H30.
Discriminate.

Rewrite <- H26 in H17.
Inversion H17.
Red.
Rewrite H31.
Inversion H29.
Discriminate.

Apply tr_pc_trinstrseq with H0.
Auto.
Save.





Lemma Preservation_add
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd,rs,rs':Reg; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (add rd rs rs') I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros S H R a rd rs rs' I P P' H0 H1 H2 H3 H4.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (add rd rs rs') I) M pc).
Intros.
Inversion H25.
Rewrite <- H30 in H28.
Inversion H28.

Unfold updatereg.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Constructor.
Intros.
Cut _r=_rd\/~_r=_rd.
Intros.
Inversion_clear H56.
Rewrite H57.
Replace (beq_reg _rd _rd) with true.
Cut v=(wi (plus rsval rsval')).
Intros.
Rewrite H56.
Replace (_R _rs) with rsval.
Replace (_R _rt) with rsval'.
Constructor.
Inversion H22.
Cut (TrWordVal L (wi rsval') (_R _rt)).
Intros.
Inversion H62.
Auto.
Apply H58 with rs'.
Auto.
Auto.
Cut (TrWordVal L (wi rsval) (_R _rs)).
Intros.
Inversion H58.
Auto.
Inversion H22.
Apply H58 with rs.
Auto.
Auto.
Apply (mlookup_eq ? ? rd v (wi (plus rsval rsval')) R').
Apply eqorneq_reg.
Replace rd with r.
Auto.
Apply trregf_eq_eq with _rd.
Rewrite <- H57.
Auto.
Auto.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Auto.
Apply sym_eq.
Apply beq_reg_eq_true.
Replace (beq_reg _r _rd) with false.
Inversion H22.
Apply H56 with r.
Apply mupdext_preserv_rev with R' rd (wi (plus rsval rsval')).
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Apply eqorneq_tisreg.
Constructor.
Replace (beq_reg _r31 _rd) with false.
Inversion H23.
Auto.
Cut ~_rd=_r31.
Intros.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Apply tr_pc_Spc with (add rd rs rs').
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.



Lemma Preservation_addi
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd,rs:Reg; i:int; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (addi rd rs i) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (addi rd rs i) I) M pc).
Intros.
Inversion H25.
Rewrite H30.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Constructor.
Intros.
Unfold updatereg.
Cut _r=_rd\/~_r=_rd.
Intros.
Inversion_clear H54.
Rewrite H55.
Replace (beq_reg _rd _rd) with true.
Cut v=(wi (plus rsval i)).
Intros.
Rewrite H54.
Replace (_R _rs) with rsval.
Constructor.
Inversion H22.
Cut (TrWordVal L (wi rsval) (_R _rs)).
Intros.
Inversion H60.
Auto.
Apply H56 with rs.
Auto.
Auto.
Apply (mlookup_eq ? ? rd v (wi (plus rsval i)) R').
Apply eqorneq_reg.
Replace rd with r.
Auto.
Apply trregf_eq_eq with _rd.
Rewrite <- H55.
Auto.
Auto.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Auto.
Apply sym_eq.
Apply beq_reg_eq_true.
Replace (beq_reg _r _rd) with false.
Inversion H22.
Apply H54 with r.
Apply mupdext_preserv_rev with R' rd (wi (plus rsval i)).
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Apply eqorneq_tisreg.
Constructor.
Unfold updatereg.
Replace (beq_reg _r31 _rd) with false.
Inversion H23.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_rd=_r31.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Apply tr_pc_Spc with (addi rd rs i).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.





Lemma heap_size_eq_label
  : (h:(Map label HeapVal); wfh:(mWF ? ? h))
    (ht:(Map label Omega); wfht:(mWF ? ? ht))
    (l:label)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (msize ? ? h (S l)) ->
    (EX hv | (EX h' | h=(melem ? ? l hv h'))).
Proof.
Intros.
Inversion H.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion H0.
Split with b.
Split with m.
Unfold unwrapMap in H2.
Rewrite <- H13 in H2.
Inversion H2.
Replace l with n0.
Replace a with n0.
Auto.
Apply sym_eq.
Apply eq_add_S.
Rewrite H16.
Unfold unwrapMap in H12.
Rewrite <- H13 in H12.
Inversion H12.
Apply (msize_eq ? ? (melem nat HeapVal a b m)).
Constructor.
Auto.
Auto.
Apply (msize_eq ? ? m).
Auto.
Auto.
Save.


Lemma heap_label_eq_size
  : (l:label; hv:HeapVal; h:(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? (melem ? ? l hv h)); G,G':HeapTy)
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? (melem ? ? l hv h) wf') G') ->
    (msize ? ? h l).
Proof.
Intros.
Inversion_clear H.
Inversion_clear H0.
Unfold unwrapMap in H2.
Unfold unwrapMap in H1.
Inversion_clear H2.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H1.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Clear H7.
Clear H12.
Clear H9.
Clear H6.
Clear H3.
Clear H.
Clear H5.
Replace l with x; Auto.
Cut x0=(S x).
Intros.
Rewrite H in H1.
Rewrite H in H8.
Rewrite H in H10.
Cut (le (S x) l)\/(lt l (S x)).
Intros.
Inversion_clear H3.
Cut (lt l (S x)).
Intros.
Absurd (le (S x) l).
Apply lt_not_le.
Auto.
Auto.
Apply H8 with hv.
Constructor.
Red in H5.
Cut l=x\/(lt (S l) (S x)).
Intros.
Inversion_clear H3.
Auto.
Cut (lt l x).
Intros.
Inversion wf'.
Absurd (mindom ? ? h l).
Apply mnotindom_imp_not_mindom.
Auto.
Auto.
Cut (EX h0:? | (mlookup ? ? h l h0)).
Intros.
Inversion_clear H14.
Apply mlookup_imp_mindom with x1.
Auto.
Apply H4.
Auto.
Apply lt_S_n.
Auto.
Cut (lt (S l) (S x))\/l=x.
Intros.
Inversion_clear H3.
Right.
Auto.
Left; Auto.
Cut (lt l x)\/l=x.
Intros.
Inversion_clear H3.
Left.
Apply lt_n_S.
Auto.
Right; Auto.
Cut (le l x).
Intros.
Apply le_lt_or_eq.
Auto.
Apply le_S_n.
Auto.
Apply le_or_lt.
Inversion H1.
Replace n with x.
Auto.
Apply (msize_eq ? ? h x n).
Auto.
Auto.
Save.




Lemma tr_extend_layout_aux_eq_preserv
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h'))
    (l:label; hv:HeapVal)
    h'=(melem ? ? l hv h) ->
    (msize ? ? h l) ->
    (Layout_aux h l)=(Layout_aux h' l).
Proof.
Induction h.
Intros.
Inversion H0.
Rewrite H.
Rewrite <- H2.
Unfold Layout_aux.
Unfold blt_nat.
Auto.
Intros.
Rewrite H0.
Unfold Layout_aux.
Replace (blt_nat l l) with false.
Auto.
Apply sym_eq.
Apply blt_eq_false.
Save.



Lemma tr_extend_layout_aux_preserv
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h'))
    (l,n:label; hv:HeapVal)
    h'=(melem ? ? l hv h) ->
    (msize ? ? h l) ->
    (lt n l) ->
    (Layout_aux h n)=(Layout_aux h' n).
Proof.
Induction h.
Intros.
Inversion H0.
Rewrite <- H3 in H1.
Inversion H1.
Intros.
Clear H.
Apply sym_eq.
Rewrite H0.
Unfold 1 Layout_aux.
Replace (blt_nat l n) with false.
Auto.
Apply sym_eq.
Apply blt_false.
Auto.
Save.


Lemma tr_extend_layout_preserv
  : (h,h':(Map label HeapVal); wfh:(mWF ? ? h); wfh':(mWF ? ? h'))
    (l,n:label; hv:HeapVal)
    (mextend ? ? h l hv h') ->
    (msize ? ? h l) ->
    (lt n l) ->
    (Layout (wfmap ? ? h' wfh') n)=(Layout (wfmap ? ? h wfh) n).
Proof.
Intros h.
Elim h.
Intros.
Inversion H0.
Rewrite <- H3 in H1. 
Inversion H1.
Clear h.
Intros a b h.
Intros.
Unfold Layout; Unfold unwrapMap.
Inversion H0.
Unfold 1 Layout_aux.
Replace (blt_nat l n) with false.
Unfold Layout_aux.
Auto.
Apply sym_eq.
Apply blt_false.
Auto.
Save.



Lemma tr_extend_heap_uninitwordlist
  : (Ts:(list Omega); H:Heap; M:Mem; l:label)
    (TrWordValList (Layout H) (makeUninitTup Ts) M l).
Proof.
Induction Ts.
Intros.
Constructor.
Intros.
Unfold makeUninitTup.
Fold makeUninitTup.
Apply trwvl_cons with (M l0).
Constructor.
Auto.
Apply H.
Save.



Lemma tr_extend_instr_preserv
  : (i:Instr; I:InstrSeq; R:RegFileTy; T:APTy;
    _i:_Instr; l:label; Ts:(list Omega); 
     h,h':(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? h'))
    (G,G':HeapTy)
    h'=(melem ? ? l (tuple (makeUninitTup Ts)) h) ->
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? h' wf') G') ->
    (WFInstrSeq G R T (iseq i I)) ->
    (TrInstr (Layout_aux h) i _i) ->
    (TrInstr (Layout_aux h') i _i).
Proof.
Induction i.
Intros.
Inversion H3.
Constructor.
Auto.
Auto.
Auto.
Intros.
Inversion H3.
Constructor.
Auto.
Auto.
Intros.
Inversion H3.
Constructor.
Auto.

Intros. Inversion H3. (* bgt *)
Replace (Layout_aux h n) with (Layout_aux h' n).
Constructor.
Auto.
Auto.
Apply sym_eq.
Apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
Auto.
Auto.
Auto.
Cut (msize ? ? h' (S l)).
Intros.
Rewrite H in H11.
Inversion H11.
Auto.
Inversion H1.
Unfold unwrapMap in H12.
Inversion_clear H12.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H21.
Replace (S l) with x.
Auto.
Apply (msize_eq ? ? h').
Auto.
Rewrite H.
Constructor.
Rewrite H in H22.
Inversion H22.
Auto.
Inversion H2.
Inversion H0.
Inversion_clear H24.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H30.
Inversion_clear H31.
Inversion_clear H32.
Inversion_clear H33.
Replace l with x.
Apply H29 with (codety G'0 T).
Auto.
Inversion H1.
Inversion_clear H35.
Inversion_clear H38.
Inversion_clear H39.
Inversion_clear H40.
Inversion_clear H41.
Inversion_clear H42.
Inversion_clear H43.
Inversion_clear H44.
Apply eq_add_S.
Replace (S x) with x0.
Replace (S l) with x0.
Auto.
Apply (msize_eq ? ? h').
Auto.
Unfold unwrapMap in H45.
Rewrite H in H45.
Rewrite H.
Constructor.
Inversion H45.
Auto.
Apply (msize_eq ? ? h').
Auto.
Rewrite H.
Constructor.
Auto.


Intros.  (* bump *)
Inversion H3.
Constructor.

Intros. Inversion H3. (* fold *)
Constructor.
Auto.
Auto.

Intros. Inversion H3. (* ld *)
Constructor.
Auto.
Auto.
Intros. Inversion H3. (* mov *)
Constructor.
Auto.
Auto.
Intros. Inversion H3. (* movi *)
Constructor.
Auto.
Intros. (* movl *)
Inversion H3.
Replace (Layout_aux h n) with (Layout_aux h' n).
Constructor.
Auto.
Apply sym_eq.
Apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
Auto.
Auto.
Auto.
Cut (msize ? ? h' (S l)).
Intros.
Rewrite H in H9.
Inversion H9.
Auto.
Inversion H1.
Inversion_clear H10.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Replace (S l) with x.
Auto.
Apply sym_eq.
Apply (msize_eq ? ? h').
Rewrite H.
Constructor.
Unfold unwrapMap in H20.
Rewrite H in H20.
Inversion H20.
Auto.
Auto.

Inversion H2.
Inversion H0.
Inversion_clear H20.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Replace l with x.
Apply H25 with t.
Auto.
Clear H30.
Inversion H1.
Inversion_clear H30.
Inversion_clear H33.
Inversion_clear H34.
Inversion_clear H35.
Inversion_clear H36.
Inversion_clear H37.
Inversion_clear H38.
Inversion_clear H39.
Apply eq_add_S.
Replace (S x) with x0.
Replace (S l) with x0.
Auto.
Apply sym_eq.
Apply (msize_eq ? ? h').
Unfold unwrapMap in H40.
Rewrite H in H40.
Rewrite H.
Constructor.
Inversion H40.
Auto.
Auto.
Apply (msize_eq ? ? h').
Auto.
Rewrite H.
Constructor.
Auto.

Intros.
Inversion H3.
Constructor.
Auto.
Auto.

Intros. Inversion H3. (* unfold *)
Constructor.
Auto.
Auto.
Save.



Lemma tr_extend_instrseq_preserv
  : (i:InstrSeq; M:Mem; l:label; Ts:(list Omega); h,h':(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? h'))
    (m:RegFileTy; a:APTy)
    (D:Word)
    (G,G':HeapTy)
    h'=(melem ? ? l (tuple (makeUninitTup Ts)) h) ->
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? h' wf') G') ->
    (WFInstrSeq G m a i) ->
    (TrInstrSeq (Layout_aux h)  i M D) ->
    (TrInstrSeq (Layout_aux h') i M D).
Proof.
Induction i.
Intros.
Inversion H4.
Apply tris_iseq with _i.
Apply (tr_extend_instr_preserv i0 i1 m a _i l Ts h h' wf wf' G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Inversion H3.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 (used n) (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' m a (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' m fresh (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). 
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). (* ld *)
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). 
Auto.
Auto.
Auto.
Auto.
Auto.
Apply (H M l Ts h h' wf wf' G'0 a (S D) G G'). 
Auto.
Auto.
Auto.
Auto.
Auto.

Intros. 
Constructor.
Inversion H3.
Replace (Layout_aux h' n) with (Layout_aux h n).
Auto.
Apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
Auto.
Auto.
Auto.
Cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
Intros.
Apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf H9 G G').
Auto.
Constructor.
Inversion H1.
Unfold unwrapMap.
Rewrite <- H.
Unfold unwrapMap in H11.
Auto.
Apply eq_map_wf with h'.
Auto.
Auto.
Inversion H2.
Inversion H0.
Inversion_clear H16.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H21.
Inversion_clear H22.
Replace l with x.
Apply H21 with (codety G'0 a).
Auto.
Unfold unwrapMap in H16.
Apply (msize_eq ? ? h x l).
Auto.
Cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
Intros.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf H22 G G'.
Auto.
Constructor.
Inversion H1.
Unfold unwrapMap.
Rewrite <- H.
Unfold unwrapMap in H25.
Auto.
Apply eq_map_wf with h'.
Auto.
Auto.

Intros. Inversion H3. (* jmp *)
Apply tris_jmp with _r.
Auto.
Auto.
Save.



Lemma wordval_strip_not_wfold
  : (v,v':WordVal) v'=(stripWV v) -> ~(EX w | (EX t | v'=(wfold w t))).
Proof.
Induction v.
Intros.
Intro.
Inversion_clear H0.
Inversion_clear H1.
Rewrite H0 in H.
Discriminate H.
Intros.
Intro.
Inversion_clear H0.
Inversion_clear H1.
Rewrite H0 in H.
Discriminate H.
Intros.
Intro.
Inversion_clear H0.
Inversion_clear H1.
Rewrite H0 in H.
Discriminate H.
Intros.
Unfold stripWV in H0.
Fold stripWV in H0.
Apply H.
Auto.
Save.



Lemma tr_wordval_strip
  : (L:LayoutF; v:WordVal; w:Word)
    (TrWordVal L v w) ->
    (EX v' | v'=(stripWV v) /\ (TrWordVal L v' w)).
Proof.
Induction v.
Intros.
Split with (wl n).
Auto.
Intros.
Split with (wi n).
Auto.
Intros.
Split with (wuninit o).
Auto.
Intros.
Inversion H0.
Cut (EX v'':? | v''=(stripWV w)/\(TrWordVal L v'' w0)).
Intros.
Inversion_clear H6.
Split with x.
Auto.
Apply H.
Auto.
Save.


Lemma tr_wordval_strip_alt
  : (L:LayoutF; v:WordVal; w:Word)
    (TrWordVal L v w) ->
    (TrWordVal L (stripWV v) w).
Proof.
Intros.
Cut (EX v':? | v'=(stripWV v)/\(TrWordVal L v' w)).
Intros.
Inversion_clear H0.
Inversion_clear H1.
Rewrite <- H0.
Auto.
Apply tr_wordval_strip.
Auto.
Save.


Lemma tr_strip_wordval
  : (L:LayoutF; v,v':WordVal; w:Word)
    v'=(stripWV v) ->
    (TrWordVal L v' w) ->
    (TrWordVal L v w).
Proof.
Induction v.
Intros.
Rewrite H in H0.
Auto.
Intros.
Rewrite H in H0.
Auto.
Intros.
Rewrite H in H0.
Auto.
Intros.
Constructor.
Apply H with v'.
Auto.
Auto.
Save.



Lemma tr_extend_wordval_preserv_aux
  : (w:Word; l:label; Ts:(list Omega); h,h':(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? h'))
    (wv',wv:WordVal)
    (HT,HT':HeapTy)
    h'=(melem ? ? l (tuple (makeUninitTup Ts)) h) ->
    (WFHeap (wfmap ? ? h wf) HT) ->
    (WFHeap (wfmap ? ? h' wf') HT') ->
    wv'=(stripWV wv) ->
    ( (l,n:nat) (stripWV wv)=(wl l) -> (msize ? ? (htunwrap HT) n) -> (lt l n))
    ->
    (TrWordVal (Layout_aux h) wv' w) ->
    (TrWordVal (Layout_aux h') wv' w).
Proof.
Induction wv'.
Intros.
Inversion H4.
Replace (Layout_aux h n) with (Layout_aux h' n).
Constructor.
Apply sym_eq.
Apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
Auto.
Auto.
Auto.
Inversion H1.
Inversion_clear H9.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Unfold unwrapMap in H19.
Rewrite H in H19.
Inversion H19.
Auto.
Apply H3.
Auto.
Induction HT.
Apply wfh_ht_size_eq with h wf m0.
Auto.
Inversion H1.
Inversion_clear H9.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Unfold unwrapMap in H19.
Rewrite H in H19.
Inversion H19.
Auto.
Intros.
Inversion H4.
Constructor.
Intros.
Constructor.
Intros.
Absurd (EX w:? | (EX t:? | (stripWV wv)=(wfold w t))).
Apply wordval_strip_not_wfold with wv.
Auto.
Split with w0.
Split with o.
Auto.
Save.


(*    (WFRegFile HT R G) ->
    (mlookup ? ? R r wv) -> *)



Lemma tr_extend_wordvali_preserv
  : (w:Word; l:label; Ts:(list Omega); h,h':(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? h'))
    (wv:WordVal; t:Omega; i:initflag)
    (G,G':HeapTy)
    h'=(melem ? ? l (tuple (makeUninitTup Ts)) h) ->
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? h' wf') G') ->
    (WFWordValinit G wv t i) ->
    (TrWordVal (Layout_aux h) wv w) ->
    (TrWordVal (Layout_aux h') wv w).
Proof.
Induction wv.
Intros.
Inversion H3.
Replace (Layout_aux h n) with (Layout_aux h' n).
Constructor.
Apply sym_eq.
Apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
Auto.
Auto.
Auto.
Cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
Intros wf''.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf'' G G'.
Auto.
Constructor.
Inversion H1.
Inversion_clear H8.
Split with x.
Unfold unwrapMap.
Rewrite <- H.
Unfold unwrapMap in H11.
Auto.
Apply eq_map_wf with h'.
Auto.
Auto.
Inversion H2.
Inversion H7.
Inversion H0.
Inversion_clear H18.
Inversion_clear H21.
Inversion_clear H22.
Inversion_clear H23.
Inversion_clear H24.
Replace l with x.
Apply H23 with t'.
Auto.
Unfold unwrapMap in H13.
Cut (msize ? ? h l).
Intros.
Apply (msize_eq ? ? h x l).
Auto.
Auto.
Cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
Intros.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf H24 G G'.
Auto.
Inversion H1.
Constructor.
Unfold unwrapMap.
Unfold unwrapMap in H27.
Rewrite <- H.
Auto.
Apply eq_map_wf with h'.
Auto.
Auto.
Intros.
Inversion H3.
Constructor.
Intros.
Constructor.

Intros.
Constructor.
Inversion H4.
Inversion H3.
Inversion H10.
Apply H with t' i G G'.
Auto.
Auto.
Auto.
Constructor.
Auto.
Auto.
Save.



Lemma tr_extend_wordvallist_preserv
  : (L:(list WordVal); tl:(list Omega); il:(list bool))
    (M:Mem; l:label; Ts:(list Omega); h,h':(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? h'))
    (D:Word)
    (G,G':HeapTy)
    h'=(melem ? ? l (tuple (makeUninitTup Ts)) h) ->
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? h' wf') G') ->
    (WFWordValinitList G L tl il) ->
    (TrWordValList (Layout_aux h)  L M D) ->
    (TrWordValList (Layout_aux h') L M D).
Proof.
Induction L.
Intros.
Constructor.
Intros.
Inversion H4.
Apply trwvl_cons with w.
Inversion H3.
Apply (tr_extend_wordvali_preserv w l0 Ts h h' wf wf' a t i G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Inversion H3.
Apply (H tl0 il0 M l0 Ts h h' wf wf' (S D) G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Save.



Lemma tr_extend_heapval_preserv_aux
  : (M:Mem; l:label; Ts:(list Omega); h,h':(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? h'))
    (n:nat; hv:HeapVal; t:Omega)
    (G,G':HeapTy)
    h'=(melem ? ? l (tuple (makeUninitTup Ts)) h) ->
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? h' wf') G') ->
    (WFHeapVal G hv t) ->
    (lt n l) ->
    (TrHeapVal (Layout_aux h)  hv M (Layout_aux h n)) ->
    (TrHeapVal (Layout_aux h') hv M (Layout_aux h' n)).
Proof.
Induction hv.
Intros.
Inversion H4.
Constructor.
Inversion H2.
Replace (Layout_aux h' n) with (Layout_aux h n).
Apply (tr_extend_wordvallist_preserv l0 tl il M l Ts h h'
         wf wf' (Layout_aux h n) G G').
Auto. Auto. Auto. Auto. Auto. Auto.
Apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
Auto. Auto. Auto.
Cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
Intros wf''.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf'' G G'.
Auto.
Inversion H1.
Constructor.
Inversion_clear H15.
Split with x.
Unfold unwrapMap.
Unfold unwrapMap in H18.
Rewrite <- H. Auto.
Apply eq_map_wf with h'.
Auto. Auto. Auto.
Intros.
Inversion H4.
Constructor.
Apply (tr_extend_instrseq_preserv i M l Ts h h' wf wf' m a (Layout_aux h' n) G G').
Auto. Auto. Auto. Inversion H2. Auto. Auto. 
Replace (Layout_aux h' n) with (Layout_aux h n).
Auto.
Apply tr_extend_layout_aux_preserv with l (tuple (makeUninitTup Ts)).
Auto. Auto. Auto.
Cut (mWF nat HeapVal (melem nat HeapVal l (tuple (makeUninitTup Ts)) h)).
Intros wf''.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf'' G G'.
Auto.
Constructor.
Inversion H1.
Inversion_clear H13.
Split with x.
Unfold unwrapMap.
Rewrite <- H.
Unfold unwrapMap in H16.
Auto.
Apply eq_map_wf with h'.
Auto.
Auto.
Auto.
Save.



Lemma tr_extend_heapval_preserv
  : (M:Mem; l:label; Ts:(list Omega))
    (H,H':Heap; G,G':HeapTy)
    (n:nat; hv:HeapVal; t:Omega)
    (WFHeap H G) ->
    (WFHeap H' G') ->
    (hextend (hunwrap H) l (tuple (makeUninitTup Ts)) (hunwrap H')) ->
    (WFHeapVal G hv t) ->
    (lt n l) ->
    (TrHeapVal (Layout H) hv M (Layout H n)) ->
    (TrHeapVal (Layout H') hv M (Layout H' n)).
Proof.
Intros.
Induction H.
Induction H'.
Unfold Layout unwrapMap.
Apply (tr_extend_heapval_preserv_aux M l Ts m m1 m0 m2 n hv t G G'). Auto.
Inversion H2.
Unfold unwrapMap in H.
Auto. Auto. Auto. Auto. Auto. Auto.
Save.




Lemma tr_heap_extend_emptytup_aux
  : (M:Mem; l:label; Ts:(list Omega); h,h':(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? h'); 
     wf'':(mWF ? ? (melem ? ? l (tuple (makeUninitTup Ts)) h)); 
     G,G':HeapTy)
    h'=(melem ? ? l (tuple (makeUninitTup Ts)) h) ->
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? h' wf') G') ->
    (TrHeap (Layout (wfmap ? ? h wf))   (wfmap ? ? h wf) M) ->
    (TrHeap (Layout (wfmap ? ? (melem ? ? l (tuple (makeUninitTup Ts)) h) wf'')) (wfmap ? ? (melem ? ? l (tuple (makeUninitTup Ts)) h) wf'') M) ->
    (TrHeap (Layout (wfmap ? ? h' wf')) (wfmap ? ? h' wf') M).
Proof.
Intros.
Unfold Layout unwrapMap in H3.
Inversion H3.
Apply trheap with s.
Unfold unwrapMap.
Unfold unwrapMap in H5.
Rewrite H.
Auto.
Unfold unwrapMap.
Unfold Layout unwrapMap.
Rewrite H.
Auto.
Save.



Lemma tr_heap_extend_emptytup
  : (M:Mem; l:label; Ts:(list Omega); h:(Map label HeapVal))
    (wf:(mWF ? ? h); wf':(mWF ? ? (melem ? ? l (tuple (makeUninitTup Ts)) h));
     G,G':HeapTy)
    (WFHeap (wfmap ? ? h wf) G) ->
    (WFHeap (wfmap ? ? (melem ? ? l (tuple (makeUninitTup Ts)) h) wf') G') ->
    (TrHeap (Layout (wfmap ? ? h wf)) (wfmap ? ? h wf) M) ->
    (TrHeap (Layout (wfmap ? ? (melem ? ? l (tuple (makeUninitTup Ts)) h) wf'))
      	    (wfmap ? ? (melem ? ? l (tuple (makeUninitTup Ts)) h) wf') M).
Proof.
Intros.
Inversion H0.
Unfold unwrapMap in H3.
Inversion_clear H3.
Apply trheap with x.
Inversion_clear H6.
Auto.
Unfold unwrapMap.
Intros.
Cut (lt n x).
Intros.
Inversion_clear H6.
Inversion H8.
Clear H12.
Clear m.
Clear H11.
Clear H6.
Clear a.
Clear b.
Move H7 after H13.
Rewrite <- H10 in H7.
Cut n=n0\/(lt n n0).
Intros.
Inversion_clear H6.
Cut n0=l.
Intros.
Rewrite H6 in H11.
Replace h0 with (tuple (makeUninitTup Ts)).
Constructor.
Apply tr_extend_heap_uninitwordlist.
Rewrite H11 in H3.
Inversion H3.
Auto.
Absurd ~l=l.
Auto.
Auto.
Cut (msize ? ? h l).
Intros.
Apply (msize_eq ? ? h n0 l).
Auto.
Auto.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wf wf' G G'.
Auto.
Auto.
Cut (EX t:? | (WFHeapVal G h0 t)).
Intros.
Elim H6; Clear H6; Intros t; Intros.
Apply tr_extend_heapval_preserv
 with n0 Ts (wfmap nat HeapVal h wf) G G' t.
Auto.
Auto.
Unfold unwrapMap.
Replace n0 with l.
Constructor.
Inversion wf'.
Auto.
Cut (msize ? ? h l).
Intros.
Apply (msize_eq ? ? h l n0).
Auto.
Auto.
Apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf wf' G G').
Auto.
Auto.
Auto.
Auto.
Inversion H1.
Apply H15.
Inversion H3.
Cut l=n0.
Intros.
Rewrite H24 in H19.
Absurd n0=n.
Cut ~n=n0; Intros; Auto.
Apply lt_neq.
Auto.
Auto.
Cut (msize ? ? h l).
Intros.
Apply (msize_eq ? ? h l n0).
Auto.
Auto.
Apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf wf' G G').
Auto.
Auto.
Auto.
Cut (EX t:? | (htlookup (htunwrap G) n t)).
Intros.
Elim H6; Clear H6; Intros t H6.
Split with t.
Inversion H.
Inversion_clear H14.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H21.
Inversion_clear H22.
Inversion_clear H23.
Apply H22 with n.
Unfold unwrapMap.
Inversion H3.
Absurd n=l.
Apply lt_neq.
Replace l with n0.
Auto.
Cut (msize ? ? h l).
Intros.
Apply (msize_eq ? ? h n0 l); Auto.
Apply (heap_label_eq_size l (tuple (makeUninitTup Ts)) h wf wf' G G').
Auto.
Auto.
Auto.
Auto.
Auto.
Inversion H.
Inversion_clear H12.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H21.
Apply H20.
Replace x0 with n0.
Auto.
Unfold unwrapMap in H12.
Apply (msize_eq ? ? h n0 x0).
Auto.
Auto.
Cut n0=n\/(lt n n0).
Intros.
Inversion_clear H6.
Left; Auto.
Right; Auto.
Apply lt_S_eq_or_lt.
Auto.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H9.
Apply H8 with h0.
Auto.
Save.



Lemma heaplookup_wfheapval 
  : (H:Heap; HT:HeapTy; l:nat; h:HeapVal)
    (WFHeap H HT) ->
    (hlookup (hunwrap H) l h) ->
    (EX t | (WFHeapVal HT h t)).
Proof.
Intro.
Elim H; Clear H; Intros h wfh.
Intro.
Elim HT; Clear HT; Intros ht wfht.
Intros.
Inversion H.
Unfold unwrapMap in H2.
Cut (EX t:? | (mlookup ? ? ht l t)).
Intros.
Inversion_clear H5.
Split with x.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Apply H11 with l.
Auto.
Auto.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Apply H9.
Apply H6 with h0.
Auto.
Save.





Lemma Preservation_alloc
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd:Reg; Ts:(list Omega); I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (alloc rd Ts) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (alloc rd Ts) I) M pc).
Intros.
Inversion H25.
Rewrite <- H30 in H28.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with (Layout H').
Auto.
Rewrite <- H44 in H4.
Inversion H4.
Clear H60.
Cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
Intros.
Rewrite H48.
Apply (tr_heap_extend_emptytup_aux M a Ts h h' wfh wfh' H60 HT HT0).
Inversion H50.
Auto.
Rewrite <- H47.
Auto.
Rewrite <- H48.
Auto.
Rewrite <- H47.
Rewrite <- H19.
Auto.
Apply (tr_heap_extend_emptytup M a Ts h wfh H60 HT HT0).
Rewrite <- H47.
Auto.
Inversion H50.
Inversion H56.
Constructor.
Inversion_clear H67.
Split with x.
Unfold unwrapMap.
Rewrite H48 in H70.
Unfold unwrapMap in H70.
Rewrite <- H65 in H70.
Auto.
Rewrite <- H47.
Rewrite <- H19.
Auto.
Apply eq_map_wf with h'.
Auto.
Inversion H50.
Auto.
Unfold updatereg.
Constructor.
Intros.
Cut _r=_rd\/~_r=_rd.
Intros.
Inversion_clear H53.
Rewrite H54.
Replace (beq_reg _rd _rd) with true.
Cut v=(wl a).
Intros.
Replace (_R _r31) with (Layout H' a).
Replace (plus (Layout H' a) O) with (Layout H' a).
Rewrite H53.
Constructor.
Apply sym_eq.
Apply plus_n_O_n.
Inversion H23.
Rewrite H19 in H55.
Replace (Layout H' a) with (Layout H a).
Auto.
Unfold Layout.
Unfold unwrapMap.
Rewrite H47.
Rewrite H48.
Apply (tr_extend_layout_aux_eq_preserv h h' wfh wfh' a
        (tuple (makeUninitTup Ts))).
Inversion H50.
Auto.
Rewrite <- H44 in H4.
Inversion H4.
Clear H68.
Cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
Intros.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wfh H68 HT HT0.
Rewrite <- H47.
Auto.
Inversion H64.
Rewrite H48 in H70.
Unfold unwrapMap in H70.
Inversion H50.
Rewrite <- H77 in H70.
Constructor.
Auto.
Apply eq_map_wf with h'.
Auto.
Inversion H50.
Auto.
Cut r=rd.
Intros.
Cut (mlookup ? ? R' rd (wl a)).
Intros.
Apply (mlookup_eq ? ? rd v (wl a) R').
Apply eqorneq_reg.
Rewrite <- H53.
Auto.
Auto.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Auto.
Rewrite H54 in H52.
Apply trregf_eq_eq with _rd.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_eq_true.
Replace (beq_reg _r _rd) with false.
Inversion H22.
Rewrite H19 in H53.
Cut (TrWordVal (Layout_aux h) v (_R _r)).
Intros.
Rewrite H48.
Unfold Layout unwrapMap.
Apply tr_strip_wordval with (stripWV v).
Auto.
Rewrite <- H44 in H4.
Inversion H4.
Apply (tr_extend_wordval_preserv_aux (_R _r) a Ts h h' wfh wfh'
        (stripWV v) v HT HT0).
Inversion H50.
Auto.
Rewrite <- H47.
Auto.
Rewrite <- H48.
Auto.
Auto.
Inversion H11.
Intros.
Apply H70 with r v.
Apply mupdext_preserv_rev with R' rd (wl a).
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply tr_wordval_strip_alt.
Auto.
Unfold Layout unwrapMap in H53.
Rewrite H47 in H53.
Apply H53 with r.
Apply mupdext_preserv_rev with R' rd (wl a).
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Apply eqorneq_tisreg.
Constructor.
Unfold updatereg.
Replace (beq_reg _r31 _rd) with false.
Inversion H23.
Rewrite H19 in H51.
Rewrite H47 in H51.
Rewrite H48.
Unfold Layout unwrapMap.
Unfold Layout unwrapMap in H51.
Replace (Layout_aux h' a) with (Layout_aux h a).
Auto.
Apply tr_extend_layout_aux_eq_preserv with (tuple (makeUninitTup Ts)).
Auto.
Auto.
Inversion H50.
Auto.
Rewrite <- H44 in H4.
Inversion H4.
Clear H64.
Cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
Intros.
Apply heap_label_eq_size with (tuple (makeUninitTup Ts)) wfh H64 HT HT0.
Rewrite <- H47.
Auto.
Inversion H60.
Rewrite H48 in H66.
Unfold unwrapMap in H66.
Inversion H50.
Rewrite <- H73 in H66.
Constructor.
Auto.
Apply eq_map_wf with h'.
Auto.
Inversion H50.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_rd=_r31.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Elim H24; Clear H24; Intros l1 H24.
Elim H24; Clear H24; Intros GO H24.
Elim H24; Clear H24; Intros T H24.
Elim H24; Clear H24; Intros I'0 H24.
Elim H24; Clear H24; Intros n' H24.
Inversion_clear H24.
Inversion_clear H52.
Inversion_clear H53.
Split with l1.
Split with GO.
Split with T.
Split with I'0.
Rewrite <- H21 in H2.
Clear H21.
Clear S.
Split with (S n').
Split.
Rewrite H48.
Unfold unwrapMap.
Apply mextend_preserv with h a (tuple (makeUninitTup Ts)).
Auto.
Rewrite H47 in H51.
Auto.
Cut ~a=l1.
Auto.
Apply (mwf_anyelem_neq ? ? a l1 (tuple (makeUninitTup Ts)) h).
Inversion H50.
Apply eq_map_wf with h'.
Auto.
Auto.
Apply mlookup_imp_mindom with (code GO T I'0).
Rewrite H47 in H51.
Auto.
Split.
Apply tr_instrseq_subdepth_S with (alloc rd Ts).
Auto.
Cut (TrInstrSeq (Layout H') I'0 M (Layout H' l1)).
Intros.
Split.
Auto.
Replace (plus (Layout H' l1) (S n')) with (plus (S n') (Layout H' l1)).
Unfold plus.
Fold plus.
Apply eq_S.
Replace (plus n' (Layout H' l1)) with (plus (Layout H' l1) n').
Replace (Layout H' l1) with (Layout H l1).
Rewrite <- H19.
Auto.
Unfold Layout unwrapMap.
Rewrite H47.
Rewrite H48.
Apply tr_extend_layout_aux_preserv with a (tuple (makeUninitTup Ts)).
Auto.
Auto.
Inversion H50.
Auto.
Rewrite <- H44 in H4.
Inversion H4.
Cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
Intros.
Apply (heap_label_eq_size a (tuple (makeUninitTup Ts)) h wfh H64 HT HT0).
Rewrite <- H47.
Auto.
Rewrite H48 in H59.
Inversion H59.
Inversion H50.
Unfold unwrapMap in H66.
Rewrite <- H73 in H66.
Constructor.
Auto.
Apply eq_map_wf with h'.
Auto.
Inversion H50.
Auto.
Inversion H10.
Inversion_clear H55.
Inversion_clear H58.
Inversion_clear H59.
Inversion_clear H60.
Inversion_clear H61.
Replace a with x.
Apply H59 with (code GO T I'0).
Auto.
Inversion H13.
Rewrite <- H67 in H12.
Inversion H12.
Apply (msize_eq ? ? (htunwrap HT)).
Auto.
Auto.
Apply plus_sym.
Apply plus_sym.
Rewrite H19 in H52.
Unfold Layout unwrapMap.
Unfold Layout unwrapMap in H52.
Rewrite H48.
Rewrite H47 in H52.
Cut (EX t:? | (WFHeapVal HT (code GO T I'0) t)).
Intros.
Inversion_clear H21.
Inversion_clear H53.
Rewrite <- H44 in H4.
Inversion H4.
Apply (tr_extend_instrseq_preserv I'0 M a Ts h h' wfh wfh' GO T)
 with HT HT0.
Inversion H50.
Auto.
Rewrite <- H47.
Auto.
Rewrite <- H48.
Auto.
Auto.
Replace (Layout_aux h' l1) with (Layout_aux h l1).
Auto.
Apply tr_extend_layout_aux_preserv with a (tuple (makeUninitTup Ts)).
Auto.
Auto.
Inversion H50.
Auto.
Cut (mWF nat HeapVal (melem nat HeapVal a (tuple (makeUninitTup Ts)) h)).
Intros.
Apply (heap_label_eq_size a (tuple (makeUninitTup Ts)) h wfh H64 HT HT0).
Rewrite <- H47.
Auto.
Rewrite H48 in H59.
Inversion H59.
Unfold unwrapMap in H66.
Inversion H50.
Rewrite <- H73 in H66.
Constructor.
Auto.
Apply eq_map_wf with h'.
Auto.
Inversion H50.
Auto.
Inversion H10.
Inversion_clear H65.
Inversion_clear H68.
Inversion_clear H69.
Inversion_clear H70.
Inversion_clear H71.
Replace a with x0.
Apply H69 with (code GO T I'0).
Auto.
Apply (msize_eq ? ? (htunwrap HT)).
Auto.
Inversion H13.
Rewrite <- H77 in H12.
Inversion H12.
Auto.
Apply heaplookup_wfheapval with (wfmap ? ? h wfh) l1.
Rewrite <- H47.
Auto.
Rewrite <- H47.
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.




Lemma Preservation_bgt
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rs,rt:Reg; l:label; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (bgt rs rt l) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (bgt rs rt l) I) M pc).
Intros.
Inversion H25.
Rewrite H30.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Replace (blt_nat (_R _rt) (_R _rs)) with false.
Apply trprogram with L.
Auto.
Auto.
Auto.
Auto.
Apply tr_pc_Spc with (bgt rs rt l).
Auto.
Apply sym_eq.
Apply le_blt_false.
Replace (_R _rs) with rsval.
Replace (_R _rt) with rtval.
Auto.
Inversion H22.
Cut (TrWordVal L (wi rtval) (_R _rt)).
Intros.
Inversion H57.
Auto.
Apply H53 with rt.
Auto.
Auto.
Cut (TrWordVal L (wi rsval) (_R _rs)).
Intros.
Inversion H53.
Auto.
Inversion H22.
Apply H53 with rs.
Auto.
Auto.
Replace (blt_nat (_R _rt) (_R _rs)) with true.
Apply trprogram with L.
Auto.
Auto.
Auto.
Auto.
Split with l.
Split with G0.
Split with T.
Split with I''.
Split with O.
Split.
Auto.
Split.
Constructor.
Split.
Inversion H20.
Cut (TrHeapVal L (code G0 T I'') M (L l)).
Intros.
Inversion H60.
Auto.
Apply H56.
Auto.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Replace (_R _rt) with rtval.
Replace (_R _rs) with rsval.
Unfold lt.
Apply gt_le_S.
Auto.
Cut (TrWordVal L (wi rsval) (_R _rs)).
Intros.
Inversion H54.
Auto.
Inversion H22.
Apply H54 with rs.
Auto.
Auto.
Cut (TrWordVal L (wi rtval) (_R _rt)).
Intros.
Inversion H54.
Auto.
Inversion H22.
Apply H54 with rt.
Auto.
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.





Lemma wf_wordlist_len_eq
  : (HT:HeapTy; V:(list WordVal); tl:(list Omega); ol:(list initflag))
    (WFWordValinitList HT  V tl ol) ->
    (length V)=(length tl).
Proof.
Induction V.
Intros.
Inversion H.
Auto.
Intros.
Inversion H0.
Unfold length.
Apply eq_S.
Apply H with il.
Auto.
Save.


Lemma wf_tuple_tupty_len_eq
  : (HT:HeapTy; V:(list WordVal); tl:(list Omega); ol:(list initflag))
    (WFHeapVal HT (tuple V) (tupty tl ol)) ->
    (length V)=(length tl).
Proof.
Intros.
Inversion H.
Apply wf_wordlist_len_eq with HT ol.
Auto.
Save.


Lemma layout_gt_eq
  : (h:(Map label HeapVal); wfh:(mWF ? ? h); a:label)
    ((b:label) (mindom ? ? h b) -> (lt b a)) ->
    (Layout_aux h a)=(Layout_aux h (S a)).
Proof.
Induction h.
Intros.
Auto.
Intros.
Unfold Layout_aux.
Replace (blt_nat a a0) with true.
Replace (blt_nat a (S a0)) with true.
Fold Layout_aux.
Cut (Layout_aux m a0)=(Layout_aux m (S a0)).
Intros.
Rewrite H1.
Auto.
Apply H.
Inversion wfh.
Auto.
Intros.
Apply H0.
Apply minnext.
Inversion wfh.
Cut ~b0=a.
Intros.
Auto.
Apply mindom_mnotindom_neq with HeapVal m.
Auto.
Apply eqorneq_nat.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Cut (lt a a0).
Intros.
Apply lt_S.
Auto.
Apply H0.
Constructor.
Apply sym_eq.
Apply blt_lt_true.
Apply H0.
Constructor.
Save.


Lemma wfheap_next_mindom_imp_lt
  : (l,a,b:label; hv:HeapVal; h,H:(Map label HeapVal); wfh:(mWF ? ? H))
    (HT:HeapTy)
    (WFHeap (wfmap ? ? H wfh) HT) ->
    H=(melem ? ? l hv h) ->
    (msize ? ? h a) ->
    (mindom ? ? h b) -> (lt b a).
Proof.
Intros.
Inversion H0.
Inversion_clear H5.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Replace a with l.
Cut l=b\/(lt b l).
Intros.
Inversion_clear H14.
Absurd l=b.
Inversion wfh.
Rewrite H1 in H14.
Absurd (mempty nat HeapVal)=(melem nat HeapVal l hv h).
Discriminate.
Auto.
Replace l with a0.
Cut ~b=a0.
Intros.
Auto.
Apply mindom_mnotindom_neq with HeapVal m.
Auto.
Apply eqorneq_nat.
Replace m with h.
Auto.
Rewrite H1 in H18.
Injection H18.
Auto.
Auto.
Rewrite H1 in H18.
Injection H18.
Auto.
Auto.
Auto.
Apply lt_S_eq_or_lt.
Replace (S l) with x.
Cut (EX hv:? | (hlookup (hunwrap (wfmap nat HeapVal H wfh)) b hv)).
Intros.
Inversion_clear H14.
Apply H9 with x0.
Auto.
Unfold unwrapMap.
Apply mindom_imp_mlookup.
Rewrite H1.
Apply minnext.
Cut ~b=l.
Auto.
Apply mindom_mnotindom_neq with HeapVal h.
Inversion wfh.
Rewrite H1 in H14.
Absurd (mempty nat HeapVal)=(melem nat HeapVal l hv h).
Discriminate.
Auto.
Replace h with m.
Auto.
Rewrite H1 in H17.
Injection H17.
Auto.
Apply eqorneq_nat.
Auto.
Inversion wfh.
Rewrite H1 in H14.
Absurd (mempty nat HeapVal)=(melem nat HeapVal l hv h).
Discriminate.
Auto.
Replace h with m.
Replace l with a0.
Auto.
Rewrite H1 in H17.
Injection H17.
Auto.
Rewrite H1 in H17.
Injection H17.
Auto.
Auto.
Apply (msize_eq ? ? H).
Auto.
Unfold unwrapMap in H15.
Rewrite H1 in H15.
Inversion H15.
Rewrite H1.
Constructor.
Auto.
Cut (S a)=x.
Intros.
Cut (S l)=x.
Intros.
Rewrite <- H14 in H16.
Apply eq_add_S.
Auto.
Unfold unwrapMap.
Auto.
Apply (msize_eq ? ? H).
Rewrite H1.
Unfold unwrapMap in H15.
Rewrite H1 in H15.
Inversion H15.
Constructor.
Auto.
Auto.
Apply (msize_eq ? ? H).
Rewrite H1.
Constructor.
Auto.
Auto.
Save.


Lemma wfh_h_size_eq
  : (h:(Map label HeapVal); wfh:(mWF ? ? h);
     ht:(Map label Omega); wfht:(mWF ? ? ht); l:nat)
    (WFHeap (wfmap ? ? h wfh) (wfmap ? ? ht wfht)) ->
    (htsize (htunwrap (wfmap ? ? ht wfht)) l) ->
    (hsize (hunwrap (wfmap ? ? h wfh)) l).
Proof.
Intros.
Inversion H.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H6.
Replace l with x.
Unfold unwrapMap in H2.
Auto.
Unfold unwrapMap in H5.
Apply (msize_eq nat Omega ht).
Auto.
Auto.
Save.



Lemma bump_aux
  : (H:Heap; HT:HeapTy; a,n:nat; w:Word; tl:(list Omega); ol:(list bool))
    (WFHeap H HT) ->
    (htsize (htunwrap HT) (S a)) ->
    (Layout H a)=w ->
    (WFWordVal HT (wl a) (tupty tl ol)) ->
    n=(length tl) ->
    (Layout H (S a))=(plus w n).
Proof.
Intros H.
Elim H; Clear H; Intros H wfH; Intros.
Cut (msize ? ? H (S a)).
Intros.
Inversion H5.
Unfold Layout unwrapMap.
Rewrite <- H7.
Cut a0=a.
Intros.
Rewrite H9.
Unfold Layout_aux.
Replace (blt_nat a (S a)) with true.
Fold Layout_aux.
Replace (Layout_aux m (S a)) with (Layout_aux m a).
Unfold Layout unwrapMap in H2.
Rewrite <- H7 in H2.
Rewrite <- H2.
Rewrite H9.
Unfold Layout_aux.
Replace (blt_nat a a) with false.
Fold Layout_aux.
Replace (HV_size b) with n.
Auto.
Inversion H3.
Cut (EX ol':? | t'=(tupty tl ol')).
Intros.
Elim H15; Clear H15; Intros ol'; Intros.
Rewrite H15 in H11.
Cut (EX V:? | (mlookup ? ? H a (tuple V))).
Intros.
Elim H16; Clear H16; Intros V H16.
Replace b with (tuple V).
Unfold HV_size.
Cut (WFHeapVal HT (tuple V) (tupty tl ol')).
Intros.
Replace (length V) with (length tl).
Auto.
Apply sym_eq.
Apply wf_tuple_tupty_len_eq with HT ol'.
Auto.
Inversion H0.
Inversion_clear H18.
Inversion_clear H21.
Inversion_clear H22.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H27.
Apply H26 with a.
Auto.
Auto.
Rewrite <- H7 in H16.
Rewrite H9 in H16.
Inversion H16.
Auto.
Absurd ~a=a; Auto.
Inversion H0.
Inversion_clear H17.
Inversion_clear H20.
Inversion_clear H21.
Inversion_clear H22.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Cut (EX h:? | (hlookup (hunwrap (wfmap ? ? H wfH)) a h)).
Intros.
Inversion H26.
Cut (WFHeapVal HT x0 (tupty tl ol')).
Intros.
Inversion H29.
Split with wl.
Rewrite H32.
Auto.
Apply H25 with a.
Auto.
Auto.
Apply H23.
Apply H22 with (tupty tl ol').
Auto.
Apply subtype_tupty with ol.
Auto.
Apply sym_eq.
Apply blt_eq_false.
Apply layout_gt_eq.
Inversion wfH.
Rewrite <- H10 in H7.
Absurd (melem nat HeapVal a0 b m)=(mempty nat HeapVal).
Discriminate.
Auto.
Replace m with m0.
Auto.
Rewrite <- H12 in H7.
Injection H7.
Auto.
Intros.
Apply (wfheap_next_mindom_imp_lt a0 a b0 b m H wfH HT).
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Apply lt_n_Sn.
Apply eq_add_S.
Inversion H0.
Inversion_clear H10.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H17.
Inversion_clear H18.
Inversion_clear H19.
Cut (S a0)=x.
Intros.
Rewrite H19.
Cut (S a)=x.
Intros.
Rewrite H21.
Auto.
Apply (msize_eq ? ? H).
Rewrite <- H7.
Constructor.
Auto.
Auto.
Apply (msize_eq ? ? H).
Unfold unwrapMap in H20.
Rewrite <- H7 in H20.
Rewrite <- H7.
Constructor.
Inversion H20.
Auto.
Auto.
Change (hsize (hunwrap (wfmap nat HeapVal H wfH)) (S a)) .
Induction HT.
Apply wfh_h_size_eq with m m0.
Auto.
Auto.
Save.




Lemma Preservation_bump
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	i:nat; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (bump i) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.

Cut (TrInstrSeq L (iseq (bump i) I) M pc). Intros.
Inversion H25.
Rewrite <- H30 in H28.
Inversion H28.
Clear H24.

Unfold updatereg.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Constructor.
Intros.
Cut _r=_r31\/~_r=_r31.
Intros.
Inversion_clear H46.
Rewrite H47 in H45.
Absurd ~_r31=_r31.
Auto.
Apply trreg_neq_31 with r.
Auto.
Replace (beq_reg _r _r31) with false.
Inversion H22.
Apply H46 with r.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Apply eqorneq_tisreg.
Constructor.
Replace (beq_reg _r31 _r31) with true.
Rewrite H19.
Inversion H23.
Inversion H13.
Rewrite <- H51 in H12.
Inversion H12.
Rewrite <- H48.
Replace l1 with (S a).
Apply sym_eq.
Apply bump_aux with HT tl ol.
Auto.
Auto.
Rewrite <- H19.
Auto.
Auto.
Auto.
Inversion H10.
Inversion_clear H61.
Inversion_clear H64.
Inversion_clear H65.
Clear H66.
Induction H.
Induction HT.
Replace (S a) with x.
Replace l1 with x.
Auto.
Apply (msize_eq ? ? m x l1).
Auto.
Auto.
Apply (msize_eq ? ? m1 x (S a)).
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_eq_true.
Rewrite <- H21 in H2.
Inversion H2.
Rewrite H19.
Rewrite <- H51.
Apply tr_pc_Spc with (bump i).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.




Lemma Preservation_fold
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd,rs:Reg; t:Omega; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (fold rd t rs) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (fold rd t rs) I) M pc).
Intros.
Inversion H25.
Rewrite H30.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Unfold updatereg.
Constructor.
Intros.
Compare _r _rd.
Intros.
Replace (beq_reg _r _rd) with true.
Replace (plus (_R _rs) O) with (plus O (_R _rs)).
Unfold plus.
Replace v with (wfold rsval t).
Constructor.
Inversion H22.
Apply H54 with rs.
Auto.
Auto.
Apply (mlookup_eq ? ? r (wfold rsval t) v R').
Apply eqorneq_reg.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Replace r with rd.
Auto.
Apply trregf_eq_eq with _r.
Rewrite e.
Auto.
Auto.
Auto.
Apply plus_sym.
Rewrite e.
Apply sym_eq.
Apply beq_reg_eq_true.
Intros.
Replace (beq_reg _r _rd) with false.
Inversion H22.
Apply H54 with r.
Apply mupdext_preserv_rev with R' rd (wfold rsval t).
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Constructor.
Unfold updatereg.
Replace (beq_reg _r31 _rd) with false.
Inversion H23.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_rd=_r31.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Apply tr_pc_Spc with (fold rd t rs).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.





Lemma Preservation_jmp
      	       	     : (St:State; H:Heap; R:RegFile; a:AP; r:Reg;
      	       	       	P,P':Program)
		       P = (H,(R,(a,(jmp r)))) ->
		       (WFProgram P) ->
		       (TrProgram P St) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step St)).
Proof.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (jmp r) M pc).
Intros.
Inversion H25.
Rewrite H29.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Auto.
Auto.
Split with l0.
Split with G0.
Split with T.
Split with I'.
Split with O.
Split.
Auto.
Split.
Constructor.
Split.
Inversion H20.
Cut (TrHeapVal L (code G0 T I') M (L l0)).
Intros.
Inversion H46.
Auto.
Apply H42.
Auto.
Replace (plus (L l0) O) with (plus O (L l0)).
Unfold plus.
Inversion H22.
Cut (TrWordVal L (wl l0) (_R _r)).
Intros.
Inversion H44.
Auto.
Apply H40 with r.
Auto.
Auto.
Apply plus_sym.
Apply tr_pc_trinstrseq with H.
Auto.
Save.



Lemma Preservation_jd
      	       	     : (St:State; H:Heap; R:RegFile; a:AP; l:label;
      	       	       	P,P':Program)
		       P = (H,(R,(a,(jd l)))) ->
		       (WFProgram P) ->
		       (TrProgram P St) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step St)).
Proof.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (jd l) M pc).
Intros.
Inversion H25.
Rewrite H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Auto.
Auto.
Split with l.
Split with G0.
Split with T.
Split with I'.
Split with O.
Split.
Auto.
Split.
Constructor.
Split.
Inversion H20.
Cut (TrHeapVal L (code G0 T I') M (L l)).
Intros.
Inversion H44.
Auto.
Apply H40.
Auto.
Replace (plus (L l) O) with (plus O (L l)).
Unfold plus.
Auto.
Apply plus_sym.

Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H27.
Rewrite <- H28.
Replace (plus (L x) x3) with (plus x3 (L x)).
Apply tr_sub_instrseq with x2.
Auto.
Auto.
Apply plus_sym.
Save.




Lemma trwordvallist_i_wordval
  : (L:LayoutF; M:Mem; V:(list WordVal); i:int; v:WordVal; w:Word)
    (TrWordValList L V M w) ->
    (ListNth ? V i v) ->
    (TrWordVal L v (M (plus w i))).
Proof.
Induction V.
Intros.
Inversion H0.
Clear V.
Intros v V.
Induction i.
Intros.
Inversion H1.
Inversion H0.
Replace (M (plus w O)) with w0.
Rewrite <- H2.
Auto.
Replace (plus w O) with (plus O w).
Auto.
Apply plus_sym.
Intros.
Replace (plus w (S n)) with (plus (S w) n).
Apply H.
Inversion H1.
Auto.
Inversion H2.
Auto.
Replace (plus w (S n)) with (plus (S n) w).
Unfold plus.
Fold plus.
Replace (plus n w) with (plus w n).
Auto.
Apply plus_sym.
Apply plus_sym.
Save.



Lemma Preservation_ld
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd,rs:Reg; i:int; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (ld rd rs i) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (ld rd rs i) I) M pc). Intros.
Inversion H25.
Rewrite H30.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Unfold updatereg.
Constructor.
Intros.
Cut _r=_rd\/~_r=_rd.
Intros.
Inversion_clear H56.
Replace (beq_reg _r _rd) with true.
Replace v0 with v.
Cut (TrHeapVal L (tuple V) M (L l1)).
Intros.
Inversion H56.
Replace (_R _rs) with (L l1).
Apply trwordvallist_i_wordval with V.
Auto.
Auto.
Inversion H22.
Cut (TrWordVal L (wl l1) (_R _rs)).
Intros.
Inversion H67.
Auto.
Apply H63 with rs.
Auto.
Auto.
Inversion H20.
Apply H59.
Auto.
Apply (mlookup_eq ? ? r v v0 R').
Apply eqorneq_reg.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Replace r with rd.
Auto.
Apply trregf_eq_eq with _r.
Rewrite H57.
Auto.
Auto.
Auto.
Rewrite H57.
Apply sym_eq.
Apply beq_reg_eq_true.
Replace (beq_reg _r _rd) with false.
Inversion H22.
Apply H56 with r.
Apply mupdext_preserv_rev with R' rd v.
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Apply eqorneq_tisreg.
Unfold updatereg.
Constructor.
Replace (beq_reg _r31 _rd) with false.
Inversion H23.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_rd=_r31.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Apply tr_pc_Spc with (ld rd rs i).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.



Lemma Preservation_mov
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd,rs:Reg; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (mov rd rs) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (mov rd rs) I) M pc). Intros.
Inversion H25.
Rewrite H30.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Unfold updatereg.
Constructor.
Intros.
Compare _r _rd.
Intros.
Replace (beq_reg _r _rd) with true.
Replace (plus (_R _rs) O) with (plus O (_R _rs)).
Unfold plus.
Replace v0 with v.
Inversion H22.
Apply H52 with rs.
Auto.
Auto.
Apply (mlookup_eq ? ? r v v0 R').
Apply eqorneq_reg.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Replace r with rd.
Auto.
Apply trregf_eq_eq with _r.
Rewrite e.
Auto.
Auto.
Auto.
Apply plus_sym.
Apply sym_eq.
Rewrite e.
Apply beq_reg_eq_true.
Intros.
Replace (beq_reg _r _rd) with false.
Inversion H22.
Apply H52 with r.
Apply mupdext_preserv_rev with R' rd v.
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Constructor.
Unfold updatereg.
Replace (beq_reg _r31 _rd) with false.
Inversion H23.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_rd=_r31.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Apply tr_pc_Spc with (mov rd rs).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.



Lemma Preservation_movi
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd:Reg; i:int; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (movi rd i) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (movi rd i) I) M pc). Intros.
Inversion H25.
Rewrite H30.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Unfold updatereg.
Constructor.
Intros.
Compare _r0 _r.
Intros.
Replace (beq_reg _r0 _r) with true.
Replace v with (wi i).
Constructor.
Apply (mlookup_eq ? ? rd (wi i) v R').
Apply eqorneq_reg.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Auto.
Replace rd with r0.
Auto.
Apply trregf_eq_eq with _r0.
Auto.
Rewrite e.
Auto.
Apply sym_eq.
Rewrite e.
Apply beq_reg_eq_true.
Intros.
Replace (beq_reg _r0 _r) with false.
Inversion H22.
Apply H50 with r0.
Apply mupdext_preserv_rev with R' rd (wi i).
Auto.
Auto.
Apply trregf_neq_neq with _r0 _r.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Constructor.
Unfold updatereg.
Replace (beq_reg _r31 _r) with false.
Inversion H23.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_r=_r31.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Apply tr_pc_Spc with (movi rd i).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.



Lemma Preservation_movl
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	r:Reg; l:label; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (movl r l) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.

Cut (TrInstrSeq L (iseq (movl r l) I) M pc). Intros.
Inversion H25.
Rewrite <- H30 in H28.
Inversion H28.
Clear H24.

Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Unfold updatereg.
Constructor.
Intros.
Cut _r0=_r\/~_r0=_r.
Intros.
Inversion_clear H49.
Rewrite H50.
Replace (beq_reg _r _r) with true.
Replace v with (wl l).
Constructor.
Inversion H13.
Apply mlookup_eq with Reg r R'.
Apply eqorneq_reg.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Auto.
Replace r with r1.
Auto.
Apply trregf_eq_eq with _r0.
Auto.
Rewrite H50.
Auto.
Apply sym_eq.
Apply beq_reg_eq_true.
Replace (beq_reg _r0 _r) with false.
Inversion H22.
Apply H49 with r1.
Apply mupdext_preserv_rev with R' r (wl l).
Auto.
Auto.
Apply trregf_neq_neq with _r0 _r.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Apply eqorneq_tisreg.
Constructor.
Unfold updatereg.
Replace (beq_reg _r31 _r) with false.
Inversion H23.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_r=_r31.
Auto.
Apply trreg_neq_31 with r.
Auto.
Rewrite <- H21 in H2.
Inversion H2.
Rewrite H19.
Rewrite <- H54.
Apply tr_pc_Spc with (movl r l).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.



Lemma tr_update_layout_preserv
  : (h,h':(Map label HeapVal); l:label; hv,hv':HeapVal)
    (mlookup ? ? h l hv) ->
    (mupdate ? ? h l hv' h') ->
    (HV_size hv)=(HV_size hv') ->
    (Layout_aux h)=(Layout_aux h').
Proof.
Induction h.
Intros.
Inversion H.
Intros.
Inversion H1.
Rewrite H3 in H0.
Inversion H0.
Unfold Layout_aux.
Rewrite H2.
Auto.
Absurd ~l=l; Auto.
Unfold Layout_aux.
Fold Layout_aux.
Replace (Layout_aux m') with (Layout_aux m).
Auto.
Apply H with l hv hv'.
Inversion H0.
Rewrite H11 in H10.
Absurd ~l=l; Auto.
Auto.
Auto.
Auto.
Save.


Lemma tr_update_tuple_size_eq
  : (V,V':(list WordVal); i:int; w:WordVal)
    (updatetuple V i w V') ->
    (HV_size (tuple V))=(HV_size (tuple V')).
Proof.
Intros.
Induction H.
Unfold HV_size.
Auto.
Unfold HV_size.
Unfold length.
Unfold HV_size in HrecH.
Auto.
Save.


Lemma tr_update_wordvallist_aux
  : (L:LayoutF; M:Mem; V:(list WordVal); n,a:int; w:Word)
    (lt a (S n)) ->
    (TrWordValList L V M (S n)) ->
    (TrWordValList L V (updatemem M a w) (S n)).
Proof.
Induction V.
Intros.
Constructor.
Intros.
Inversion H1.
Apply trwvl_cons with w0.
Auto.
Unfold updatemem.
Replace (beq_nat (S n) a0) with false.
Auto.
Apply beq_neq_false.
Cut ~a0=(S n).
Auto.
Apply lt_neq.
Auto.
Apply H.
Auto.
Auto.
Save.


Lemma tr_update_wordvallist_aux2
  : (L:LayoutF; M:Mem; i,a,n:nat; V,V':(list WordVal); wv:WordVal; w:Word)
    (TrWordVal L wv w) ->
    a=n ->
    (updatetuple V i wv V') ->
    (TrWordValList L V M n) ->
    (TrWordValList L V' (updatemem M (plus a i) w) n).
Proof.
Induction i.
Intros.
Inversion H1.
Apply trwvl_cons with w.
Auto.
Unfold updatemem.
Replace (beq_nat n (plus a O)) with true.
Auto.
Replace (plus a O) with n.
Apply beq_nat_refl.
Rewrite H0.
Auto.
Rewrite <- H3 in H2.
Inversion H2.
Apply tr_update_wordvallist_aux.
Replace (plus a O) with n.
Apply lt_n_Sn.
Rewrite H0.
Auto.
Auto.
Intros.
Inversion H2.
Rewrite <- H5 in H3.
Inversion H3.
Apply trwvl_cons with w1.
Auto.
Unfold updatemem.
Replace (beq_nat n0 (plus a (S n))) with false.
Auto.
Apply beq_neq_false.
Rewrite H1.
Apply plus_S_neq.
Replace (plus a (S n)) with (plus (S a) n).
Apply H with V0 wv.
Auto.
Auto.
Auto.
Auto.
Apply plus_Snm_nSm.
Save.




Lemma lt_Sm_n : (m,n:nat) (lt (S m) n) -> (lt m n).
Proof.
Induction m.
Intros.
Inversion H.
Auto.
Apply lt_O_Sn.
Intros.
Inversion H0.
Auto.
Apply lt_n_S.
Apply H.
Unfold lt.
Apply le_trans_S.
Auto.
Save.


Lemma label_lt_leftinc_neq 
  : (l,l':nat; m:nat) (lt (plus l m) l') -> (k:nat)(le k m) -> (lt (plus l k) l').
Proof.
Induction m.
Intros.
Inversion H0.
Auto.
Intros.
Inversion H1.
Auto.
Apply H.
Apply lt_Sm_n.
Replace (S (plus l n)) with (plus l (S n)).
Auto.
Auto.
Auto.
Save.


Lemma label_lt_rightinc_neq
  : (l,l':nat; n:nat) (lt (plus l' n) l) -> (k:nat)(lt (plus l' n) (plus l k)).
Proof.
Induction k.
Replace (plus l O) with l.
Auto.
Auto.
Intros.
Replace (plus l (S n0)) with (S (plus l n0)).
Apply lt_S.
Auto.
Auto.
Save.


Lemma lt_plus_S : (n,k:nat) (lt n (plus n (S k))).
Proof.
Induction n.
Intros.
Unfold plus.
Apply lt_O_Sn.
Intros.
Replace (plus (S n0) (S k)) with (S (plus n0 (S k))).
Apply lt_n_S.
Apply H.
Auto.
Save.


Lemma lt_prop_aux
  : (l,l',i,m,n:nat)
    (le i n) ->
    (lt (plus l m) l') \/ (lt (plus l' n) l) ->
    (lt (plus l m) (plus l' i)) \/ (lt (plus l' i) l).
Proof.
Intros.
Inversion_clear H0.
Left.
Induction i.
Replace (plus l' O) with l'.
Auto.
Auto.
Replace (plus l' (S i)) with (S (plus l' i)).
Apply lt_S.
Apply Hreci.
Apply le_trans_S.
Auto.
Auto.
Right.
Inversion H.
Auto.
Cut (lt i n).
Intros.
Apply lt_trans with (plus l' n).
Apply le_lt_plus_plus.
Constructor.
Auto.
Auto.
Unfold lt.
Rewrite <- H2.
Apply le_n_S.
Auto.
Save.



Lemma wf_propsubheap_wf
  : (h,h':HeapMap) (mWF ? ? h) -> (PropSubHeap h' h) -> (mWF ? ? h').
Proof.
Induction h.
Intros.
Inversion H0.
Intros.
Inversion H1.
Inversion H0.
Auto.
Apply H.
Inversion H0.
Auto.
Auto.
Save.


Lemma wf_subheap_wf
  : (h,h':HeapMap) (mWF ? ? h) -> (SubHeap h' h) -> (mWF ? ? h').
Proof.
Intros.
Inversion H0.
Auto.
Apply wf_propsubheap_wf with h.
Auto.
Auto.
Save.

Lemma ord_propsubheap_ord
  : (h,h':HeapMap) (OrdHeap h) -> (PropSubHeap h' h) -> (OrdHeap h').
Proof.
Induction h.
Intros.
Inversion H0.
Intros.
Inversion H1.
Inversion H0.
Auto.
Apply H.
Inversion H0.
Auto.
Auto.
Save.


Lemma ord_subheap_ord
  : (h,h':HeapMap) (OrdHeap h) -> (SubHeap h' h) -> (OrdHeap h').
Proof.
Intros.
Inversion H0.
Auto.
Apply ord_propsubheap_ord with h.
Auto.
Auto.
Save.


Lemma size_propsubheap_lt
  : (h,h':HeapMap)
    (PropSubHeap h' h) ->
    (s,s':nat)
    (OrdHeap h) ->
    (msize ? ? h s) ->
    (msize ? ? h' s') ->
    (lt s' s).
Proof.
Intros h h' H.
Induction H.
Intros.
Replace s with (S s').
Auto.
Apply (msize_eq ? ? (melem nat HeapVal l hv h)).
Constructor.
Auto.
Auto.
Intros.
Inversion H1.
Apply lt_S.
Apply HrecH.
Inversion H0.
Auto.
Auto.
Auto.
Save.


Lemma ord_propsubheap_mlookup
  : (h,h':HeapMap; s,l:nat; hv:HeapVal)
    (OrdHeap h) -> (PropSubHeap h' h) -> 
    (msize ? ? h' s) ->
    (lt l s) ->
    (mlookup ? ? h l hv) ->
    (mlookup ? ? h' l hv).
Proof.
Induction h.
Intros.
Inversion H0.
Intros.
Inversion H1.
Inversion H4.
Absurd l=a.
Apply lt_neq.
Replace a with s.
Auto.
Inversion H0.
Rewrite <- H9 in H16.
Apply (msize_eq ? ? h').
Auto.
Auto.
Auto.
Auto.
Apply H with s.
Inversion H0.
Auto.
Auto.
Auto.
Auto.
Inversion H4.
Absurd l=a.
Apply lt_neq.
Apply lt_trans with s.
Auto.
Inversion H0.
Apply size_propsubheap_lt with m h'.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Save.


Lemma ord_propsubheap_mindom
  : (h,h':HeapMap; s,l:nat)
    (OrdHeap h) -> (PropSubHeap h' h) -> 
    (msize ? ? h' s) ->
    (lt l s) ->
    (mindom ? ? h l) ->
    (mindom ? ? h' l).
Proof.
Induction h.
Intros.
Inversion H0.
Intros.
Inversion H1.
Inversion H4.
Absurd l=a.
Apply lt_neq.
Replace a with s.
Auto.
Inversion H0.
Rewrite <- H9 in H15.
Apply (msize_eq ? ? h').
Auto.
Auto.
Auto.
Auto.
Apply H with s.
Inversion H0.
Auto.
Auto.
Auto.
Auto.
Inversion H4.
Absurd l=a.
Apply lt_neq.
Apply lt_trans with s.
Auto.
Inversion H0.
Apply size_propsubheap_lt with m h'.
Auto.
Auto.
Auto.
Auto.
Auto.
Auto.
Save.



Lemma ord_heap_label_notin
  : (h:HeapMap; s,t:nat)
    (OrdHeap h) ->
    (msize ? ? h s) ->
    (le s t) ->
    (mnotindom ? ? h t).
Proof.
Induction h.
Intros.
Constructor.
Intros.
Constructor.
Inversion H1.
Cut (lt n t).
Intros.
Replace a with n.
Apply lt_neq.
Auto.
Apply (msize_eq ? ? m).
Auto.
Inversion H0.
Auto.
Rewrite <- H4 in H2.
Auto.
Apply H with a.
Inversion H0.
Auto.
Inversion H0.
Auto.
Apply le_trans with s.
Inversion H1.
Replace a with n.
Auto.
Apply (msize_eq ? ? m).
Auto.
Inversion H0.
Auto.
Auto.
Save.



Lemma layout_aux_eq_top_label
  : (h:HeapMap)
    (wfh:(mWF ? ? h); ordh:(OrdHeap h))
    (l,s:label)
    (msize ? ? h s) ->
    (lt l s) ->
    (EX m | 
      (  (msize ? ? m l) /\ (SubHeap m h) 
      /\ (Layout_aux h l)=(Layout_aux m l))).
Proof.
Induction h.
Intros.
Inversion H.
Rewrite <- H2 in H0.
Inversion H0.
Intros.
Inversion H1.
Cut l=a.
Intros.
Rewrite H3.
Split with m.
Split.
Inversion ordh.
Auto.
Split.
Constructor.
Constructor.
Unfold 1 Layout_aux.
Replace (blt_nat a a) with false.
Auto.
Apply sym_eq.
Apply blt_eq_false.
Cut (S a)=s.
Intros.
Rewrite <- H2 in H3.
Apply eq_add_S.
Auto.
Apply (msize_eq ? ? (melem ? ? a b m)).
Constructor.
Inversion ordh.
Auto.
Auto.
Cut (EX m0:? |
       (msize ? ? m0 l)
       /\(SubHeap m0 m)/\(Layout_aux m l)=(Layout_aux m0 l)).
Intros.
Inversion_clear H4.
Split with x.
Split.
Inversion_clear H5.
Auto.
Split.
Inversion_clear H5.
Inversion_clear H6.
Inversion H5.
Constructor.
Constructor.
Constructor.
Constructor.
Auto.
Inversion_clear H5.
Inversion_clear H6.
Unfold 1 Layout_aux.
Replace (blt_nat a l) with false.
Auto.
Apply sym_eq.
Apply blt_false.
Rewrite <- H3 in H1.
Replace a with m0.
Auto.
Cut (S a)=s.
Intros.
Apply eq_add_S.
Rewrite H3.
Rewrite H6.
Auto.
Apply (msize_eq ? ? (melem ? ? a b m)).
Constructor.
Inversion ordh.
Auto.
Auto.
Apply H with m0.
Inversion wfh.
Auto.
Inversion ordh.
Auto.
Rewrite <- H3 in H0.
Inversion H0.
Auto.
Auto.
Save.



Lemma layout_aux_lt_eq
  : (h:HeapMap; a,a',s:nat)
    (OrdHeap h) ->
    (msize ? ? h s) ->
    (le s a) ->
    (le s a') ->
    (Layout_aux h a)=(Layout_aux h a').
Proof.
Induction h.
Intros.
Auto.
Intros.
Unfold Layout_aux.
Replace (blt_nat a a0) with true.
Replace (blt_nat a a') with true.
Fold Layout_aux.
Replace (Layout_aux m a0) with (Layout_aux m a').
Auto.
Induction m.
Auto.
Apply H with a.
Inversion H0.
Auto.
Inversion H0.
Auto.
Apply le_trans_S.
Replace (S a) with s.
Auto.
Apply (msize_eq ? ? (melem nat HeapVal a b (melem nat HeapVal a1 b0 m))).
Auto.
Constructor.
Inversion H0.
Auto.
Apply le_trans_S.
Replace (S a) with s.
Auto.
Apply (msize_eq ? ? (melem nat HeapVal a b (melem nat HeapVal a1 b0 m))).
Auto.
Constructor.
Inversion H0.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Unfold lt.
Replace (S a) with s.
Auto.
Apply (msize_eq ? ? (melem nat HeapVal a b m)).
Auto.
Constructor.
Inversion H0.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Unfold lt.
Replace (S a) with s.
Auto.
Apply (msize_eq ? ? (melem nat HeapVal a b m)).
Auto.
Constructor.
Inversion H0.
Auto.
Save.


Lemma layout_aux_le_sub
  : (h:HeapMap; l:label)
    (OrdHeap h) ->
    (msize ? ? h (S l)) ->
    (le (Layout_aux h l) (Layout_aux h (S l))).
Proof.
Induction h.
Intros.
Inversion H0.
Intros.
Replace a with l.
Unfold 1 Layout_aux.
Replace (blt_nat l l) with false.
Fold Layout_aux.
Unfold 2 Layout_aux.
Replace (blt_nat l (S l)) with true.
Fold Layout_aux.
Replace (Layout_aux m (S l)) with (Layout_aux m l).
Apply le_plus_l.
Apply layout_aux_lt_eq with l.
Inversion H0.
Auto.
Inversion H1.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Auto.
Apply sym_eq.
Apply blt_eq_false.
Inversion H0.
Inversion H1.
Apply (msize_eq ? ? m).
Auto.
Auto.
Save.



Lemma layout_aux_plus_le_sub
  : (h:HeapMap; l:label; hv:HeapVal)
    (OrdHeap h) ->
    (msize ? ? h (S l)) ->
    (mlookup ? ? h l hv) ->
    (le (plus (Layout_aux h l) (HV_size hv)) (Layout_aux h (S l))).
Proof.
Induction h.
Intros.
Inversion H0.
Intros.
Replace a with l.
Unfold 1 Layout_aux.
Replace (blt_nat l l) with false.
Fold Layout_aux.
Unfold 2 Layout_aux.
Replace (blt_nat l (S l)) with true.
Fold Layout_aux.
Replace (Layout_aux m (S l)) with (Layout_aux m l).
Replace b with hv.
Auto.
Cut a=l.
Intros.
Rewrite H3 in H2.
Inversion H2.
Auto.
Absurd ~l=l; Auto.
Apply eq_add_S.
Apply (msize_eq ? ? (melem ? ? a b m)).
Constructor.
Inversion H0.
Auto.
Auto.
Apply layout_aux_lt_eq with l.
Inversion H0.
Auto.
Inversion H1.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Auto.
Apply sym_eq.
Apply blt_eq_false.
Apply eq_add_S.
Apply (msize_eq ? ? (melem ? ? a b m)).
Auto.
Constructor.
Inversion H0.
Auto.
Save.


Lemma ordheap_mindom_lt_size
  : (h:HeapMap; l,s:nat)
    (OrdHeap h) ->
    (msize ? ? h s) ->
    (mindom ? ? h l) ->
    (lt l s).
Proof.
Induction h.
Intros.
Inversion H1.
Intros.
Cut a=l\/~a=l.
Intros.
Inversion_clear H3.
Rewrite <- H4.
Inversion H1.
Replace n with a.
Auto.
Apply (msize_eq ? ? m).
Inversion H0.
Auto.
Auto.
Inversion H1.
Apply lt_S.
Apply H.
Inversion H0.
Auto.
Auto.
Inversion H2.
Rewrite H9 in H4.
Absurd ~l=l; Auto.
Auto.
Apply eqorneq_nat.
Save.


Lemma layout_aux_propsubheap_eq
  : (h,h':HeapMap; l,s:label)
    (OrdHeap h) ->
    (PropSubHeap h' h) ->
    (msize ? ? h' s) ->
    (lt l s) ->
    (Layout_aux h' l)=(Layout_aux h l).
Proof.
Induction h.
Intros.
Inversion H0.
Intros.
Inversion H1.
Unfold 2 Layout_aux.
Replace (blt_nat a l) with false.
Auto.
Apply sym_eq.
Apply blt_false.
Replace a with s.
Auto.
Apply (msize_eq ? ? m).
Rewrite <- H8.
Auto.
Inversion H0.
Auto.
Unfold 2 Layout_aux.
Replace (blt_nat a l) with false.
Fold Layout_aux.
Apply H with s.
Inversion H0.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply blt_false.
Apply lt_trans with s.
Auto.
Apply size_propsubheap_lt with m h'.
Auto.
Inversion H0.
Auto.
Inversion H0.
Auto.
Auto.
Save.


Lemma layout_aux_top_label_plus_nonoverlap
  : (s,l:nat; h:HeapMap; hv:HeapVal)
    (mWF ? ? h) -> (OrdHeap h) ->
    (msize ? ? h (S s)) ->
    (mlookup ? ? h l hv) ->
    (lt l s) ->
    (le (plus (Layout_aux h l) (HV_size hv)) (Layout_aux h s)).
Proof.
Induction s.
Intros.
Inversion H3.
Intros.
Inversion H2.
Cut (S n)=a.
Intros.
Unfold 2 Layout_aux.
Replace (blt_nat a (S n)) with false.
Fold Layout_aux.
Cut l=n\/~l=n.
Intros.
Inversion_clear H9.
Rewrite H10.
Rewrite <- H8.
Inversion H7.
Replace a0 with n.
Replace hv with b0.
Unfold 1 Layout_aux.
Replace (blt_nat (S n) n) with false.
Replace (blt_nat n n) with false.
Fold Layout_aux.
Unfold 2 Layout_aux.
Replace (blt_nat n (S n)) with true.
Fold Layout_aux.
Replace (Layout_aux m0 n) with (Layout_aux m0 (S n)).
Auto.
Apply layout_aux_lt_eq with n.
Rewrite <- H11 in H6.
Rewrite <- H6 in H1.
Inversion H1.
Inversion H17.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply blt_lt_true.
Auto.
Apply sym_eq.
Apply blt_eq_false.
Apply sym_eq.
Apply blt_false.
Auto.
Apply (mlookup_eq ? ? l b0 hv h).
Apply eqorneq_nat.
Rewrite <- H6.
Rewrite <- H11.
Replace a0 with l.
Apply mlooknext.
Apply mlookfound.
Rewrite <- H8.
Rewrite H10.
Auto.
Rewrite H10.
Replace a0 with n.
Auto.
Apply (msize_eq ? ? m0).
Auto.
Rewrite <- H11 in H6.
Rewrite <- H6 in H1.
Inversion H1.
Inversion H17.
Auto.
Auto.
Apply (msize_eq ? ? m0).
Auto.
Rewrite <- H11 in H6.
Rewrite <- H6 in H1.
Inversion H1.
Inversion H17.
Auto.
Apply le_trans with (Layout_aux m n).
Replace (Layout_aux (melem nat HeapVal a b m) l) with (Layout_aux m l).
Apply H.
Rewrite <- H6 in H0.
Inversion H0.
Auto.
Rewrite <- H6 in H1.
Inversion H1.
Auto.
Auto.
Rewrite <- H6 in H3.
Inversion H3.
Absurd l=a.
Rewrite <- H8.
Apply lt_neq.
Auto.
Auto.
Auto.
Apply lt_S_neq.
Auto.
Auto.
Rewrite <- H8.
Unfold 2 Layout_aux.
Replace (blt_nat (S n) l) with false.
Auto.
Apply sym_eq.
Apply blt_false.
Auto.
Apply layout_aux_le_sub.
Rewrite <- H6 in H1.
Inversion H1.
Auto.
Auto.
Apply eqorneq_nat.
Apply sym_eq.
Rewrite H8.
Apply blt_eq_false.
Rewrite <- H6 in H1.
Inversion H1.
Apply (msize_eq ? ? m).
Auto.
Auto.
Save.


Lemma layout_aux_label_plus_nonoverlap
  : (h:HeapMap; l,l':label; hv:HeapVal)
    (mWF ? ? h) -> (OrdHeap h) ->
    (mlookup ? ? h l hv) ->
    (mindom ? ? h l') ->
    (lt l l') ->
    (le (plus (Layout_aux h l) (HV_size hv)) (Layout_aux h l')).
Proof.
Intros.
Cut (EX h':(Map nat HeapVal) |
       (msize nat HeapVal h' l')
       /\(SubHeap h' h)/\(Layout_aux h l')=(Layout_aux h' l')).
Intros.
Elim H4; Clear H4; Intros h' H4.
Inversion_clear H4.
Inversion_clear H6.
Inversion H4.
Absurd (mnotindom ? ? h l').
Apply mindom_imp_not_mnotindom.
Auto.
Auto.
Apply ord_heap_label_notin with l'.
Auto.
Rewrite <- H8.
Auto.
Auto.
Clear H8.
Clear H9.
Clear h0.
Clear h'0.
Cut (EX s:? | (msize ? ? h s)).
Intros.
Elim H8; Clear H8; Intros s H8.
Replace (Layout_aux h l) with (Layout_aux h' l).
Rewrite H7.
Inversion H5.
Absurd (mnotindom ? ? h' l).
Apply mindom_imp_not_mnotindom.
Apply wf_subheap_wf with h.
Auto.
Auto.
Apply ord_propsubheap_mindom with h l'.
Auto.
Auto.
Auto.
Auto.
Apply mlookup_imp_mindom with hv.
Auto.
Rewrite <- H9.
Constructor.
Cut n=a.
Intros.
Cut l=a\/~l=a.
Intros.
Inversion_clear H13.
Rewrite H14.
Rewrite H12.
Apply layout_aux_plus_le_sub.
Rewrite H10.
Apply ord_subheap_ord with h.
Auto.
Auto.
Auto.
Rewrite H10.
Rewrite <- H12.
Rewrite H11.
Auto.
Rewrite H10.
Rewrite <- H14.
Apply ord_propsubheap_mlookup with h l'.
Auto.
Auto.
Auto.
Auto.
Auto.
Cut (lt l n).
Intros.
Rewrite H12.
Apply le_trans with (Layout_aux (melem ? ? a b m) a).
Apply layout_aux_top_label_plus_nonoverlap.
Apply wf_subheap_wf with h.
Auto.
Rewrite H10.
Auto.
Apply ord_subheap_ord with h.
Auto.
Rewrite H10.
Auto.
Constructor.
Rewrite <- H12.
Auto.
Rewrite H10.
Apply ord_propsubheap_mlookup with h l'.
Auto.
Auto.
Auto.
Auto.
Auto.
Rewrite <- H12.
Auto.
Apply layout_aux_le_sub.
Apply ord_subheap_ord with h.
Auto.
Rewrite H10.
Auto.
Constructor.
Rewrite <- H12.
Auto.
Apply lt_S_neq.
Rewrite H11.
Auto.
Rewrite H12.
Auto.
Apply eqorneq_nat.
Apply (msize_eq ? ? m).
Auto.
Cut (OrdHeap (melem ? ? a b m)).
Intros.
Inversion H12.
Auto.
Apply ord_subheap_ord with h.
Auto.
Rewrite H10.
Auto.
Apply layout_aux_propsubheap_eq with l'.
Auto.
Auto.
Auto.
Auto.
Apply msize_always.
Cut (EX s:nat | (msize nat HeapVal h s)).
Intros.
Inversion_clear H4.
Apply layout_aux_eq_top_label with x.
Auto.
Auto.
Auto.
Apply ordheap_mindom_lt_size with h.
Auto.
Auto.
Auto.
Apply msize_always.
Save.


Lemma updatetuple_list_len_eq
  : (V,V':(list WordVal); i:nat; wv:WordVal)
    (updatetuple V i wv V') ->
    (length V)=(length V').
Proof.
Induction V.
Intros.
Inversion H.
Intros.
Inversion H0.
Unfold length.
Apply eq_S.
Auto.
Unfold length.
Apply eq_S.
Apply H with i0 wv.
Auto.
Save.


Lemma tuple_length_le_aux
  : (i:nat; V,V':(list WordVal); wv:WordVal)
    (updatetuple V i wv V') ->
    (lt i (length V')).
Proof.
Induction i.
Intros.
Inversion H.
Unfold length.
Apply lt_O_Sn.
Intros.
Inversion H0.
Unfold length.
Fold length.
Apply lt_n_S.
Apply H with V0 wv.
Auto.
Save.


Lemma tuple_length_le
  : (i,j,k:nat) (lt i j) -> (le (S (plus k i)) (plus k j)).
Proof.
Induction i.
Intros.
Replace (plus k O) with k.
Inversion H.
Replace (plus k (S O)) with (S k).
Auto.
Replace (plus k (S O)) with (plus (S O) k).
Auto.
Apply plus_sym.
Replace (plus k (S m)) with (S (plus k m)).
Apply le_n_S.
Apply le_plus_l.
Replace (plus k (S m)) with (plus (S m) k).
Unfold 2 plus.
Replace (plus k m) with (plus m k).
Auto.
Apply plus_sym.
Apply plus_sym.
Auto.
Intros.
Inversion H0.
Replace (plus k (S (S n))) with (S (S (plus k n))).
Replace (S (plus k (S n))) with (S (S (plus k n))).
Auto.
Replace (plus k (S n)) with (plus (S n) k).
Unfold 2 plus.
Fold plus.
Replace (plus n k) with (plus k n).
Auto.
Apply plus_sym.
Apply plus_sym.
Replace (plus k (S (S n))) with (plus (S (S n)) k).
Unfold 2 plus.
Fold plus.
Replace (plus n k) with (plus k n).
Auto.
Apply plus_sym.
Apply plus_sym.
Inversion H1.
Replace (plus k (S n)) with (plus (S n) k).
Replace (plus k (S (S (S n)))) with (plus (S (S (S n))) k).
Unfold plus.
Auto.
Apply plus_sym.
Apply plus_sym.
Replace (S (plus k (S n))) with (S (S (plus k n))).
Replace (plus k (S (S m0))) with (S (plus k (S m0))).
Apply le_n_S.
Apply H.
Unfold lt.
Apply le_S.
Apply le_trans with (S (S n)).
Auto.
Auto.
Replace (plus k (S m0)) with (plus (S m0) k).
Replace (plus k (S (S m0))) with (plus (S (S m0)) k).
Unfold plus.
Auto.
Apply plus_sym.
Apply plus_sym.
Replace (plus k (S n)) with (plus (S n) k).
Replace (plus k n) with (plus n k).
Unfold plus.
Auto.
Apply plus_sym.
Apply plus_sym.
Save.


Lemma layout_non_overlap
  : (L:LayoutF; H:Heap; HT:HeapTy)
    (l,l':label; hv,hv':HeapVal)
    (WFHeap H HT) ->
    L=(Layout H) ->
    (hlookup (hunwrap H) l hv) ->
    (hlookup (hunwrap H) l' hv') ->
    ~l=l' ->
    (le (plus (L l) (HV_size hv)) (L l'))
    \/ (le (plus (L l') (HV_size hv')) (L l)).
Proof.
Intros.
Cut (lt l l')\/(lt l' l).
Intros.
Inversion_clear H5.
Left.
Rewrite H1.
Unfold Layout.
Apply layout_aux_label_plus_nonoverlap.
Induction H.
Unfold unwrapMap.
Auto.
Inversion H0.
Inversion_clear H7.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Auto.
Auto.
Apply mlookup_imp_mindom with hv'.
Auto.
Auto.
Right.
Rewrite H1.
Unfold Layout.
Apply layout_aux_label_plus_nonoverlap.
Induction H.
Unfold unwrapMap.
Auto.
Inversion H0.
Inversion_clear H7.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Inversion_clear H16.
Auto.
Auto.
Apply mlookup_imp_mindom with hv.
Auto.
Auto.
Apply nat_total_order.
Auto.
Save.



Lemma tr_update_wordvallist_preserv
  : (L:LayoutF; M:Mem; l':Word; w:Word)
    (V:(list WordVal))
    (l:Word)
    (le (plus l (length V)) l') \/ (lt l' l) ->
    (TrWordValList L V M l) ->
    (TrWordValList L V (updatemem M l' w) l).
Proof.
Induction V.
Intros.
Constructor.
Intros.
Inversion H1.
Apply trwvl_cons with w0.
Auto.
Unfold updatemem.
Replace (beq_nat l0 l') with false.
Auto.
Apply beq_neq_false.
Inversion H0.
Apply lt_neq.
Cut (plus l0 (length (cons a l)))=(S (plus l0 (length l))).
Intros.
Rewrite H11 in H10.
Unfold lt.
Apply le_trans with (S (plus l0 (length l))).
Apply le_n_S.
Apply le_plus_l.
Auto.
Unfold 1 length.
Auto.
Cut ~l'=l0.
Auto.
Apply lt_neq.
Auto.
Apply H.
Inversion H0.
Left.
Replace (plus (S l0) (length l)) with (plus l0 (length (cons a l))).
Auto.
Unfold length.
Unfold plus.
Auto.
Right.
Apply lt_S.
Auto.
Auto.
Save.





Lemma tr_update_instrseq_preserv
  : (L:LayoutF; M:Mem; l':Word; w:Word)
    (IS:InstrSeq)
    (l:Word)
    (le (plus l (lenInstrSeq IS)) l') \/ (lt l' l) ->
    (TrInstrSeq L IS M l) ->
    (TrInstrSeq L IS (updatemem M l' w) l).
Proof.
Induction IS.
Intros.
Inversion H1.
Apply tris_iseq with _i.
Auto.
Unfold updatemem.
Replace (beq_nat l l') with false.
Auto.
Apply beq_neq_false.
Inversion H0.
Cut (plus l (lenInstrSeq (iseq i i0)))=(S (plus l (lenInstrSeq i0))).
Intros.
Rewrite H11 in H10.
Apply lt_neq.
Unfold lt.
Apply le_trans with (S (plus l (lenInstrSeq i0))).
Apply le_n_S.
Apply le_plus_l.
Auto.
Unfold 1 lenInstrSeq.
Auto.
Cut ~l'=l.
Auto.
Apply lt_neq.
Auto.
Apply H.
Inversion H0.
Left.
Replace (plus (S l) (lenInstrSeq i0))
 with (plus l (lenInstrSeq (iseq i i0))).
Auto.
Replace (plus (S l) (lenInstrSeq i0))
 with (plus l (lenInstrSeq (iseq i i0))).
Auto.
Unfold lenInstrSeq.
Unfold plus.
Auto.
Right.
Apply lt_S.
Auto.
Auto.

Intros.
Constructor.
Inversion H0.
Unfold updatemem.
Replace (beq_nat l l') with false.
Auto.
Apply beq_neq_false.
Inversion H.
Apply lt_neq.
Unfold lenInstrSeq in H6.
Cut (plus l (S O))=(S l).
Intros.
Rewrite H7 in H6.
Auto.
Replace (plus l (S O)) with (plus (S O) l).
Auto.
Apply plus_sym.
Cut ~l'=l.
Auto.
Apply lt_neq.
Auto.

Intros.
Inversion H0.
Apply tris_jmp with _r.
Auto.
Unfold updatemem.
Replace (beq_nat l l') with false.
Auto.
Apply beq_neq_false.
Inversion H.
Apply lt_neq.
Unfold lenInstrSeq in H7.
Cut (plus l (S O))=(S l).
Intros.
Rewrite H8 in H7.
Auto.
Replace (plus l (S O)) with (plus (S O) l).
Auto.
Apply plus_sym.
Cut ~l'=l.
Auto.
Apply lt_neq.
Auto.
Save.




Lemma Preservation_st
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd,rs:Reg; i:int; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (st rd i rs) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (st rd i rs) I) M pc).
Intros.
Inversion H26.
Rewrite <- H31 in H29.
Inversion H29.
Clear H25.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Rewrite H20.
Unfold Layout unwrapMap.
Rewrite H49.
Rewrite H51.
Apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
Auto.
Auto.
Apply tr_update_tuple_size_eq with i w.
Auto.
Inversion H21.
Apply trheap with s.
Unfold unwrapMap.
Rewrite H51.
Apply mupdate_size_same with h l1 (tuple V').
Apply eqorneq_nat.
Auto.
Rewrite H49 in H58.
Auto.
Intros.
Cut n=l1\/~n=l1.
Intros.
Rewrite H51 in H63.
Unfold unwrapMap in H63.
Inversion_clear H64.
Replace h0 with (tuple V').
Cut (TrHeapVal L (tuple V) M (L n)).
Intros.
Constructor.
Inversion H64.
Apply tr_update_wordvallist_aux2 with V w.
Inversion H23.
Apply H71 with rs.
Auto.
Auto.
Cut (TrWordVal L (wl n) (_R _rd)).
Intros.
Inversion H71.
Auto.
Inversion H23.
Apply H71 with rd.
Rewrite H65.
Auto.
Auto.
Auto.
Auto.
Apply H59.
Rewrite H49.
Unfold unwrapMap.
Rewrite H65.
Auto.

Apply mlookup_eq with nat n h'.
Apply eqorneq_nat.
Apply mupdate_imp_mlookup with h.
Apply eqorneq_nat.
Rewrite H65.
Auto.
Auto.
Cut (TrHeapVal L h0 M (L n)).
Intros.

Induction h0.
Constructor.
Apply tr_update_wordvallist_preserv.
Cut (le (plus (L n) (HV_size (tuple l2))) (L l1))
    \/(le (plus (L l1) (HV_size (tuple V'))) (L n)).
Intros.
Inversion_clear H66.
Left.
Unfold HV_size in H67.
Replace (_R _rd) with (L l1).
Apply le_trans with (L l1).
Auto.
Apply le_plus_l.
Inversion H23.
Cut (TrWordVal L (wl l1) (_R _rd)).
Intros.
Inversion H71.
Auto.
Apply H66 with rd.
Auto.
Auto.
Right.
Unfold lt.
Replace (_R _rd) with (L l1).
Apply le_trans with (plus (L l1) (HV_size (tuple V'))).
Unfold HV_size.
Apply tuple_length_le.
Apply tuple_length_le_aux with V w.
Auto.
Auto.
Inversion H23.
Cut (TrWordVal L (wl l1) (_R _rd)).
Intros.
Inversion H71.
Auto.
Apply H66 with rd.
Auto.
Auto.
Rewrite <- H50 in H4.
Inversion H4.
Apply layout_non_overlap with H' HT0.
Auto.
Rewrite H20.
Unfold Layout.
Apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
Rewrite H49.
Auto.
Rewrite H49.
Rewrite H51.
Auto.
Unfold HV_size.
Apply updatetuple_list_len_eq with i w.
Auto.
Rewrite H51.
Auto.
Rewrite H51.
Unfold unwrapMap.
Apply mupdate_imp_mlookup with h.
Apply eqorneq_nat.
Auto.
Auto.

Inversion H64.
Auto.
Inversion H64.
Constructor.
Apply tr_update_instrseq_preserv.
Cut (le (plus (L n) (HV_size (code m a0 i3))) (L l1))
    \/(le (plus (L l1) (HV_size (tuple V'))) (L n)).
Intros.
Inversion_clear H73.
Left.
Unfold HV_size in H74.
Replace (_R _rd) with (L l1).
Apply le_trans with (L l1).
Auto.
Apply le_plus_l.
Inversion H23.
Cut (TrWordVal L (wl l1) (_R _rd)).
Intros.
Inversion H78.
Auto.
Apply H73 with rd.
Auto.
Auto.
Right.
Unfold lt.
Replace (_R _rd) with (L l1).
Apply le_trans with (plus (L l1) (HV_size (tuple V'))).
Unfold HV_size.
Apply tuple_length_le.
Apply tuple_length_le_aux with V w.
Auto.
Auto.
Inversion H23.
Cut (TrWordVal L (wl l1) (_R _rd)).
Intros.
Inversion H78.
Auto.
Apply H73 with rd.
Auto.
Auto.
Rewrite <- H50 in H4.
Inversion H4.
Apply layout_non_overlap with H' HT0.
Auto.
Rewrite H20.
Unfold Layout.
Apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
Rewrite H49.
Auto.
Rewrite H49.
Rewrite H51.
Auto.
Unfold HV_size.
Apply updatetuple_list_len_eq with i w.
Auto.
Rewrite H51.
Auto.
Rewrite H51.
Unfold unwrapMap.
Apply mupdate_imp_mlookup with h.
Apply eqorneq_nat.
Auto.
Auto.
Auto.
Apply H59.
Rewrite H49.
Unfold unwrapMap.
Apply mupdate_preserv_rev with h' l1 (tuple V').
Apply eqorneq_nat.
Auto.
Auto.
Auto.
Apply eqorneq_nat.
Auto.

Auto.
Rewrite <- H22 in H2.
Inversion H2.
Inversion_clear H69.
Inversion_clear H70.
Inversion_clear H69.
Inversion_clear H70.
Inversion_clear H69.
Inversion_clear H70.
Inversion_clear H71.
Inversion_clear H72.
Split with x.
Split with x0.
Split with x1.
Split with x2.
Split with (S x3).
Cut ~x=l1.
Intros.
Split.
Apply mupdate_preserv with h l1 (tuple V').
Rewrite H51.
Auto.
Rewrite H49 in H69.
Auto.
Auto.
Split.
Apply tr_instrseq_subdepth_S with (st rd i rs).
Auto.
Split.
Apply tr_update_instrseq_preserv.
Cut (_R _rd)=(L l1).
Intros.
Cut (le (plus (L x) (HV_size (code x0 x1 x2))) (L l1))
    \/(le (plus (L l1) (HV_size (tuple V'))) (L x)).
Intros.
Inversion_clear H75.
Left.
Unfold HV_size in H76.
Rewrite H74.
Apply le_trans with (L l1).
Auto.
Apply le_plus_l.
Right.
Unfold lt; Rewrite H74.
Apply le_trans with (plus (L l1) (HV_size (tuple V'))).
Unfold HV_size.
Apply tuple_length_le.
Apply tuple_length_le_aux with V w.
Auto.
Auto.
Rewrite <- H50 in H4.
Inversion H4.
Apply layout_non_overlap with H' HT0.
Auto.
Rewrite H20.
Unfold Layout.
Apply tr_update_layout_preserv with l1 (tuple V) (tuple V').
Rewrite H49.
Auto.
Rewrite H49.
Rewrite H51.
Auto.
Unfold HV_size.
Apply updatetuple_list_len_eq with i w.
Auto.
Rewrite H51.
Apply mupdate_preserv with h l1 (tuple V').
Auto.
Rewrite H49 in H69.
Auto.
Auto.
Rewrite H51.
Apply mupdate_imp_mlookup with h.
Apply eqorneq_nat.
Auto.
Auto.
Inversion H23.
Cut (TrWordVal L (wl l1) (_R _rd)).
Intros.
Inversion H78.
Auto.
Apply H74 with rd.
Auto.
Auto.
Rewrite H20.
Rewrite <- H64.
Auto.
Replace (plus (L x) (S x3)) with (plus (S x3) (L x)).
Unfold plus.
Fold plus.
Replace (plus x3 (L x)) with (plus (L x) x3).
Apply eq_S.
Rewrite H20.
Rewrite <- H64.
Auto.
Apply plus_sym.
Apply plus_sym.
Cut x=l1\/~x=l1.
Intros.
Inversion_clear H72.
Absurd (code x0 x1 x2)=(tuple V).
Discriminate.
Apply (mlookup_eq ? ? x (code x0 x1 x2) (tuple V) h).
Apply eqorneq_nat.
Rewrite H49 in H69.
Auto.
Rewrite H74.
Auto.
Auto.
Apply eqorneq_nat.
Apply tr_pc_trinstrseq with H.
Auto.
Save.




Lemma Preservation_unfold
      	       	     : (S:State; H:Heap; R:RegFile; a:AP; 
      	       	       	rd,rs:Reg; I:InstrSeq; 
      	       	       	P,P':Program)
		       P = (H,(R,(a,(iseq (unfold rd rs) I)))) ->
		       (WFProgram P) ->
		       (TrProgram P S) ->
		       (Eval P P') ->
		       (WFProgram P') ->
		       (TrProgram P' (Step S)).
Proof.
Intros St.
Intros.
Rewrite H0 in H1.
Inversion H1.
Clear H14.
Rewrite H0 in H2.
Inversion H2.
Unfold Step.
Cut (TrInstrSeq L (iseq (unfold rd rs) I) M pc).
Intros.
Inversion H25.
Rewrite H30.
Inversion H28.
Rewrite H0 in H3.
Inversion H3.
Apply trprogram with L.
Auto.
Auto.
Unfold updatereg.
Constructor.
Intros.
Compare _r _rd.
Intros.
Replace (beq_reg _r _rd) with true.
Replace (plus (_R _rs) O) with (plus O (_R _rs)).
Unfold plus.
Replace v0 with v.
Inversion H22.
Cut (TrWordVal L (wfold v t0) (_R _rs)).
Intros.
Inversion H56.
Auto.
Apply H52 with rs.
Auto.
Auto.
Apply (mlookup_eq ? ? r v v0 R').
Apply eqorneq_reg.
Replace r with rd.
Apply mupdext_imp_mlookup with R.
Apply eqorneq_reg.
Auto.
Apply trregf_eq_eq with _r.
Rewrite e.
Auto.
Auto.
Auto.
Apply plus_sym.
Rewrite e.
Apply sym_eq.
Apply beq_reg_eq_true.
Intros.
Replace (beq_reg _r _rd) with false.
Inversion H22.
Apply H52 with r.
Apply mupdext_preserv_rev with R' rd v.
Auto.
Auto.
Apply trregf_neq_neq with _r _rd.
Auto.
Auto.
Auto.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Auto.
Constructor.
Unfold updatereg.
Replace (beq_reg _r31 _rd) with false.
Inversion H23.
Auto.
Apply sym_eq.
Apply beq_reg_neq_false.
Cut ~_rd=_r31.
Auto.
Apply trreg_neq_31 with rd.
Auto.
Apply tr_pc_Spc with (unfold rd rs).
Auto.
Apply tr_pc_trinstrseq with H.
Auto.
Save.




Theorem Preservation : (S:State) (Inv S) -> (Inv (Step S)).
Proof.
Intros.
Elim H; Clear H; Intros P H.
Elim H; Clear H; Intros D H.
Cut (EX P':? | (Eval P P')).
Intros.
Elim H0; Clear H0; Intros P' H0.
Cut (WFProgram P').
Intros D'.
Split with P'.
Split with D'.
Inversion D.
Inversion H5.
Apply Preservation_add with H1 R l rd rs rs' I0 P.
Rewrite H15.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_addi with H1 R l rd rs i I0 P.
Rewrite H14.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_alloc with H1 R l rd V I0 P.
Rewrite H14.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_bgt with H1 R l rs rt l0 I0 P.
Rewrite H16.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_bump with H1 R l n I0 P.
Rewrite H12.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_fold with H1 R l rd rs (recty f) I0 P.
Rewrite H15.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_jd with H1 R l l0 P.
Rewrite H13.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_jmp with H1 R l r P.
Rewrite H13.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_ld with H1 R l rd rs i I0 P.
Rewrite H16.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_mov with H1 R l rd rs I0 P.
Rewrite H14.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_movi with H1 R l rd i I0 P.
Rewrite H13.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_movl with H1 R l rd l0 I0 P.
Rewrite H15.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_st with H1 R l rd rs i I0 P.
Rewrite H17.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply Preservation_unfold with H1 R l rd rs I0 P.
Rewrite H16.
Auto.
Auto.
Auto.
Auto.
Auto.
Apply preserv with P.
Auto.
Auto.
Apply progress.
Auto.
Save.


(*

Proofs needed:
  H1 : (Inv S0).
  H2 : (S:State) (Inv S) -> (SP S).
  H3 : (S:State) (Inv S) -> (Inv (Step S)).

  H1 must be supplied for each individual program.
  H2 and H3 can be proven once and used for all TAL programs compiled to TIS.

  Thus:
*)

Theorem Safety_aux : (S:State) (n:nat) (Inv S) -> (Inv (MultiStep n S)).
Proof.
Induction n.
Intros.
Unfold MultiStep.
Apply Preservation; Auto.

Intros.
Unfold MultiStep.
Fold MultiStep.
Apply Preservation.
Apply H.
Auto.
Save.


Theorem Safety : (S:State) (Inv S) -> (Safe S).
Proof.
Unfold Safe.
Induction n.
Apply Progress.
Unfold MultiStep.
Apply Preservation; Auto.

Intros.
Unfold MultiStep.
Fold MultiStep.
Apply Progress.
Apply Preservation.
Apply Safety_aux.
Auto.
Save.





(***************************************************************************)
(***************************************************************************)

